Package: PICB
Title: piRNA Cluster Builder
Version: 1.4.0
Authors@R: c(person("Pavol", "Genzor",  role = c("aut")),
    person("Aleksandr", "Friman", email="haase-lab-bioinfo@nih.gov",  role = c("aut")),
    person("Daniel", "Stoyko", role = c("aut")),
    person("Parthena", "Konstantinidou", email="haase-lab-bioinfo@nih.gov", role = c("aut")),
    person("Franziska", "Ahrend", email="haase-lab-bioinfo@nih.gov",  role = c("aut", "cre"), comment=c(ORCID = "0009-0004-7464-3444")),
    person("Zuzana", "Loubalova", email="haase-lab-bioinfo@nih.gov",  role = c("aut")), 
    person("Yuejun", "Wang", email="haase-lab-bioinfo@nih.gov",  role = c("aut")),
    person("Hernan", "Lorenzi", email="haase-lab-bioinfo@nih.gov",  role = c("aut")),
    person("Astrid D", "Haase", email="haase-lab-bioinfo@nih.gov",  role = c("aut")))
Description: piRNAs (short for PIWI-interacting RNAs) and their PIWI protein partners play a key role in fertility 
    and maintaining genome integrity by restricting mobile genetic elements (transposons) in germ cells. 
    piRNAs originate from genomic regions known as piRNA clusters. The piRNA Cluster Builder (PICB) is 
    a versatile toolkit designed to identify genomic regions with a high density of piRNAs. It constructs 
    piRNA clusters through a stepwise integration of unique and multimapping piRNAs and offers wide-ranging 
    parameter settings, supported by an optimization function that allows users to test different parameter 
    combinations to tailor the analysis to their specific piRNA system. The output includes extensive metadata 
    columns, enabling researchers to rank clusters and extract cluster characteristics. 
License: CC0
Encoding: UTF-8
Imports: utils, Seqinfo, GenomicRanges, GenomicAlignments, Biostrings,
        Rsamtools, data.table, IRanges, seqinr, stats, openxlsx, dplyr,
        S4Vectors, methods
Suggests: GenomeInfoDb, knitr, rtracklayer, testthat, BiocStyle,
        prettydoc, BSgenome, BSgenome.Dmelanogaster.UCSC.dm6,
        BiocManager, rmarkdown, ggplot2
VignetteBuilder: knitr
BugReports: https://github.com/HaaseLab/PICB/issues
URL: https://github.com/HaaseLab/PICB
biocViews: Genetics, GenomeAnnotation, Sequencing,
        FunctionalPrediction, Coverage, Transcriptomics
RoxygenNote: 7.3.2
git_url: https://git.bioconductor.org/packages/PICB
git_branch: RELEASE_3_23
git_last_commit: e2dfa54
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 23:53:36 UTC; biocbuild
Author: Pavol Genzor [aut],
  Aleksandr Friman [aut],
  Daniel Stoyko [aut],
  Parthena Konstantinidou [aut],
  Franziska Ahrend [aut, cre] (ORCID:
    <https://orcid.org/0009-0004-7464-3444>),
  Zuzana Loubalova [aut],
  Yuejun Wang [aut],
  Hernan Lorenzi [aut],
  Astrid D Haase [aut]
Maintainer: Franziska Ahrend <haase-lab-bioinfo@nih.gov>

# Generated by roxygen2: do not edit by hand

export(PICBannotate)
export(PICBbuild)
export(PICBexporttoexcel)
export(PICBgetchromosomes)
export(PICBimportfromexcel)
export(PICBload)
export(PICBloadfasta)
export(PICBoptimize)
export(PICBstrandanalysis)

<!-- Style of markdown used from https://github.com/aregtech/areg-sdk --> 
<!-- Recommended live preview: https://markdownlivepreview.com/ --> 
<!-- Logo-->
<h1 align="center" style="display: block; font-size: 2.5em; font-weight: bold; margin-block-start: 1em; margin-block-end: 1em;">

<div align="right">
    <a href="https://www.niddk.nih.gov/research-funding/at-niddk/labs-branches/laboratory-cell-molecular-biology/rna-biology-section">
    <img src="https://upload.wikimedia.org/wikipedia/commons/a/a8/National_Institute_of_Diabetes_and_Digestive_and_Kidney_Diseases_%28NIDDK%29_Logo.png" alt="Visit our Lab Page" height="60">
    </a>
</div>

  <br /><br /><strong>PICB - <u>pi</u>RNA <u>C</u>luster <u>B</u>uilder</strong>
</h1>

## Introduction[![](vignettes/double-helix-svgrepo-com.svg)](#introduction)

**PICB** (*piRNA Cluster Builder*) is a flexible toolkit for assembling, prioritizing, and characterizing piRNA clusters. 

<!--[source to read more](https://). --> 

---

## Table of contents[![](vignettes/double-helix-svgrepo-com.svg)](#table-of-contents)
- [Motivation](#motivation)
- [Getting started](#getting-started)
- [How to run PICB](#how-to-run-PICB)
    - [Data preparation](#data-preparation)
    - [Running PICB](#running-PICB)
- [Parameter adjustments](#parameter-adjustments)
- [Output](#output)
- [Other PICB functions](#other-picb-functions)
- [Let's give it a try - An Example](#lets-give-it-a-try---an-example)
- [Authors, Citation and Acknowledgments](#authors-citation-and-acknowledgments)

---

## Motivation[![](vignettes/double-helix-svgrepo-com.svg)](#motivation)

piRNAs (PIWI-interacting RNAs) and their PIWI protein partners play a key role in fertility and maintaining genome integrity by restricting mobile genetic elements (transposons) in germ cells. piRNAs originate from genomic regions which are called _piRNA clusters_.

PICB identifies genomic regions with a high density of piRNAs. This construction of piRNA clusters is performed through stepwise integration of unique and multimapping piRNAs. 

<div align="center"><a href="https://www.sciencedirect.com/science/article/pii/S2211124724011288#sec2"><img src="vignettes/PICB_stepwiseIntegration.jpg" alt="Stepwise Integration for PICB" style="width:50%;height:50%"/></a></div>

Figure 1: PICB considers unique mapping piRNAs (NH=1), primary alignments of multimapping piRNAs (NH>1), and all possible alignments stepwise to build seeds, cores, and clusters. Find additional information in our <a href="https://www.sciencedirect.com/science/article/pii/S2211124724011288" target="_blank"> recent publication</a>.

Only very limited programming knowledge is needed to run PICB. Check out our step-by-step instructions and our [demonstration](#lets-give-it-a-try---an-example) below. 

Please visit our <a href="https://www.sciencedirect.com/science/article/pii/S2211124724011288" target="_blank">publication</a> for full context.

<div align="right">[ <a href="#table-of-contents">↑ Back to top ↑</a> ]</div>

---

## Getting started[![](vignettes/double-helix-svgrepo-com.svg)](#getting-started)

PICB runs in R versions <span>&#8805;</span> 4.2. to 4.4.

It is possible to run PICB in RStudio, in an R script on your local machine or with High Performance Computing (HPC) resources or in Jupyter Notebook (using R). Keep in mind that as for any handling of large-scale sequencing data you need to have sufficient memory on your device (or cluster) allocated. 

<b>1. Load dependencies in R environment</b>

You will need to install and load the following required R packages:

```R
install.packages(c("data.table", "seqinr", "openxlsx", "dplyr"))
if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install(c("IRanges", "GenomicRanges", "GenomicAlignments", "Rsamtools", "Biostrings", "Seqinfo", "BSgenome", "rtracklayer"))
```
> 💡 In case you have not worked with GRanges yet, we recommend reading the following <a href=https://bioconductor.org/packages/release/bioc/vignettes/GenomicRanges/inst/doc/GenomicRangesIntroduction.html target="_blank">GRanges Introduction</a>.



<b>2. Install PICB</b>

PICB is available to install from any of the following sources. If you choose to install from GitHub, make sure to have either devtools or remotes installed.

| Where        | Source   | Command                                                                     |
|-------------|----------|-----------------------------------------------------------------------------|
| R | GitHub   | From GitHub repository: `devtools::install_github("HaaseLab/PICB")` or `remotes::install_github("HaaseLab/PICB")` |
| Web Browser | GitHub   | <a href="https://github.com/HaaseLab/PICB/archive/refs/heads/main.zip">Download GitHub repository here.</a> Now unzip the file and run `install.packages("Downloads/PICB-main", repos=NULL, type="source")` in R.  |
| Terminal | GitHub   | From GitHub repository: Clone Source Code: `git clone https://github.com/HaaseLab/PICB.git` <!--<br> In R: `install.packages()`--> |
<!--| R     | Bioconductor     | Soon: BiocManager::install("PICB") |-->


Now load PICB in your R environment: 
```R
library('PICB')
```


From now on it gets even easier. 

---

## How to run PICB[![](vignettes/double-helix-svgrepo-com.svg)](#how-to-run-PICB)

There are just two required inputs: the **BAM File** and the **Reference Genome**.

### Data preparation

Checklist for having the right **BAM File**

- [ ] NH and NM tags are included
- [ ] Indexed (.bai file required) 

PICB stops when these requirements are not met and provides a descriptive error message.

**Four options for providing the _Reference Genome_**


1. Your genome is part of the <a href= https://kasperdanielhansen.github.io/genbioconductor/html/BSgenome.html target="_blank" rel="noopener noreferrer">BSgenome</a> package
  
```R
# Example for Drosophila melanogaster, replace with your own genome - previous installation of BSgenome required
library(BSgenome.Dmelanogaster.UCSC.dm6)
myGenome <- "BSgenome.Dmelanogaster.UCSC.dm6"
```

2. Chromosome names and lengths according to the _BAM file_</label>


```R
myGenome <- Seqinfo::Seqinfo(
    seqnames = c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY"), 
    seqlengths = c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352))
```

3. **Using a Seqinfo object**

Or use an existing `Seqinfo` object:

```R
myGenome <- Seqinfo::Seqinfo(genome = "dm6")
```


4. _Fasta_ with the assembled genome sequence.
```R
myGenome <- PICBloadfasta(FASTA.NAME="/path/to/your/genome.fa")
```




<!--</details>-->

### Running PICB


1. Load your mapped piRNAs with `PICBload`

```R
myAlignments <- PICBload(
    BAMFILE = bam_file,
    REFERENCE.GENOME = myGenome
)
```


2. Build piRNA clusters with `PICBbuild`

```R
myClusters <- PICBbuild(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome
)$clusters
```

> 💡 Both `PICBload` and `PICBbuild` allow wide-ranging adjustments: Read the below section [Parameter adjustments](#parameter-adjustments) to adapt to sparse reference genomes and specific limitations of the data set.

> 💡 As described [here](#motivation), `PICBbuild` integrates unique mapping (seeds), primary multimapping (cores) and secondary alignments stepwise. You can access the outputs of the previous steps by using `$seeds` or `$cores` instead of `$clusters`.

<!--</details>-->

---

### Parameter adjustments[![](vignettes/double-helix-svgrepo-com.svg)](#parameter-adjustments)


PICB allows wide-ranging parameter adjustments to adapt to e.g. sparse reference genomes and specific limitations of the data set. Tables of adjustments for both functions are shown below. 

<details close><summary title="Click to show/hide details">Click to show / hide: Parameters for <code>PICBload</code></summary><br/>

Required parameters: 
* BAMFILE
* REFERENCE.GENOME 

<br>

Adjustable parameters:


| Parameter Name | Possible Values | Default Value | Explanation |
|----------------|-----------------|---------------|-------------|
| VERBOSE | TRUE, FALSE  | TRUE | Allows disabling progress messages while running `PICBload`. |
| IS.SECONDARY.ALIGNMENT | TRUE, FALSE, NA | NA  (all alignments) | Determines which alignment types (primary multimappers and secondary multimappers) will be loaded. |
| SEQ.LEVELS.STYLE | _character_ | "UCSC" | Allows changing the style of chromosome names in the output. Supported styles are listed in GenomeInfoDb::genomeStyles(). NA allows keeping all chromosomes from REFERENCE.GENOME and does not filter by standard linear chromosome names. |
| STANDARD.CONTIGS.ONLY   | TRUE, FALSE  | TRUE        | Determines whether alignments from non-standard contigs are used. |
| FILTER.BY.FLAG   | TRUE, FALSE | TRUE | Allows only those alignments with flag values present in the vector of allowed flags SELECT.FLAG. Default values of SELECT.FLAG are 0, 16, 272 and 256 (primary and secondary alignments on plus and minus strands). If FALSE, includes all flags. |
| USE.SIZE.FILTER   | TRUE, FALSE | TRUE | Allows filtering of alignments based on size. Default value is 18-50 nt. |
| TAGS   | _vector_  | c("NH","NM")  | Indicates list of tags to be extracted from given bam file. The “NH” tag is required to deduce if the alignment is unique mapping or multimapping. The “NM” is required to identify mismatches if required in further analysis. |
| GET.ORIGINAL.SEQUENCE | TRUE, FALSE  | FALSE  | Allows extraction of original read sequence from the bam file. |
| SIMPLE.CIGAR  | TRUE, FALSE  | TRUE  | Allows filtering out spliced alignments. |
| PERFECT.MATCH.ONLY   | TRUE, FALSE  | FALSE  | Allows filtering out alignments with mismatches. |
| WHAT   | _vector_ | c(“flag”)         | Allows importing flags. Corresponds to “what” from ScanBamParam-class [Morgan M, Pagès H, Obenchain V, Hayden N (2023). Rsamtools: Binary alignment (BAM), FASTA, variant call (BCF), and tabix file import]  |


</details>



<details close><summary title="Click to show/hide details">Click to show / hide: Parameters for <code>PICBbuild</code></summary><br/>

Required parameters: 
* IN.ALIGNMENTS (output of `PICBload`)
* REFERENCE.GENOME (reference genome previously used in `PICBload`)

The library size can be adjusted as shown in our [PICB demo](#lets-give-it-a-try---an-example), however in most cases this is not necessary.

| Parameter Name | Possible Values | Default Value | Explanation |
|----------------|-----------------|---------------|-------------|
| LIBRARY.SIZE | _numeric_  |  number of unique mapping alignments + number of primary multimapping alignments  | number of reads in the library |
| VERBOSITY | 0,1,2,3  |  2  | Allows choosing the quantity of progress messages while running `PICBbuild`. Depending on VERBOSITY's value, printed messages are missing (0), include current processing step (1), include additionally current processing sub-step (2) or include chosen parameters for `PICBbuild`)|
| PROVIDE.NON.NORMALIZED | TRUE, FALSE  |  FALSE  | Includes non-normalized statistics in the output annotations|
| COMPUTE.1U.10A.FRACTIONS | TRUE, FALSE  |  FALSE  | Calculates the fractions of piRNAs with a 1U (Uridine at the 5' most piRNA position) and a 10A (adenine at position 10). Requirement: GET.ORIGINAL.SEQUENCE set to TRUE in `PICBload`|

<br>

PICBbuild integrates unique mapping (seeds), primary multimapping (cores) and secondary alignments stepwise using the sliding window algorithm. Each step allows following parameter adjustments. 

Adjustable parameters for processing **unique mapping alignments**:


| Parameter Name | Possible Values | Default Value | Explanation |
|----------------|-----------------|---------------|-------------|
| UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH | _numeric_  | 350 nt   | Sets length of sliding windows for uniquely mapping alignments. |
| UNIQUEMAPPERS.SLIDING.WINDOW.STEP | _numeric_  | UNIQUEMAPPERS.SLIDING. WINDOW.WIDTH divided by 10 and rounded to the nearest numeric | Sets distance between starts of the windows for uniquely mapping alignments. |
| MIN.UNIQUE.ALIGNMENTS.PER.WINDOW | _numeric_ | Value corresponding to 2 FPKM  | Sets coverage threshold for seed discovery. Corresponds to normalized counts of uniquely mapping alignments throughout the window (in FPKM). |
| MIN.UNIQUE.SEQUENCES.PER.WINDOW | _numeric_  | 1 per 50 nt of window length or MIN.UNIQUE.ALIGNMENTS.PER.WINDOW whichever is smaller | Minimum number of distinct piRNA sequences within the sliding window. This parameter allows identification of clusters supported by a diverse set of piRNA sequences rather than a high number of reads from one or a few piRNA sequences.  |

The called windows are reduced into **seeds** and further filtered: 

| Parameter Name | Possible Values | Default Value | Explanation |
|----------------|-----------------|---------------|-------------|
| THRESHOLD.SEEDS.GAP  | _numeric_  | 0 nt | Removes gaps between seeds if below the given length value. |
| MIN.SEED.LENGTH | _numeric_  | 2 * UNIQUEMAPPERS.SLIDING. WINDOW.WIDTH + 100 nt = 800 nt | Removes seeds below a certain length (Default requires an actual piRNA coverage of at least 100 nt: UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH (see above, default: 350 nt) at both the 5’ end and the 3’ end, thus making a 350+350+100=800 nt long seed). |
| MIN.COVERED.SEED.LENGTH   | _numeric_ | 0 nt | Allows filtering by minimum number of seed bases covered by unique mapping alignments. |

The next step includes **primary multimapping alignments** using a similar but simplified algorithm. Following parameters can be adjusted:


| Parameter Name | Possible Values | Default Value | Explanation |
|----------------|-----------------|---------------|-------------|
| PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH  | _numeric_  | 350 nt  | Sets lengths of sliding windows for primary multimapping alignments. |
| PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP  | _numeric_ | PRIMARY.MULTIMAPPERS. SLIDING.WINDOW.WIDTH divided by 10 and rounded to the nearest integer | Sets distances between starts of windows for primary multimapping alignments. |
| MIN.PRIMARY.MULTIMAPING.ALIGNMENTS.PER.WINDOW  | _numeric_ | Value corresponding to 4 FPKM | Sets coverage threshold for calling primary multimapping alignments windows. Corresponds to normalized counts of primary multimapping alignments throughout the window (in FPKM).|

These resulting windows are being reduced into **cores** and are further filtered:

| Parameter Name | Possible Values | Default Value | Explanation |
|----------------|-----------------|---------------|-------------|
| THRESHOLD.CORES.GAP  | _numeric_ | 0 nt  | Removes gaps between cores if below the given length value. |

**Seeds** and **primary multimapping windows** overlapping **seeds** merge into **cores**. Standalone **seeds** are also considered **cores**. **Cores** not overlapping with a **seed** are being filtered out as their transcription cannot be verified.

In the next step **secondary alignments** are processed identically as primary multimapping alignments. This step is dependent on whether `PICBload` loaded the secondary alignments (parameter *IS.SECONDARY.ALIGNMENT*).

| Parameter Name | Possible Values | Default Value | Explanation |
|----------------|-----------------|---------------|-------------|
| SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH  | _numeric_  | 1000 nt  | Sets lengths of sliding windows for secondary alignments. |
| SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP  | _numeric_ | 100 nt | Sets distances between starts of the windows for secondary alignments. |
| MIN.SECONDARY.MULTIMAPING.ALIGNMENTS.PER.WINDOW  | _numeric_ | Value corresponding to  0.2 FPKM | Sets coverage threshold for calling secondary alignments windows. Corresponds to normalized counts of secondary alignments throughout the window (in FPKM). |

**Cores** and **secondary alignments** overlapping **cores** merge into **clusters**. Standalone **cores** are also considered **clusters**. piRNA **clusters** were formed and include all alignment information. 


</details>

---

## Output[![](vignettes/double-helix-svgrepo-com.svg)](#output)

The output of `PICBbuild` includes **seeds**, **cores** and  **clusters**, each in *GenomicRanges* format. In [Running PICB](#running-PICB), we extracted directly the **clusters** using `$clusters`. Extracting the **seeds** and **cores** can be done similarly using `$seeds` and `$cores`. 

The **clusters** follow *GRanges* convention including the genomic coordinates (seqnames, ranges, and strand) and metacolumns. There are in total 9 metacolumns when running PICB with default parameters:

<details close><summary title="Click to show/hide details">Click to show / hide: Output columns of PICB</summary><br/>


| Output Name | Explanation |
|----------------|-----------------|
| type | Type of cluster with possible values: 'SingleCore', 'ExtendedCore', 'MultiCore' |
| width_in_nt | Width of the seed/core/cluster |
| uniq_reads_FPM | Number of uniquely mapping piRNA reads aligned to the seed/core/cluster (normalized to the number of all aligned reads) |
| multimapping_reads_primary_alignments_FPM | Number of primary multimapping alignments overlapping to the seed/core/cluster (normalized to the number of all aligned reads) |
| all_reads_primary_alignments_FPM | Number of primary alignments overlapping the seed/core/cluster (normalized to the number of all aligned reads) |
| uniq_reads_FPKM | Number of unique reads normalized to the number of all aligned reads and to the size of the corresponding seed/core/cluster length (unique reads FPM per kilobase) |
| multimapping_reads_primary_alignments_FPKM | Number of multimapping reads normalized to the number of all aligned reads and to the size of the corresponding seed/core/cluster length (primary alignments of multimapping reads FPM per kilobase) |
| all_reads_primary_alignments_FPKM | Number of all reads primary alignments normalized to the number of all aligned reads and to the size of the corresponding seed/core/cluster length (primary alignments of all reads FPM per kilobase) |
| fraction_of_width_covered_by_unique_alignments | Fraction of seed/core/cluster length coreved by at least 1 unique mapping piRNA read |

<!-- 
| uniq_reads | Number of uniquely mapping piRNA reads aligned to the seed/core/cluster |
| uniq_reads_FPM | Uniquely mapping piRNA reads aligned to the seed/core/cluster (normalized to the number of all aligned reads) |
| multimapping_reads_primary_alignments_FPM | Primary multimapping alignments overlapping the seed/core/cluster (normalized to the number of all aligned reads) |
| all_reads_primary_alignments_FPM | All primary alignments overlapping the seed/core/cluster (normalized to the number of all aligned reads)|
| width_covered_by_unique_alignments | Number of base pairs of seed/core/cluster length coreved by at least 1 unique mapping piRNA read | 
| uniq_sequences | Number of different unique mapping piRNA sequences aligned to the seed/core/cluster |-->


</details>
<div align="right">[ <a href="#table-of-contents">↑ Back to top ↑</a> ]</div>

---

## Other PICB functions[![](vignettes/double-helix-svgrepo-com.svg)](#other-picb-functions)

### Optimization of parameters

You can use `PICBoptimize` to find the optimal parameters for `PICBbuild` for your dataset.
This is especially useful if you are not sure about the quality of your data or your reference genome.
The goal is to find a set of parameters that maximize the number of piRNAs accomodated in the clusters while minimizing genomic space occupied by the clusters. 

Make sure to provide the values for the parameters as a vector. An example is shown below:

```R
parameterExploration <- PICBoptimize(
    IN.ALIGNMENTS = myAlignments, 
    REFERENCE.GENOME=myGenome, 
    MIN.UNIQUE.ALIGNMENTS.PER.WINDOW=c(1,2,3,4,5)
)
```

This function works on numerical parameters and returns a dataframe with any combination of your chosen parameters and their corresponding number of clusters, the total width of the clusters in nucleotides, the number of reads explained by the clusters, the mean RPKM of the clusters, the fraction of reads explained by the clusters and fraction of genomic space occupied by the clusters.

Be aware that for most datasets, the default parameters will yield great results and you do not need to optimize parameters. 

### Strand analysis

Next to the standard analysis, `PICBstrandanalysis` allows strand-specific analysis of piRNA clusters. The function computes the sense/antisense ratio of piRNAs per cluster (non-collapsed) and adds this information to a new metadata column (`s_as_ratio`). 

```R
myClustersWithStrandAnalysis <- PICBstrandanalysis(
    IN.ALIGNMENTS = myAlignments,
    IN.RANGES = myClusters
)
```

To visualize the sense/antisense ratio of the clusters, you can plot the sense/antisense ratio of the clusters in a violin plot. Higher numbers indicate a higher ratio of sense to antisense piRNAs, while values close to 1 indicate an equal ratio of sense and antisense piRNAs.


--- 


## Let's give it a try - An Example[![](vignettes/double-helix-svgrepo-com.svg)](#lets-give-it-a-try---an-example)

In the following, we would like to show you how easy it is to run PICB!

<b>We created a demo to show you the most basic workflow with PICB.</b> Download the sample subset of mapped reads and the corresponding R-based code in R Script or Jupyter Notebook format from the folder `demo`. Though, if you feel extra brave today, feel free to direcly use your own bam file!


<b>Step 1: Getting started with PICB </b>
* PICB-Tutorial 1/3: Installation of PICB 

Follow the steps in [Getting started](#getting-started) to install PICB. Do not forget to load PICB with `library(PICB)`.

<b>Step 2: Data preparation </b>
* PICB-Tutorial 2/3: Preparation of input

We showed different ways on how to load your genome. In the following the variant with using the  <a href= https://kasperdanielhansen.github.io/genbioconductor/html/BSgenome.html target="_blank" rel="noopener noreferrer">BSgenome</a> package. You will need to install your specific genome first. In our demo it is the <i>Drosophila melanogaster</i> genome. 

```R
library("BSgenome.Dmelanogaster.UCSC.dm6")
```
Now, let's store the genome into the variable `myGenome`.

```R
myGenome <- "BSgenome.Dmelanogaster.UCSC.dm6"
```

<b>Step 3: Running PICB</b>
* PICB-Tutorial 3/3: Running PICBload and PICBbuild

Load the alignments with `PICBload`.
```R
myAlignments <- PICBload(
    BAMFILE = system.file("extdata","Fly_Ov1_chr2L_20To21mb_filtered.bam", package="PICB"), 
    REFERENCE.GENOME = myGenome)
```
Next, we want to form the piRNA clusters using the `PICBbuild` function. Usually you would not need to include the size of the library (`LIBRARY.SIZE`) since PICB calculates it automatically. However, just for this demo, please include this parameter to adjust to the subset we chose to create this demo. 

```R
myClusters <- PICBbuild(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    LIBRARY.SIZE = 12799826 #usually not necessary
)$clusters
```

<!--Possibilities of optimizing the parameters to adjust to the quality of sequencing reads and reference genome are described in the [protocol]().--> We note that ranking piRNA clusters is essential for proper interpretation. 

> 💡 Keep in mind that you just run a demo. These are not representative clusters! Though you can apply these steps to your organism of choice and sequencing data. Have fun with PICB and your piRNA clusters!

---



## Authors, Citation and Acknowledgments[![](vignettes/double-helix-svgrepo-com.svg)](#authors-citation-and-acknowledgments)

Special thanks to all contributors and supporters that starred this repository.

**Our authors:**

<a href="https://github.com/alexfriman"><img src="https://avatars.githubusercontent.com/u/20302553?v=4" alt="Aleksandr (Alex) Friman, M.Sc. M.Eng." style="width:8%"/></a>
<a href="https://github.com/frahrend"><img src="https://avatars.githubusercontent.com/u/15805954?v=4" alt="Franziska Ahrend, M.Sc." style="width:8%"/></a>

**Our PICB-team:**

<a href="https://github.com/thenoulav"><img src="https://avatars.githubusercontent.com/u/79937315?v=4" alt="
Parthena (Thenia) Konstantinidou, Ph.D." style="width:8%"/></a>
<a href="https://github.com/LoubalovaZ"><img src="https://avatars.githubusercontent.com/u/120222151?v=4" alt="Zuzana Loubalova, Ph.D." style="width:8%"/></a>
<a href="https://github.com/frahrend"><img src="https://avatars.githubusercontent.com/u/15805954?v=4" alt="Franziska Ahrend, M.Sc." style="width:8%"/></a>
<a href="https://github.com/alexfriman"><img src="https://avatars.githubusercontent.com/u/20302553?v=4" alt="Aleksandr (Alex) Friman, M.Sc. M.Eng." style="width:8%"/></a>
<a href="https://www.niddk.nih.gov/about-niddk/staff-directory/biography/haase-astrid"><img src="https://www.niddk.nih.gov/-/media/Images/About-NIDDK/Profile-Images/Profile-Images---Intramural/Haase-Astrid_300x300.jpg?h=300&iar=0&w=300&hash=EDEF30948C30B9F1E0922288AFCB4C9F" alt="Astrid D. Haase, M.D., Ph.D." style="width:8%"/></a><br>
Visit the lab website of 
<a href="https://www.niddk.nih.gov/about-niddk/staff-directory/biography/haase-astrid">Astrid D. Haase, M.D., Ph.D.</a>

**Our Co-authors and support:**

<a href="https://www.linkedin.com/in/miguel-vasconcelos-almeida-bb93b6140/?originalSubdomain=uk"><img src="https://scholar.googleusercontent.com/citations?view_op=view_photo&user=ON00Iz8AAAAJ&citpid=1" alt="Miguel Vasconcelos Almeida" style="width: 62px; height: 62px; object-fit: cover;"/></a>
<a href="https://github.com/PouletAxel"><img src="https://avatars.githubusercontent.com/u/6670853?v=4" alt="Axel Poulet" style="width: 8%;"/></a>
<a href="https://www.img.cas.cz/group/petr-svoboda/"><img src="https://bioinfo.medils.hr/images/5_Filip.png" alt="Filip Horvat" style="width: 8%;"/></a>
<a href="https://github.com/yuejun-j-wang"><img src="https://avatars.githubusercontent.com/u/110154696?v=4" alt="Yuejun Wang" style="width: 8%;"/></a>
<a href="https://umdphysics.umd.edu/people/faculty/current/item/318-wlosert.html"><img src="https://cmns.umd.edu/sites/default/files/styles/square/public/images/news/wolfgang_losert-featnews_1.jpg?itok=CxiFSmSQ" alt="Wolfgang Losert" style="width: 8%;"/></a>
<a href="https://www.niddk.nih.gov/about-niddk/staff-directory/biography/lorenzi-hernan"><img src="https://www.niddk.nih.gov/-/media/Images/About-NIDDK/Profile-Images/Profile-Images---Intramural/Lorenzi_Hernan_600x600.jpg?h=600&iar=0&w=600&hash=98CDD384341B767DA39C3BA3D2F298C5" alt="Hernan Lorenzi" style="width: 8%;"/></a>
<a href="https://www.img.cas.cz/group/petr-svoboda/"><img src="https://www.img.cas.cz/files/2022/02/portrait-svoboda.jpg" alt="Petr Svoboda" style="width: 8%;"/></a>
<a href="https://www.gen.cam.ac.uk/directory/eric-miska"><img src="https://www.sanger.ac.uk/userpics/400/em13.jpg" alt="Eric A. Miska" style="width: 8%;"/></a>
<a href="https://mcdb.yale.edu/people/josien-van-wolfswinkel"><img src="https://vanwolfswinkellab.org/wp-content/uploads/bb-plugin/cache/WebsitePic_alt-circle.png" alt="Josien van Wolfswinkel" style="width: 8%;"/></a>

<br>

This project is licensed under the CC0.1.0 license - see the [LICENSE](/LICENSE) file for details.<br/>

Do you like this project? Please join us or [give a ⭐](/stargazers). Let us make piRNA clusters more comparable and easy to build!<br/>

<div align="right">[ <a href="#table-of-contents">↑ Back to top ↑</a> ]</div>

---

### List of papers utilizing PICB:

- [A comparative roadmap of PIWI-interacting RNAs across seven species reveals insights into de novo piRNA-precursor formation in mammals](https://www.sciencedirect.com/science/article/pii/S2211124724011288) (2024) by Parthena Konstantinidou*, Zuzana Loubalova*, Franziska Ahrend*, Aleksandr Friman*, Miguel Vasconcelos Almeida, Axel Poulet, Filip Horvat, Yuejun Wang, Wolfgang Losert, Hernan Lorenzi, Petr Svoboda, Eric A. Miska, Josien C. van Wolfswinkel, Astrid D. Haase*.
- [Dynamic co-evolution of transposable elements and the piRNA pathway in African cichlid fishes](https://www.biorxiv.org/content/10.1101/2024.04.01.587621v1) (2024) by Miguel Vasconcelos Almeida*, Moritz Blumer, Chengwei Ulrika Yuan, Pío Sierra, Jonathan L. Price, Fu Xiang Quah, Aleksandr Friman, Alexandra Dallaire, Grégoire Vernaz, Audrey L. K. Putman, Alan M. Smith, Domino A. Joyce, Falk Butter, Astrid D. Haase, Richard Durbin, M. Emília Santos, Eric A. Miska*.

<b> 

<div align="right">[ <a href="#table-of-contents">↑ Back to top ↑</a> ]</div>

<br> <br><br>

<br><br>
<br><br>
<!-- markdownlint-enable -->

# =======================================
# Script Name: PICB_demo.R
# Purpose: Demonstration of the PICB workflow using a processed subset 
#          of small RNA data from Drosophila ovaries.
# Authors: PICB Team of the Haase Lab
# Date: 2024-12-04
#
# Description:
# This script demonstrates analysis on a reduced dataset, 
# Fly_Ov1_chr2L_20To21mb_filtered.bam, created from small RNA reads 
# aligned to Drosophila melanogaster genome (Dm6). The data preparation 
# includes trimming, filtering, mapping, and downsampling as detailed below.
#
# Compatible Environments:
# - ✅ R versions 4.2 - 4.4
# - ✅ R with R base
# - ✅ R with IDE RStudio
# - ✅ R with IDE Cursor AI
#
# =======================================

# Data Preparation Context:
# - Processed data: Fly_Ov1_chr2L_20To21mb_filtered.bam
# - Details:
#   - Trimmed 3' adaptors, filtered reads ≥24 nt.
#   - Removed abundant cellular RNAs (rRNAs, tRNAs, snRNAs, snoRNAs).
#   - Mapped reads to Dm6 genome (chr2L:20000000-21000000) using STAR.
#   - Retained essential tags (NH, NM) and compressed BAM file.
#   - Downsampled to 22% of reads for a smaller, manageable dataset.
#   - Indexed and renamed the file.

# Disclaimer:
# The processed data is intended for demonstration purposes only and 
# is not representative of biologically relevant samples due to 
# extensive downsampling and regional restriction.

# Additional Notes:
# The Jupyter Notebook version of this script is available in the 
# corresponding folder. Data preparation commands can be found on the 
# [PICB GitHub page](https://github.com/HaaseLab/PICB).

# =======================================

# Check R version compatibility
message("Using R version: ", R.version.string)

# Install required packages
# =======================================
# CRAN packages
cran_packages <- c("data.table", "seqinr", "openxlsx", "dplyr")
if (length(setdiff(cran_packages, installed.packages()[,"Package"])) > 0) 
    install.packages(setdiff(cran_packages, installed.packages()[,"Package"]), 
    quiet = TRUE) else 
    message("All required CRAN packages are already installed.")
# Bioconductor packages
if (!require("BiocManager", quietly = TRUE)) install.packages("BiocManager")

bio_packages <- c("BiocManager", "IRanges", "GenomicRanges", "GenomicAlignments", "Rsamtools", "Biostrings", "Seqinfo", "BSgenome", "rtracklayer")
if (length(setdiff(bio_packages, installed.packages()[,"Package"])) > 0) BiocManager::install(setdiff(bio_packages, installed.packages()[,"Package"]), ask = FALSE) else message("All required Bioconductor packages are already installed.")


# Install PICB
# =======================================
# Select one of the following installation methods for the PICB package:
# Uncomment the preferred method.
# devtools::install_github("HaaseLab/PICB")
# remotes::install_github("HaaseLab/PICB")
# install.packages("../PICB", repos = NULL, type = "source")
# Use terminal to clone and build if required:
# git clone https://github.com/HaaseLab/PICB.git

# Load PICB package
library("PICB") 

#Core-PICB

# Load reference genome
# =======================================
# Provide the reference genome. Select one option:
# a. Install and load a BSgenome
# BiocManager::install("BSgenome.Dmelanogaster.UCSC.dm6")
library("BSgenome.Dmelanogaster.UCSC.dm6")
myGenome <- "BSgenome.Dmelanogaster.UCSC.dm6"
# b. Provide a Seqinfo object containing the chromosome names and lengths according to the BAM file.
# myGenome <- Seqinfo::Seqinfo(
#     seqnames = c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY"),
#     seqlengths = c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352)
# )
# c. Provide supported genome as Seqinfo object.
# myGenome <- Seqinfo::Seqinfo(genome = "dm6")
# d. Provide a fasta file containing the reference genome sequence.
# myGenome <- PICBloadfasta("dm6.fa")


# Load mapped piRNA data
# =======================================
# Load the demo BAM file and process it into GRanges format
bam_file <- system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB")
myAlignments <- PICBload(
    BAMFILE = bam_file,
    REFERENCE.GENOME = myGenome
)


# Build piRNA clusters
# =======================================
# Run PICBbuild to construct piRNA clusters.
myClusters <- PICBbuild(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    LIBRARY.SIZE = 12799826
)$clusters
# Note: Library size usually doesn't need to be set, but we are dealing with a sample that does not contain all reads, so normalization is altered.

# Optimize PICB parameters
# =======================================
parameterExploration <- PICBoptimize(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2, 3, 4, 5),
    LIBRARY.SIZE = 12799826,
    VERBOSITY = 2
)
# Note: Expect a warning due to the modified LIBRARY.SIZE, which does not match the number of reads in IN.ALIGNMENTS.

parameterExploration

# A thorough parameter exploration for a subset of alignments is not particularly useful here. Nevertheless, we provide the code below to visualize the exploration of different values for MIN.UNIQUE.ALIGNMENTS.PER.WINDOW for the reads ("library") and the genome space covered by piRNA clusters.
library(ggplot2)

#Specify parameter to optimize for.
x_column <- "MIN.UNIQUE.ALIGNMENTS.PER.WINDOW" #change parameter to optimize, if applicable

scaling_factor <- max(parameterExploration$fraction.of.library.explained.by.clusters) / max(parameterExploration$fraction.of.genome.space.clusters)
ggplot(parameterExploration, aes(x = .data[[x_column]])) +
    geom_line(aes(y = fraction.of.library.explained.by.clusters * 100, color = "piRNAs Explained"), linewidth = 1) +
    geom_point(aes(y = fraction.of.library.explained.by.clusters * 100, color = "piRNAs Explained"), size = 3) +
    geom_line(aes(y = fraction.of.genome.space.clusters * scaling_factor * 100, color = "Genome Space"), linewidth = 1) +
    geom_point(aes(y = fraction.of.genome.space.clusters * scaling_factor * 100, color = "Genome Space"), size = 3) +
    scale_y_continuous(name = "piRNAs Explained by Clusters (%, piRNA sample)", sec.axis = sec_axis(~ . / scaling_factor, name = "Total piRNA cluster-length (Genome, %)")) +
    scale_x_reverse(name = paste0("Parameter chosen: ", x_column), breaks = parameterExploration[[x_column]], labels = parameterExploration[[x_column]]) +
    scale_color_manual(name = "Metrics", values = c("piRNAs Explained" = "#00a100", "Genome Space" = "black")) +
    theme_classic() +
    theme(axis.title.y.left = element_text(color = "#00a100"), axis.title.y.right = element_text(color = "black"), legend.position = "top")

# if needed modify paramater in PICBbuild (not needed for PICB-demonstration)


# Rank piRNA clusters
# =======================================
# one option is to rank by primary alignments FPM:
myClusters <- myClusters[order(-mcols(myClusters)$all_reads_primary_alignments_FPM)]
mcols(myClusters)$rank_all_reads_primary_alignments_FPM <- rank(-mcols(myClusters)$all_reads_primary_alignments_FPM)


# Export results
# =======================================
# RDS format 
saveRDS(myClusters, file="clusters_PICB_demo.RDS")
# Microsoft Excel format
PICBexporttoexcel(IN.RANGES = myClusters, EXCEL.FILE.NAME = "clusters_PICB_demo.xlsx")
# GTF format 
myClusters$transcript_id <- paste0("PICB_ rank_", myClusters$rank_all_reads_primary_alignments_FPM)
rtracklayer::export(myClusters, "clusters_PICB_demo.gtf")

#' Get SeqInfo object from a fasta file
#'
#' @param FASTA.NAME path to the fasta file
#'
#' @return SeqInfo object with all chromosome names and lengths from the fasta file
#' @export
#' @author Aleksandr Friman
#'
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#'
#' # create temporary fasta file
#' chr2L_seq <- BSgenome.Dmelanogaster.UCSC.dm6[["chr2L"]]
#' chr2L_seq_set <- DNAStringSet(chr2L_seq)
#' names(chr2L_seq_set) <- "chr2L"
#' temp_fasta <- tempfile(fileext = ".fasta")
#' writeXStringSet(chr2L_seq_set, temp_fasta)
#'
#' myGenome <- PICBloadfasta(FASTA.NAME = temp_fasta)
#' unlink(temp_fasta)
PICBloadfasta <- function(FASTA.NAME = NULL) {
    if (is.null(FASTA.NAME)) stop("Please provide FASTA.NAME !")
    FAdata <- seqinr::read.fasta(FASTA.NAME)
    FAnames <- names(FAdata)
    FAlengths <- lengths(FAdata)
    return(Seqinfo::Seqinfo(seqnames = FAnames, seqlengths = FAlengths))
}

test_that("example data (bam and corresponding bai file) is present", {
    expect_true(file.exists(system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB")))
    expect_true(file.exists(system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam.bai", package = "PICB")))
})

test_that("PICBload loads BAM file with default parameters correctly", {
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    )
    expect_type(result, "list")
    expect_named(result, c("unique", "multi.primary", "multi.secondary"))
    expect_true(all(sapply(result, function(x) inherits(x, "GRanges"))))
    expect_equal(length(result$unique), 31165)
    expect_equal(length(result$multi.primary), 47741)
    expect_equal(length(result$multi.secondary), 369219)
    expect_equal(ncol(mcols(result$unique)), 3)
    expect_equal(ncol(mcols(result$multi.primary)), 3)
    expect_equal(ncol(mcols(result$multi.secondary)), 3)
    expect_equal(min(start(result$unique)), 20063379)
    expect_equal(min(start(result$multi.primary)), 20006039)
    expect_equal(min(start(result$multi.secondary)), 20005922)
    expect_equal(max(end(result$unique)), 20984760)
    expect_equal(max(end(result$multi.primary)), 20980831)
    expect_equal(max(end(result$multi.secondary)), 20981429)
})

test_that("PICBload throws error when BAMFILE is NULL", {
    expect_error(
        PICBload(
        BAMFILE = NULL,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6"
        ),
        "Please provide full path to a .bam file!"
    )
})

test_that("PICBload throws error when REFERENCE.GENOME is NULL", {
    expect_error(
        PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = NULL
        ),
        "Please provide REFERENCE.GENOME"
    )
})

test_that("PICBload throws error when REFERENCE.GENOME is NULL", {
    expect_error(
        PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = NULL
        ),
        "Please provide REFERENCE.GENOME"
    )
})

test_that("PICBload handles SIMPLE.CIGAR = FALSE correctly", {
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        SIMPLE.CIGAR = FALSE
    )
    expect_type(result, "list")
    expect_named(result, c("unique", "multi.primary", "multi.secondary"))
    expect_true(all(sapply(result, function(x) inherits(x, "GRanges"))))
    expect_equal(length(result$unique), 31207)
    expect_equal(length(result$multi.primary), 47844)
    expect_equal(length(result$multi.secondary), 369964)
    expect_equal(ncol(mcols(result$unique)), 3)
    expect_equal(ncol(mcols(result$multi.primary)), 3)
    expect_equal(ncol(mcols(result$multi.secondary)), 3)
})

test_that("PICBload loads secondary alignments when IS.SECONDARY.ALIGNMENT = TRUE", {
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        IS.SECONDARY.ALIGNMENT = TRUE
    )
    expect_type(result, "list")
    expect_named(result, c("multi.secondary"))
    expect_null(result$unique)
    expect_null(result$multi.primary)
    expect_true(inherits(result$multi.secondary, "GRanges"))
    expect_equal(length(result$multi.secondary), 369219)
    expect_equal(ncol(mcols(result$multi.secondary)), 3)
})

test_that("PICBload loads primary alignments when IS.SECONDARY.ALIGNMENT = FALSE", {
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        IS.SECONDARY.ALIGNMENT = FALSE
    )
    expect_type(result, "list")
    expect_named(result, c("unique", "multi.primary"))
    expect_null(result$multi.secondary)
    expect_true(inherits(result$unique, "GRanges"))
    expect_true(inherits(result$multi.primary, "GRanges"))
    expect_equal(length(result$unique), 31165)
    expect_equal(length(result$multi.primary), 47741)
    expect_equal(ncol(mcols(result$unique)), 3)
    expect_equal(ncol(mcols(result$multi.primary)), 3)
})

test_that("PICBload loads BAM file with default parameters the same with using the same reference genome but different methods", {
    expectedResult <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6"
    )
    # Using chromosome names and lengths
    seqnames <- c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY")
    seqlengths <- c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352)
    myGenome <- Seqinfo(seqnames = seqnames, seqlengths = seqlengths)

    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = myGenome
    )
    expect_equal(result, expectedResult)

    # Using genome name
    myGenome <- Seqinfo::Seqinfo(genome = "dm6")
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = myGenome
    )
    expect_equal(result, expectedResult)

    # Using a FASTA file
    chr2L_seq <- BSgenome.Dmelanogaster.UCSC.dm6[["chr2L"]]
    chr2L_seq_set <- DNAStringSet(chr2L_seq)
    names(chr2L_seq_set) <- "chr2L"
    temp_fasta <- tempfile(fileext = ".fasta")
    writeXStringSet(chr2L_seq_set, temp_fasta)
    myGenome <- PICBloadfasta(FASTA.NAME = temp_fasta)
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = myGenome
    )
    expect_equal(result, expectedResult)
    unlink(temp_fasta)
})

test_that("PICBload filters alignments based on custom READ.SIZE.RANGE", {
    custom_size_range <- c(24, 27) # not recommended size range! Just here for testing
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        USE.SIZE.FILTER = TRUE,
        READ.SIZE.RANGE = custom_size_range
    )
    expect_type(result, "list")
    expect_named(result, c("unique", "multi.primary", "multi.secondary"))
    expect_true(all(sapply(result, function(x) inherits(x, "GRanges"))))
    expect_equal(length(result$unique), 28742)
    expect_equal(length(result$multi.primary), 44773)
    expect_equal(length(result$multi.secondary), 347254)
    expect_equal(ncol(mcols(result$unique)), 3)
    expect_equal(ncol(mcols(result$multi.primary)), 3)
    expect_equal(ncol(mcols(result$multi.secondary)), 3)
    # Check that all alignments have width within the specified range
    all_widths <- c(width(result$unique), width(result$multi.primary), width(result$multi.secondary))
    expect_true(all(all_widths >= custom_size_range[1] & all_widths <= custom_size_range[2]))
})

test_that("PICBload retrieves original sequences when GET.ORIGINAL.SEQUENCE = TRUE", {
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        GET.ORIGINAL.SEQUENCE = TRUE
    )
    expect_type(result, "list")
    expect_equal(length(result$unique), 31165)
    expect_equal(length(result$multi.primary), 47741)
    expect_equal(length(result$multi.secondary), 369219)
    expect_equal(ncol(mcols(result$unique)), 4)
    expect_equal(ncol(mcols(result$multi.primary)), 4)
    expect_equal(ncol(mcols(result$multi.secondary)), 4)
    # Check that 'seq' column is present in metadata
    expect_true("seq" %in% colnames(mcols(result$unique)))
    expect_true("seq" %in% colnames(mcols(result$multi.primary)))
    expect_true("seq" %in% colnames(mcols(result$multi.secondary)))
})

test_that("PICBload changes seqlevels style when SEQ.LEVELS.STYLE is specified", {
    custom_style <- "NCBI"
    result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        SEQ.LEVELS.STYLE = custom_style
    )
    expect_type(result, "list")
    # Check that seqlevels style is updated
    expect_equal(GenomeInfoDb::seqlevelsStyle(result$unique), custom_style)
    expect_equal(GenomeInfoDb::seqlevelsStyle(result$multi.primary), custom_style)
    expect_equal(GenomeInfoDb::seqlevelsStyle(result$multi.secondary), custom_style)
})

test_that("PICBload works correctly with a combination of custom parameters", {
     result <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        SIMPLE.CIGAR = FALSE,
        IS.SECONDARY.ALIGNMENT = FALSE,
        STANDARD.CONTIGS.ONLY = TRUE,
        PERFECT.MATCH.ONLY = TRUE,
        FILTER.BY.FLAG = FALSE,
        SELECT.FLAG = c(0, 16),
        USE.SIZE.FILTER = TRUE,
        READ.SIZE.RANGE = c(20, 30),
        TAGS = c("NH", "NM"),
        WHAT = c("flag", "qwidth"),
        SEQ.LEVELS.STYLE = "NCBI",
        GET.ORIGINAL.SEQUENCE = TRUE
    )
    expect_type(result, "list")
    expect_named(result, c("unique", "multi.primary"))
    # Check GRanges content
    expect_true(all(mcols(result$unique)$NH == 1))
    expect_true(all(mcols(result$multi.primary)$NH > 1))
    # Check 'seq' column
    expect_true("seq" %in% colnames(mcols(result$unique)))
    expect_true("seq" %in% colnames(mcols(result$multi.primary)))
    # Check read sizes
    all_widths_unique <- width(result$unique)
    all_widths_multi <- width(result$multi.primary)
    expect_true(all(all_widths_unique >= 20 & all_widths_unique <= 30))
    expect_true(all(all_widths_multi >= 20 & all_widths_multi <= 30))
    # Check that metadata columns are extended (seq and qwidth)
    expect_equal(ncol(mcols(result$unique)), 5)
    expect_equal(ncol(mcols(result$multi.primary)), 5)
    # Check seqlevels style
    expect_equal(GenomeInfoDb::seqlevelsStyle(result$unique), "NCBI")
})

test_that("PICBload VERBOSE parameter works correctly", {
    # Capture messages when VERBOSE is TRUE
    verbose_messages <- capture_messages({
    PICBload(
        BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        VERBOSE = TRUE
    )
    })

    # Capture messages when VERBOSE is FALSE
    silent_messages <- capture_messages({
    PICBload(
        BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        VERBOSE = FALSE
    )
    })

    # Test expectations
    expect_true(length(verbose_messages) > 0, "VERBOSE=TRUE should produce messages")
    expect_true(any(grepl("Processing", verbose_messages)), "Should contain 'Processing' message when VERBOSE=TRUE")
    expect_identical(length(silent_messages), 0L, "VERBOSE=FALSE should not produce any messages")
})

# tests/testthat/test-PICBloadfasta.R
test_that("PICBloadfasta loads a valid FASTA file correctly", {
    temp_fasta <- tempfile(fileext = ".fasta")
    writeLines(c(">chr1", "ATCG", ">chr2", "GGT"), temp_fasta)

    # Expected output (mocked for illustration)
    expected_output <- Seqinfo::Seqinfo(
        seqnames = c("chr1", "chr2"),
        seqlengths = c(4, 3),
        isCircular = c(NA, NA),
        genome = c(NA, NA)
    )

    expect_equal(PICBloadfasta(temp_fasta), expected_output)
    unlink(temp_fasta)
})

test_that("PICBloadfasta handles invalid input of FASTA file", {
    expect_error(PICBloadfasta(NULL), "Please provide FASTA.NAME !")

    temp_fasta <- tempfile(fileext = ".fasta")
    writeLines("", temp_fasta)
    expect_error(PICBloadfasta(temp_fasta), "no line starting with a > character found")
    writeLines("ATCG", temp_fasta)
    expect_error(PICBloadfasta(temp_fasta), "no line starting with a > character found")
    unlink(temp_fasta)
})

test_that("PICBbuild throws error when IN.ALIGNMENTS is NULL", {
    expect_error(
        PICBbuild(
            IN.ALIGNMENTS = NULL,
            REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
            VERBOSITY = 0
        ),
        "Please provide IN.ALIGNMENTS !"
    )
})

test_that("PICBbuild throws error when REFERENCE.GENOME is NULL", {
    expect_error(
        PICBbuild(
            IN.ALIGNMENTS = test_alignments,
            REFERENCE.GENOME = NULL,
            VERBOSITY = 0
        ), "Please provide REFERENCE.GENOME !"
    ) 
})


test_that("PICBbuild throws error when IN.ALIGNMENTS is missing 'unique' column", {
    test_alignments_unique_missing <- test_alignments
    test_alignments_unique_missing$unique <- NULL
    expect_error(
        PICBbuild(
            IN.ALIGNMENTS = test_alignments_unique_missing,
            REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6"
        ),
        regexp = "^IN.ALIGNMENTS must contain 'unique' column! "
    )
})

test_that("PICBbuild throws error when IN.ALIGNMENTS is missing 'multi.primary' column", {
    test_alignments_multi_primary_missing <- test_alignments
    test_alignments_multi_primary_missing$multi.primary <- NULL
    expect_error(
        PICBbuild(
            IN.ALIGNMENTS = test_alignments_multi_primary_missing,
            REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
            VERBOSITY = 0
        ),
        regexp = "^IN.ALIGNMENTS must contain 'multi.primary' column! "
    )
})

test_that("PICBbuild warns when IN.ALIGNMENTS is missing 'multi.secondary' column", {
    test_alignments_multi_secondary_missing <- test_alignments
    test_alignments_multi_secondary_missing$multi.secondary <- NULL
    expect_warning(
        PICBbuild(
            IN.ALIGNMENTS = test_alignments_multi_secondary_missing,
            REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
            VERBOSITY = 0
        ),
        regexp = "^IN.ALIGNMENTS does not contain secondary multimappers"
    )
})

test_that("PICBbuild builds piRNA seeds, cores, and clusters with default parameters correctly", {
    expect_type(test_ranges, "list")
    expect_named(test_ranges, c("seeds", "cores", "clusters"))
    expect_true(all(sapply(test_ranges, function(x) inherits(x, "GRanges"))))

    expect_equal(length(test_ranges$seeds), 42)
    expect_equal(length(test_ranges$cores), 38)
    expect_equal(length(test_ranges$clusters), 27)
    # Check metadata columns
    expectedOutputCols <- c("width_in_nt", "uniq_reads_FPM", "multimapping_reads_primary_alignments_FPM", "all_reads_primary_alignments_FPM", "uniq_reads_FPKM", "multimapping_reads_primary_alignments_FPKM", "all_reads_primary_alignments_FPKM", "fraction_of_width_covered_by_unique_alignments")
    expect_equal(colnames(mcols(test_ranges$seeds)), expectedOutputCols)
    expect_equal(colnames(mcols(test_ranges$cores)), expectedOutputCols)
    expect_equal(colnames(mcols(test_ranges$clusters)), c(expectedOutputCols, "type"))
})


test_that("PICBbuild example subset builds piRNA seeds, cores, and clusters correctly with defined LIBRARY.SIZE", {
    result <- PICBbuild(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        LIBRARY.SIZE = 12799826
    )
    expect_type(result, "list")
    expect_named(result, c("seeds", "cores", "clusters"))
    expect_true(all(sapply(result, function(x) inherits(x, "GRanges"))))

    expect_equal(length(result$seeds), 25)
    expect_equal(length(result$cores), 18)
    expect_equal(length(result$clusters), 4)
    # Check metadata columns
    expectedOutputCols <- c("width_in_nt", "uniq_reads_FPM", "multimapping_reads_primary_alignments_FPM", "all_reads_primary_alignments_FPM", "uniq_reads_FPKM", "multimapping_reads_primary_alignments_FPKM", "all_reads_primary_alignments_FPKM", "fraction_of_width_covered_by_unique_alignments")
    expect_equal(colnames(mcols(result$seeds)), expectedOutputCols)
    expect_equal(colnames(mcols(result$cores)), expectedOutputCols)
    expect_equal(colnames(mcols(result$clusters)), c(expectedOutputCols, "type"))
})

#optional additional testing
#test_that("PICBbuild handles custom sliding window sizes correctly", {
#    custom_uniq_window_width <- 500
#    custom_uniq_window_step <- 50
#    custom_primary_mult_window_width <- 400
#    custom_primary_mult_window_step <- 40
#    custom_secondary_mult_window_width <- 1200
#    custom_secondary_mult_window_step <- 120
#    result <- PICBbuild(
#        IN.ALIGNMENTS = test_alignments,
#        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#        UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH = custom_uniq_window_width,
#        UNIQUEMAPPERS.SLIDING.WINDOW.STEP = custom_uniq_window_step,
#        PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH = custom_primary_mult_window_width,
#        PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP = custom_primary_mult_window_step,
#        SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH = custom_secondary_mult_window_width,
#        SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP = custom_secondary_mult_window_step
#    )
#    expect_type(result, "list")
#    expect_named(result, c("seeds", "cores", "clusters"))
#    expect_true(all(sapply(result, function(x) inherits(x, "GRanges"))))
#
#    expect_equal(length(result$seeds), 43)
#    expect_equal(length(result$cores), 36)
#    expect_equal(length(result$clusters), 28)
#    # Check metadata columns
#    expectedOutputCols <- c("width_in_nt", "uniq_reads_FPM", "multimapping_reads_primary_alignments_FPM", "all_reads_primary_alignments_FPM", "uniq_reads_FPKM", "multimapping_reads_primary_alignments_FPKM", "all_reads_primary_alignments_FPKM", "fraction_of_width_covered_by_unique_alignments")
#    expect_equal(colnames(mcols(result$seeds)), expectedOutputCols)
#    expect_equal(colnames(mcols(result$cores)), expectedOutputCols)
#    expect_equal(colnames(mcols(result$clusters)), c(expectedOutputCols, "type"))
#})


#test_that("PICBbuild filters based on custom thresholds correctly", {
#    custom_min_overlap <- 10
#    custom_threshold_seeds_gap <- 100
#    custom_threshold_cores_gap <- 200
#    custom_threshold_clusters_gap <- 300
#    result <- PICBbuild(
#        IN.ALIGNMENTS = test_alignments,
#        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#        MIN.OVERLAP = custom_min_overlap,
#        THRESHOLD.SEEDS.GAP = custom_threshold_seeds_gap,
#        THRESHOLD.CORES.GAP = custom_threshold_cores_gap,
#        THRESHOLD.CLUSTERS.GAP = custom_threshold_clusters_gap
#    )
#    expect_type(result, "list")
#    expect_named(result, c("seeds", "cores", "clusters"))
#    expect_true(all(sapply(result, function(x) inherits(x, "GRanges"))))
#
#
#    expect_equal(length(result$seeds), 41)
#    expect_equal(length(result$cores), 31)
#    expect_equal(length(result$clusters), 20)
#    # check metacolumns
#    expectedOutputCols <- c("width_in_nt", "uniq_reads_FPM", "multimapping_reads_primary_alignments_FPM", "all_reads_primary_alignments_FPM", "uniq_reads_FPKM", "multimapping_reads_primary_alignments_FPKM", "all_reads_primary_alignments_FPKM", "fraction_of_width_covered_by_unique_alignments")
#    expect_equal(colnames(mcols(result$seeds)), expectedOutputCols)
#    expect_equal(colnames(mcols(result$cores)), expectedOutputCols)
#    expect_equal(colnames(mcols(result$clusters)), c(expectedOutputCols, "type"))
#})


#test_that("PICBbuild provides non-normalized statistics when PROVIDE.NON_NORMALIZED = TRUE", {
#    result <- PICBbuild(
#        IN.ALIGNMENTS = test_alignments,
#        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#        PROVIDE.NON.NORMALIZED = TRUE
#    )
#    expect_type(result, "list")
#    expect_named(result, c("seeds", "cores", "clusters"))
#
#   # Check metadata columns
#    expectedOutputCols <- c("width_in_nt", "uniq_reads_FPM", "multimapping_reads_primary_alignments_FPM", "all_reads_primary_alignments_FPM", "uniq_reads_FPKM", "multimapping_reads_primary_alignments_FPKM", "all_reads_primary_alignments_FPKM", "fraction_of_width_covered_by_unique_alignments")
#    addNonNormOutputCols <- c("uniq_reads", "multimapping_reads_primary_alignments", "all_reads_primary_alignments", "width_covered_by_unique_alignments", "uniq_sequences")
#    expect_equal(length(setdiff(colnames(mcols(result$seeds)), c(expectedOutputCols, addNonNormOutputCols))), 0)
#    expect_equal(length(setdiff(colnames(mcols(result$cores)), c(expectedOutputCols, addNonNormOutputCols))), 0)
#    expect_equal(length(setdiff(colnames(mcols(result$seeds)), c(expectedOutputCols, addNonNormOutputCols, "type"))), 0)
#})

test_that("PICBbuild loads BAM file with default parameters the same with using the same reference genome but different methods", {
    # Using BSgenome as myGenome shown above. 

    # Using chromosome names and lengths
    seqnames <- c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY")
    seqlengths <- c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352)
    myGenome <- Seqinfo(seqnames = seqnames, seqlengths = seqlengths)

    result <- PICBbuild(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = myGenome
    )
    expect_equal(result, test_ranges)

    # Using genome name
    myGenome <- Seqinfo::Seqinfo(genome = "dm6")
    result <- PICBbuild(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = myGenome,
        VERBOSITY = 0
    )
    expect_equal(result, test_ranges)

    # Using a FASTA file
    chr2L_seq <- BSgenome.Dmelanogaster.UCSC.dm6[["chr2L"]]
    chr2L_seq_set <- DNAStringSet(chr2L_seq)
    names(chr2L_seq_set) <- "chr2L"
    temp_fasta <- tempfile(fileext = ".fasta")
    writeXStringSet(chr2L_seq_set, temp_fasta)
    myGenome <- PICBloadfasta(FASTA.NAME = temp_fasta)
    result <- PICBbuild(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = myGenome,
        VERBOSITY = 0
    )
    # seqinfo is expected to be different (all standard chromosomes in above examples vs just chr2L in subset genome)
    seqInfoExp <- seqinfo(test_ranges$seeds)
    seqinfo(result$seeds) <- seqInfoExp
    seqinfo(result$cores) <- seqInfoExp
    seqinfo(result$clusters) <- seqInfoExp

    expect_equal(result, test_ranges)
    unlink(temp_fasta)
})


test_that("PICBbuild changes seqlevels style when SEQ.LEVELS.STYLE is specified", {
    custom_style <- "NCBI"
    result <- PICBbuild(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        SEQ.LEVELS.STYLE = custom_style,
        VERBOSITY = 0
    )
    expect_type(result, "list")
    expect_named(result, c("seeds", "cores", "clusters"))
    # Check if seqlevels style is updated
    expect_equal(GenomeInfoDb::seqlevelsStyle(result$seeds)[1], custom_style)
    expect_equal(GenomeInfoDb::seqlevelsStyle(result$cores)[1], custom_style)
    expect_equal(GenomeInfoDb::seqlevelsStyle(result$clusters)[1], custom_style)
})

test_that("PICBbuild computes 1U and 10A fractions when COMPUTE.1U.10A.FRACTIONS = TRUE", {
    # for error when column seq not provided in PICBload() output (test_alignments)
    expect_error(
        PICBbuild(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        COMPUTE.1U.10A.FRACTIONS = TRUE
        ),
        regexp = "^Alignments 'unique' does not contain sequence"
    )
    # proper calculations with column seq in test_alignments
    ouputPICBloadWseq <- PICBload(
        BAMFILE = test_bam,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        GET.ORIGINAL.SEQUENCE = TRUE,
        VERBOSE = FALSE
    )

    result <- PICBbuild(
        IN.ALIGNMENTS = ouputPICBloadWseq,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        COMPUTE.1U.10A.FRACTIONS = TRUE
    )

    expect_type(result, "list")
    expect_named(result, c("seeds", "cores", "clusters"))

    # Check that the fraction columns are present in metadata
    expectedOutputCols <- c("width_in_nt", "uniq_reads_FPM", "multimapping_reads_primary_alignments_FPM", "all_reads_primary_alignments_FPM", "uniq_reads_FPKM", "multimapping_reads_primary_alignments_FPKM", "all_reads_primary_alignments_FPKM", "fraction_of_width_covered_by_unique_alignments")
    fraction1U10ACols <- c("oneU.frac.unique", "tenA.frac.unique", "oneU.frac.multi.primary", "tenA.frac.multi.primary", "oneU.frac.multi.secondary", "tenA.frac.multi.secondary")
    expect_equal(length(setdiff(colnames(mcols(result$seeds)), c(expectedOutputCols, fraction1U10ACols))), 0)
    expect_equal(length(setdiff(colnames(mcols(result$cores)), c(expectedOutputCols, fraction1U10ACols))), 0)
    expect_equal(length(setdiff(colnames(mcols(result$seeds)), c(expectedOutputCols, fraction1U10ACols, "type"))), 0)
})

test_that("PICBbuild handles empty alignments gracefully", {
    # Create empty GRanges objects
    empty_gr <- GRanges()
    alignments_empty <- list(
        unique = empty_gr,
        multi.primary = empty_gr,
        multi.secondary = empty_gr
    )
    expect_error(
        PICBbuild(
        IN.ALIGNMENTS = alignments_empty,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6"
        ), "The IN.ALIGNMENTS must contain NH information !"
    )
})

test_that("PICBbuild VERBOSITY parameter works correctly", {
  
  # Test VERBOSITY = 0 (silent)
  silent_messages <- capture_messages({
    PICBbuild(
      IN.ALIGNMENTS = test_alignments,
      REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
      VERBOSITY = 0
    )
  })
  
  # Test VERBOSITY = 1 (basic messages)
  basic_messages <- capture_messages({
    PICBbuild(
      IN.ALIGNMENTS = test_alignments,
      REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
      VERBOSITY = 1
    )
  })
  
  # Test VERBOSITY = 2 (detailed messages)
  detailed_messages <- capture_messages({
    PICBbuild(
      IN.ALIGNMENTS = test_alignments,
      REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
      VERBOSITY = 3
    )
  })
  
  # Test expectations
  expect_identical(length(silent_messages), 0L, "VERBOSITY=0 should not produce any messages")
  
  expect_true(length(basic_messages) == 4, "VERBOSITY=1 should produce messages")
  expect_true(
    any(grepl("Starting", basic_messages)) && 
    any(grepl("Done", basic_messages)),
    "VERBOSITY=1 should contain basic progress messages"
  )
  
  expect_true(length(detailed_messages) > length(basic_messages), 
              "VERBOSITY=2 should produce more messages than VERBOSITY=1")
  expect_true(
    any(grepl("Sliding window analysis", detailed_messages)) && 
    any(grepl("Annotating", detailed_messages)),
    "VERBOSITY=2 should contain detailed progress messages"
  )
})

#' Add sense/antisense ratio of unique piRNAs per piRNA cluster
#'
#' @param IN.ALIGNMENTS list of alignments from PICBload
#' @param IN.RANGES single GRanges object (seeds, cores or clusters from PICBbuild)
#' @param VERBOSE enables progress output. True by default.
#'
#' @return GRanges object with an additional annotation column
#' @author Parthena Konstantinidou
#' @author Zuzana Loubalova
#' @author Franziska Ahrend
#' @export
#'
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#' myAlignmentsFromPICBload <- PICBload(
#'     BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     VERBOSE = FALSE
#' )
#'
#' outputOfPICBbuild <- PICBbuild(
#'     IN.ALIGNMENTS = myAlignmentsFromPICBload,
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     VERBOSITY = 0
#' )
#'
#' outputOfPICBbuild$clusters <- PICBstrandanalysis(
#'     IN.ALIGNMENTS = myAlignmentsFromPICBload,
#'     IN.RANGES = outputOfPICBbuild$clusters
#' )
PICBstrandanalysis <- function(IN.ALIGNMENTS, IN.RANGES, VERBOSE = TRUE) {
    if (is.null(IN.ALIGNMENTS) || length(IN.ALIGNMENTS$unique) == 0) stop("IN.ALIGNMENTS is either NULL or contains no rows. Please provide valid IN.ALIGNMENTS from PICBload!")
    if (is.null(IN.RANGES) || length(IN.RANGES) == 0) stop("IN.RANGES is either NULL or contains no rows. Use the GRanges object seeds, cores, or clusters from PICBbuild to create valid IN.RANGES!")
    if (is.list(IN.RANGES)) stop("IN.RANGES must be a single GRanges object: seeds, cores, or clusters from PICBbuild!")


    if (VERBOSE == TRUE) message("PICB v", utils::packageVersion("PICB"), " Starting with strand analysis ... ")

    if (VERBOSE == TRUE) message("\tUsing unique alignments only")
    # use unique alignments only
    IN.ALIGNMENTS <- IN.ALIGNMENTS$unique[IN.ALIGNMENTS$unique$NH == 1]
    # get strandedness
    IN.ALIGNMENTS.PLUS <- IN.ALIGNMENTS[GenomicRanges::strand(IN.ALIGNMENTS) == "+"]
    IN.ALIGNMENTS.MINUS <- IN.ALIGNMENTS[GenomicRanges::strand(IN.ALIGNMENTS) == "-"]

    if (VERBOSE == TRUE) message("\tFinding overlaps of clusters with piRNAs from plus strand")
    # DT_plus <- as.data.frame(GenomicRanges::findOverlaps(IN.RANGES, IN.ALIGNMENTS.PLUS, ignore.strand = TRUE))
    overlaps_plus <- GenomicRanges::findOverlaps(IN.RANGES, IN.ALIGNMENTS.PLUS, ignore.strand = TRUE)
    DT_plus <- data.frame(
        queryHits = S4Vectors::queryHits(overlaps_plus),
        subjectHits = S4Vectors::subjectHits(overlaps_plus)
    )
    if (nrow(DT_plus) == 0) {
        warning("No overlaps detected between IN.ALIGNMENTS and clusters on the plus strand. Please verify that IN.ALIGNMENTS and IN.RANGES are valid and correctly correspond to each other. Continuing ...")
        DT_plus.2 <- data.frame(queryHits = integer(0), subjectHits = integer(0))
    } else {
        DT_plus.2 <- stats::aggregate(subjectHits ~ queryHits, data = DT_plus, FUN = length)
    }
    # Add numbers of plus-strand piRNAs to clusters
    IN.RANGES$plus_piRNAs <- 0
    IN.RANGES$plus_piRNAs[DT_plus.2$queryHits] <- DT_plus.2$subjectHits

    if (VERBOSE == TRUE) message("\tFinding overlaps of clusters with piRNAs from minus strand")
    # DT_minus <- as.data.frame(GenomicRanges::findOverlaps(IN.RANGES, IN.ALIGNMENTS.MINUS, ignore.strand = TRUE))
    overlaps_minus <- GenomicRanges::findOverlaps(IN.RANGES, IN.ALIGNMENTS.MINUS, ignore.strand = TRUE)
    DT_minus <- data.frame(
        queryHits = S4Vectors::queryHits(overlaps_minus),
        subjectHits = S4Vectors::subjectHits(overlaps_minus)
    )
    if (nrow(DT_minus) == 0) {
        warning("No overlaps detected between IN.ALIGNMENTS and clusters on the minus strand. Please verify that IN.ALIGNMENTS and IN.RANGES are valid and correctly correspond to each other. Continuing ...")
        DT_minus.2 <- data.frame(queryHits = integer(0), subjectHits = integer(0))
    } else {
        DT_minus.2 <- stats::aggregate(subjectHits ~ queryHits, data = DT_minus, FUN = length)
    }
    # Add numbers of plus-strand piRNAs to clusters
    IN.RANGES$minus_piRNAs <- 0
    IN.RANGES$minus_piRNAs[DT_minus.2$queryHits] <- DT_minus.2$subjectHits

    # to avoid deviding by 0, add 1 piRNA to each cluster (plus and minus strands)
    IN.RANGES$plus_piRNAs <- IN.RANGES$plus_piRNAs + 1
    IN.RANGES$minus_piRNAs <- IN.RANGES$minus_piRNAs + 1

    if (VERBOSE == TRUE) message("\tAdding ratio of sense/antisense piRNAs per cluster")
    IN.RANGES$s_as_ratio <- 1
    IN.RANGES[GenomicRanges::strand(IN.RANGES) == "+"]$s_as_ratio <- (IN.RANGES[GenomicRanges::strand(IN.RANGES) == "+"]$plus_piRNAs / IN.RANGES[GenomicRanges::strand(IN.RANGES) == "+"]$minus_piRNAs)
    IN.RANGES[GenomicRanges::strand(IN.RANGES) == "-"]$s_as_ratio <- (IN.RANGES[GenomicRanges::strand(IN.RANGES) == "-"]$minus_piRNAs / IN.RANGES[GenomicRanges::strand(IN.RANGES) == "-"]$plus_piRNAs)

    # remove plus_piRNAs and minus_piRNAs
    IN.RANGES$plus_piRNAs <- NULL
    IN.RANGES$minus_piRNAs <- NULL

    if (VERBOSE == TRUE) message("Done!")

    return(IN.RANGES)
}

# expect_equal(myFunction(testData), expectedOutput)

# invalid IN.ALIGNMENTS
test_that("PICBstrandanalysis stops with invalid IN.ALIGNMENTS", {
    alignments <- list(unique = GenomicRanges::GRanges())
    ranges <- GenomicRanges::GRanges()

    expect_error(PICBstrandanalysis(NULL, ranges), "IN.ALIGNMENTS is either NULL or contains no rows. Please provide valid IN.ALIGNMENTS from PICBload!")
})

# invalid IN.RANGES
test_that("PICBstrandanalysis stops with invalid IN.RANGES", {
    alignments <- list(unique = GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(1000, 2000), width = 50),
        strand = c("+", "-"),
        NH = c(1, 1)
    ))
    ranges <- GenomicRanges::GRanges()
    expect_error(PICBstrandanalysis(alignments, NULL), "IN.RANGES is either NULL or contains no rows. Use the GRanges object seeds, cores, or clusters from PICBbuild to create valid IN.RANGES!")
    expect_error(PICBstrandanalysis(alignments, ranges), "IN.RANGES is either NULL or contains no rows. Use the GRanges object seeds, cores, or clusters from PICBbuild to create valid IN.RANGES!")
    expect_error(PICBstrandanalysis(alignments, list(ranges)), "IN.RANGES must be a single GRanges object: seeds, cores, or clusters from PICBbuild!")
})

# valid IN.ALIGNMENTS and IN.RANGES expect_equal
test_that("PICBstrandanalysis returns expected output when alignments on both strands", {
    alignments <- list(unique = GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1", "chr1", "chr1", "chr1", "chr1", "chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(15, 27, 30, 19, 43, 52, 75, 80), width = c(26, 27, 27, 27, 27, 26, 27, 27)),
        strand = c("+", "+", "+", "+", "-", "-", "-", "-"),
        NH = c(1, 1, 1, 1, 1, 1, 1, 1)
    ), multi.primary = GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1", "chr1", "chr1", "chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(20, 30, 40, 60, 68, 81), width = c(26, 27, 27, 27, 27, 27)),
        strand = c("+", "-", "-", "-", "+", "-"),
        NH = c(3, 2, 4, 100, 34, 90)
    ))

    ranges <- GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(10, 40), width = c(40, 60)),
        strand = c("+", "-")
    )
    # piC on plus strand: sense/antisense = 4/1;
    # piC on minus strand: sense/antisense = 4/4;

    # Note: With piRNA input, we anticipate a large number of alignments.
    # To prevent division by zero when calculating the sense-to-antisense ratio (s_as_ratio), we add a pseudocount of 1 to both the sense and antisense counts.
    # Consequently, the expected value in the s_as_ratio column is adjusted to 2.5 and 1.0 instead of 4.0 and 1.0 in this mock dataset.
    # This adjustment ensures numerical stability during computation and does not affect the accuracy of results when working with real piRNA data, where counts are sufficiently large.

    # piC on plus strand: sense+1/antisense+1 = 5/2 = 2.5
    # piC on minus strand: sense+1/antisense+1 = 5/5 = 1.0

    expect_equal(
        PICBstrandanalysis(alignments, ranges),
        GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(10, 40), width = c(40, 60)),
        strand = c("+", "-"),
        s_as_ratio = c(2.5, 1.0)
        )
    )
})


# valid IN.ALIGNMENTS and IN.RANGES but just one strand in IN.RANGES, expect_equal
test_that("PICBstrandanalysis returns expected output when no clusters on plus strand", {
    alignments <- list(unique = GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1", "chr1", "chr1", "chr1", "chr1", "chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(15, 27, 30, 19, 43, 52, 75, 80), width = c(26, 27, 27, 27, 27, 26, 27, 27)),
        strand = c("+", "+", "+", "+", "-", "-", "-", "-"),
        NH = c(1, 1, 1, 1, 1, 1, 1, 1)
    ), multi.primary = GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1", "chr1", "chr1", "chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(20, 30, 40, 60, 68, 81), width = c(26, 27, 27, 27, 27, 27)),
        strand = c("+", "-", "-", "-", "+", "-"),
        NH = c(3, 2, 4, 100, 34, 90)
    ))

    ranges <- GenomicRanges::GRanges(
        seqnames = c("chr1"),
        ranges = IRanges::IRanges(start = c(10), width = c(40)),
        strand = c("+")
    )

    expect_equal(
        PICBstrandanalysis(alignments, ranges),
        GenomicRanges::GRanges(
        seqnames = c("chr1"),
        ranges = IRanges::IRanges(start = c(10), width = c(40)),
        strand = c("+"),
        s_as_ratio = c(2.5)
        )
    )

    ranges <- GenomicRanges::GRanges(
        seqnames = c("chr1"),
        ranges = IRanges::IRanges(start = c(40), width = c(60)),
        strand = c("-")
    )

    expect_equal(
        PICBstrandanalysis(alignments, ranges),
        GenomicRanges::GRanges(
        seqnames = c("chr1"),
        ranges = IRanges::IRanges(start = c(40), width = c(60)),
        strand = c("-"),
        s_as_ratio = c(1.0)
        )
    )
})

# valid IN.RANGES but IN.ALIGNMENTS not corresponding to IN.RANGES, expect_equal
test_that("PICBstrandanalysis returns expected output when IN.ALIGNMENTS not corresponding to IN.RANGES", {
    alignments <- list(unique = GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1", "chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(15, 27, 30, 19), width = c(26, 27, 27, 27)),
        strand = c("+", "+", "+", "+"),
        NH = c(1, 1, 1, 1)
    ), multi.primary = GenomicRanges::GRanges(
        seqnames = c("chr1", "chr1"),
        ranges = IRanges::IRanges(start = c(20, 68), width = c(26, 27)),
        strand = c("+", "+"),
        NH = c(3, 34)
    ))

    ranges <- GenomicRanges::GRanges(
        seqnames = c("chr1"),
        ranges = IRanges::IRanges(start = c(40), width = c(60)),
        strand = c("-")
    )

    result <- expect_warning(
        PICBstrandanalysis(alignments, ranges),
        "No overlaps detected between IN.ALIGNMENTS and clusters on the minus strand. Please verify that IN.ALIGNMENTS and IN.RANGES are valid and correctly correspond to each other. Continuing ..."
    )
    
    #invertStrand 
    alignmentsAS <- alignments
    alignmentsAS$unique <- GenomicRanges::invertStrand(alignments$unique)
    alignmentsAS$multi.primary <- GenomicRanges::invertStrand(alignments$multi.primary)
    rangesAS <- GenomicRanges::invertStrand(ranges)

    resultAS <- expect_warning(
        PICBstrandanalysis(alignmentsAS, rangesAS),
        "No overlaps detected between IN.ALIGNMENTS and clusters on the plus strand. Please verify that IN.ALIGNMENTS and IN.RANGES are valid and correctly correspond to each other. Continuing ..."
    )

    # piC on minus strand: sense+1/antisense+1 = 1/5 = 0.2
    expect_equal(
        result,
        GenomicRanges::GRanges(
        seqnames = c("chr1"),
        ranges = IRanges::IRanges(start = c(40), width = c(60)),
        strand = c("-"),
        s_as_ratio = c(0.2)
        )
    )

    expect_equal(
        resultAS,
        GenomicRanges::GRanges(
        seqnames = c("chr1"),
        ranges = IRanges::IRanges(start = c(40), width = c(60)),
        strand = c("+"),
        s_as_ratio = c(0.2)
        )
    )
})


# test with example data

test_that("PICBload loads BAM file with default parameters correctly", {
    expect_type(test_alignments, "list")
    expect_named(test_alignments, c("unique", "multi.primary", "multi.secondary"))
    expect_true(all(sapply(test_alignments, function(x) inherits(x, "GRanges"))))

    myClusters <- PICBbuild(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        VERBOSITY = 0,
        LIBRARY.SIZE = 12799826
    )$clusters
    expect_equal(
        length(PICBstrandanalysis(IN.ALIGNMENTS = test_alignments, IN.RANGES = myClusters)$s_as_ratio), 4
    )
    expect_equal(
        round(PICBstrandanalysis(IN.ALIGNMENTS = test_alignments, IN.RANGES = myClusters)$s_as_ratio, digits = 3), round(c(3.91609, 1.68109, 0.25495, 0.59485), digits = 3)
    )
})

#' Import cluster object from an Excel file
#'
#' @param EXCEL.FILE.NAME file name to import from
#'
#' @author Aleksandr Friman
#'
#' @return list of annotated GRanges objects named "seeds" for seeds,
#' "cores" for cores,
#' "clusters" for clusters
#' @export
#' @examples
#' importedClusters <- PICBimportfromexcel(
#'     EXCEL.FILE.NAME = system.file("extdata", "myClusters_demonstration.xlsx", package = "PICB")
#' )
PICBimportfromexcel <- function(EXCEL.FILE.NAME = NULL) {
    availbsheets <- openxlsx::getSheetNames(EXCEL.FILE.NAME)
    output <- list()
    sheetsToCheck <- intersect(availbsheets, c(uniqueonly, uniqueandprimary, allalignments))
    for (t in sheetsToCheck) {
        output[[t]] <- GenomicRanges::GRanges(openxlsx::read.xlsx(EXCEL.FILE.NAME, sheet = t))
    }
    return(output)
}

test_that("PICBimportfromexcel correctly imports Excel files", {
  # Setup: Create test Excel files using export
  temp_file <- tempfile(fileext = ".xlsx")
  single_temp_file <- tempfile(fileext = ".xlsx")
  
  # Export test data
  PICBexporttoexcel(IN.RANGES = test_ranges, EXCEL.FILE.NAME = temp_file)
  PICBexporttoexcel(IN.RANGES = list(clusters = test_ranges$clusters), 
                    EXCEL.FILE.NAME = single_temp_file)
  
  # Test importing multi-sheet Excel
  imported_ranges <- PICBimportfromexcel(EXCEL.FILE.NAME = temp_file)
  
  # Check structure
  expect_type(imported_ranges, "list")
  expect_equal(length(imported_ranges), 3)
  expect_true(all(c("seeds", "cores", "clusters") %in% names(imported_ranges)))
  
  # Check that each element is a GRanges object
  for (name in names(imported_ranges)) {
    expect_s4_class(imported_ranges[[name]], "GRanges")
    # Compare with original
    expect_equal(length(imported_ranges[[name]]), length(test_ranges[[name]]))
    expect_equal(as.character(seqnames(imported_ranges[[name]])), as.character(seqnames(test_ranges[[name]])))
    expect_equal(ranges(imported_ranges[[name]]), ranges(test_ranges[[name]]))
  }
  
  # Test importing single-sheet Excel
  imported_single <- PICBimportfromexcel(EXCEL.FILE.NAME = single_temp_file)
  
  # Check structure of single import
  expect_type(imported_single, "list")
  expect_equal(length(imported_single), 1)
  expect_true("clusters" %in% names(imported_single))
  
  # Check content of single import
  expect_s4_class(imported_single$clusters, "GRanges")
  expect_equal(length(imported_single$clusters), length(test_ranges$clusters))
  expect_equal(as.character(seqnames(imported_single$clusters)), as.character(seqnames(test_ranges$clusters)))
  expect_equal(ranges(imported_single$clusters), ranges(test_ranges$clusters))
  
  # Clean up
  unlink(temp_file)
  unlink(single_temp_file)
})

#' Get SeqInfo object from standard non-circular chromosome names from your genome
#'
#' @param REFERENCE.GENOME name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6", or directly a SeqInfo object.
#' @param SEQ.LEVELS.STYLE naming of chromosomes style. "UCSC" by default.
#'
#' @return SeqInfo object with standard non-circular chromosome names
#' @export
#'
#' @author Aleksandr Friman
#' @author Franziska Ahrend
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#' mySI <- PICBgetchromosomes("BSgenome.Dmelanogaster.UCSC.dm6", "UCSC")
PICBgetchromosomes <- function(REFERENCE.GENOME, SEQ.LEVELS.STYLE = "UCSC") {    
    # Get SeqInfo object first, regardless of SEQ.LEVELS.STYLE
    if (methods::is(REFERENCE.GENOME, "character")) {  #BSgenome object
        SI <- Seqinfo::seqinfo(x = eval(parse(text = REFERENCE.GENOME)))
    } else if (methods::is(REFERENCE.GENOME, "Seqinfo")) {
        SI <- REFERENCE.GENOME
    } else {
        stop("REFERENCE.GENOME must be a character BSgenome object (e.g. 'BSgenome.Dmelanogaster.UCSC.dm6') or a Seqinfo object")
    }
    
    if (is.na(SEQ.LEVELS.STYLE)) {
        # If SEQ.LEVELS.STYLE is NA, return full SeqInfo without filtering
        return(list(SeqInfo = SI, chrom_mismatch = TRUE))
    }

    # Filter for standard chromosomes
    SI <- GenomeInfoDb::keepStandardChromosomes(SI) 
    
    # Check if any standard chromosomes were found
    if (length(SI) == 0) {
        return(list(SeqInfo = REFERENCE.GENOME, chrom_mismatch = TRUE))
    }
    
    # Update style and drop circular chromosomes
    GenomeInfoDb::seqlevelsStyle(SI) <- SEQ.LEVELS.STYLE
    SI <- GenomeInfoDb::dropSeqlevels(x = SI, value = names(which(Seqinfo::isCircular(SI))))

    return(list(SeqInfo = SI, chrom_mismatch = FALSE))
}

#' Export cluster object into an Excel file
#'
#' @param IN.RANGES clustering object to export
#' @param EXCEL.FILE.NAME file name to save
#'
#' @author Aleksandr Friman
#' @author Franziska Ahrend
#' @return no values returned
#' @export
#'
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#' myAlignmentsFromPICBload <- PICBload(
#'     BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     VERBOSE = FALSE
#' )
#'
#' outputOfPICBbuild <- PICBbuild(
#'     IN.ALIGNMENTS = myAlignmentsFromPICBload,
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     LIBRARY.SIZE = 12799826, #usually not necessary
#'     VERBOSITY = 0
#' )
#' PICBexporttoexcel(
#'     IN.RANGES = outputOfPICBbuild,
#'     EXCEL.FILE.NAME = "inst/extdata/myClusters_demonstration.xlsx"
#' )
PICBexporttoexcel <- function(IN.RANGES = NULL, EXCEL.FILE.NAME = NULL) {
    # Check if IN.RANGES is a single GRanges object and convert to list
    if (inherits(IN.RANGES, "GRanges")) {
        message("IN.RANGES is provided as a single GRanges object. Converting to a list with the name 'clusters'. If you intend to save seeds or cores, please input the data as a list (e.g., list(seeds = IN.RANGES)).")
        IN.RANGES <- list(clusters = IN.RANGES)
    }
    valid_names <- c(uniqueonly, uniqueandprimary, allalignments)
    invalid_names <- setdiff(names(IN.RANGES), valid_names) 
    if (length(invalid_names) > 0) {
        stop("Invalid names detected in IN.RANGES: ", paste(invalid_names, collapse = ", "), ". Expected names are 'seeds', 'cores', and/or 'clusters'.")
    }

    allowedSheets <- names(IN.RANGES)[names(IN.RANGES) %in% c(uniqueonly, uniqueandprimary, allalignments)]
    openxlsx::write.xlsx(IN.RANGES[allowedSheets], file = EXCEL.FILE.NAME)
    message("Saved to Excel.")
}

test_that("PICBexporttoexcel creates correct Excel files", {
  # Test full export with multiple sheets
  temp_file <- tempfile(fileext = ".xlsx")
  PICBexporttoexcel(IN.RANGES = test_ranges, EXCEL.FILE.NAME = temp_file)
  
  # Read the Excel file back
  xlsx_data <- openxlsx::loadWorkbook(temp_file)
  
  # Check number of sheets
  expect_equal(length(names(xlsx_data)), 3)
  
  # Check content of sheets
  for (sheet in names(xlsx_data)) { 
    sheet_data <- openxlsx::read.xlsx(temp_file, sheet = sheet)
    expect_gt(nrow(sheet_data), 1) # More than just header row
    expect_gt(ncol(sheet_data), 0) # Has columns
  }
  
  # Test single GRanges export
  single_temp_file <- tempfile(fileext = ".xlsx")
  PICBexporttoexcel(IN.RANGES = test_ranges$clusters, 
                    EXCEL.FILE.NAME = single_temp_file)
  
  # Read single sheet Excel file back
  single_xlsx_data <- openxlsx::loadWorkbook(single_temp_file)
  
  # Check single sheet export
  expect_equal(length(names(single_xlsx_data)), 1)
  expect_equal(names(single_xlsx_data), "clusters")
  
  sheet_data <- openxlsx::read.xlsx(single_temp_file, sheet = 1)
  expect_gt(nrow(sheet_data), 1)
  expect_gt(ncol(sheet_data), 0)
  
  # Clean up
  unlink(temp_file)
  unlink(single_temp_file)
})

#' Runs PICBbuild multiple times with provided parameters and returns optimization data frame.
#'
#' @param IN.ALIGNMENTS list of alignments from PICBload
#' @param REFERENCE.GENOME name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"
#' @param LIBRARY.SIZE number of reads in the library. By default computed as number of unique mapping alignments + number of primary multimapping alignments.
#' @param VERBOSITY verbosity level 0/1/2/3. 2 by default.
#' @param PROVIDE.INFO.SEEDS.AND.CORES FALSE by default.
#' @param SEQ.LEVELS.STYLE naming of chromosomes style. "UCSC" by default.
#' @param ... rest of the parameters used by PICBbuild and provided as iteratable vectors
#'
#' @author Aleksandr Friman
#' @return Optimization values dataframe
#' @export
#'
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#' myAlignmentsFromPICBload <- PICBload(
#'     BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     VERBOSE = FALSE
#' )
#'
#' PICBoptimize(
#'     IN.ALIGNMENTS = myAlignmentsFromPICBload,
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2, 3, 4, 5)
#' )
PICBoptimize <- function(
    IN.ALIGNMENTS,
    REFERENCE.GENOME,
    LIBRARY.SIZE = length(IN.ALIGNMENTS$unique) + length(IN.ALIGNMENTS$multi.primary),
    VERBOSITY = 2,
    PROVIDE.INFO.SEEDS.AND.CORES = FALSE,
    SEQ.LEVELS.STYLE = "UCSC",
    ...) {

    if (VERBOSITY > 0) message("PICB v", utils::packageVersion("PICB"), " Starting ... ")

    # checking the inputs
    if (VERBOSITY > 0) {
        message("Checking the inputs")
    }
    if (is.null(IN.ALIGNMENTS)) stop("Please provide IN.ALIGNMENTS !")
    if (is.null(REFERENCE.GENOME)) stop("Please provide REFERENCE.GENOME !")

    numberOfallReadsExplained <- function(gr, alignments) {
        foundnames <- c()
        if (typeof(alignments) == "list") {
            for (clmn in c("unique", "multi.primary", "multi.secondary")) {
                usedAlignments <- IRanges::subsetByOverlaps(alignments[[clmn]], gr)
                foundnames <- c(foundnames, names(usedAlignments))
            }
        } else {
            usedAlignments <- IRanges::subsetByOverlaps(alignments, gr)
            foundnames <- c(foundnames, names(usedAlignments))
        }
        numbr <- length(unique(foundnames))
        return(numbr)
    }

    inputArgs <- as.list(match.call(expand.dots = TRUE))
    PICBbuildArgs <- formals(PICBbuild)
    ArgsToIterate <- NULL
    for (argName in names(PICBbuildArgs)) {
        # checking if the argument need to be iterated over
        if (is.numeric(PICBbuildArgs[[argName]]) ||
            (is.language(PICBbuildArgs[[argName]]) && PICBbuildArgs[[argName]] != "")) {
            # iteration is possible
            if (argName %in% names(inputArgs)) {
                # need to iterate
                if (length(inputArgs[[argName]]) > 1) {
                    if (VERBOSITY > 0) {
                        message("Adding ", argName, " to iteration targets")
                    }
                    if (is.null(ArgsToIterate)) {
                        ArgsToIterate <- argName
                    } else {
                        ArgsToIterate <- c(ArgsToIterate, argName)
                    }
                }
            }
        }
    }

    
    if (is.null(ArgsToIterate)) {
        stop("Provide arguments to iterave over. See example.")
    }

    totalReads <- LIBRARY.SIZE
    if (length(unique(c(
        names(IN.ALIGNMENTS$unique),
        names(IN.ALIGNMENTS$multi.primary)
    ))) != totalReads) {
        warning("The total number of primary alignments is not equal to the total number of read names. This discrepancy may occur when using a subset of data, such as in the PICB demonstration, where you specified the LIBRARY.SIZE but the actual number of reads is reduced. For full piRNA datasets, this indicates a potential issue with read assignment and your numbers of explained reads may be wrong.")
    }
    # building the parameter combinations
    if (VERBOSITY > 0) {
        message("Building the parameter combinations")
    }
    listOfArgLists <- list()
    for (argName in ArgsToIterate) {
        firstArg <- length(listOfArgLists) == 0
        argVals <- eval(inputArgs[[argName]])
        if (firstArg) { # first arg
            for (curVal in argVals) {
                tmpList <- list()
                tmpList[[argName]] <- curVal
                listOfArgLists[[length(listOfArgLists) + 1]] <- tmpList
            }
        } else {
            # add the argName to each combination
            newlistOfArgLists <- list()
            for (combList in listOfArgLists) {
                for (curVal in argVals) {
                    combList[[argName]] <- curVal
                    newlistOfArgLists[[length(newlistOfArgLists) + 1]] <- combList
                }
            }
            listOfArgLists <- newlistOfArgLists
        }
    }
    IN.ALIGNMENTS.ALL <- c(
        IN.ALIGNMENTS$unique, IN.ALIGNMENTS$multi.primary,
        IN.ALIGNMENTS$multi.secondary
    )
    # preparing output dataframe
    outDF <- data.frame(combinationNumber = seq_along(listOfArgLists))
    for (argName in names(listOfArgLists[[1]])) {
        # adding columns we iterate over
        outDF[[argName]] <- NA
    }
    if (PROVIDE.INFO.SEEDS.AND.CORES) {
        for (locustype in c(uniqueonly, uniqueandprimary, allalignments)) {
            outDF[[paste0("number.of.", locustype)]] <- NA
        }
        for (locustype in c(uniqueonly, uniqueandprimary, allalignments)) {
            outDF[[paste0("total.width.", locustype)]] <- NA
        }
        for (locustype in c(uniqueonly, uniqueandprimary, allalignments)) {
            outDF[[paste0("reads.explained.by.", locustype)]] <- NA
        }
    } else {
        outDF[[paste0("number.of.clusters")]] <- NA
        outDF[[paste0("total.width.clusters")]] <- NA
        outDF[[paste0("reads.explained.by.clusters")]] <- NA
    }
    # running PICBbuild for every possible combation
    i <- 1
    for (combList in listOfArgLists) {
        if (VERBOSITY > 0) {
            message("Iteration ", i, " out of ", length(listOfArgLists))
        }
        for (argName in names(combList)) {
            # adding values we iterate over
            outDF[[argName]][i] <- combList[[argName]]
        }
        listToRun <- combList
        listToRun[["IN.ALIGNMENTS"]] <- IN.ALIGNMENTS
        listToRun[["REFERENCE.GENOME"]] <- REFERENCE.GENOME
        listToRun[["VERBOSITY"]] <- VERBOSITY
        if (VERBOSITY > 1) {
            message("Running PICBbuild")
        }
        tmrLoci <- do.call(what = PICBbuild, args = listToRun)
        if (VERBOSITY > 1) {
            message("PICBbuild done")
        }
        if (PROVIDE.INFO.SEEDS.AND.CORES) {
        for (locustype in c(uniqueonly, uniqueandprimary, allalignments)) {
            outDF[[paste0("number.of.", locustype)]][i] <- length(tmrLoci[[locustype]])
            outDF[[paste0("total.width.", locustype)]][i] <- as.numeric(sum(GenomicRanges::width(tmrLoci[[locustype]])))
            outDF[[paste0("reads.explained.by.", locustype)]][i] <- as.numeric(numberOfallReadsExplained(tmrLoci[[locustype]], IN.ALIGNMENTS.ALL))
        }
        } else {
            outDF[["number.of.clusters"]][i] <- length(tmrLoci[[allalignments]])
            outDF[["total.width.clusters"]][i] <- as.numeric(sum(GenomicRanges::width(tmrLoci[[allalignments]])))
            outDF[["reads.explained.by.clusters"]][i] <- as.numeric(numberOfallReadsExplained(tmrLoci[[allalignments]], IN.ALIGNMENTS.ALL))
        }
        i <- i + 1
    }
    outDF$combinationNumber <- NULL
    # adding normaized values
    if (VERBOSITY > 0) {
        message("Values normalization")
    }
    if (PROVIDE.INFO.SEEDS.AND.CORES) {
        for (locustype in c(uniqueonly, uniqueandprimary, allalignments)) {
            outDF[[paste0("fraction.of.library.explained.by.", locustype)]] <- outDF[[paste0("reads.explained.by.", locustype)]] / totalReads
        }
        for (locustype in c(uniqueonly, uniqueandprimary, allalignments)) {
            outDF[[paste0("mean.RPKM.", locustype)]] <- outDF[[paste0("reads.explained.by.", locustype)]] * 1e9 / (outDF[[paste0("total.width.", locustype)]] * totalReads)
            outDF[[paste0("mean.RPKM.", locustype)]][is.na(outDF[[paste0("mean.RPKM.", locustype)]])] <- 0
        }
        SI <- PICBgetchromosomes(REFERENCE.GENOME, SEQ.LEVELS.STYLE)$SeqInfo
        GenomeSize <- sum(Seqinfo::seqlengths(SI))
        for (locustype in c(uniqueonly, uniqueandprimary, allalignments)) {
        outDF[[paste0("fraction.of.genome.space.", locustype)]] <- outDF[[paste0("total.width.", locustype)]] / (2 * GenomeSize)
        }
    } else {
        outDF[["fraction.of.library.explained.by.clusters"]] <- outDF[["reads.explained.by.clusters"]] / totalReads
        outDF[["mean.RPKM.clusters"]] <- outDF[["reads.explained.by.clusters"]] * 1e9 / (outDF[["total.width.clusters"]] * totalReads)
        outDF[["mean.RPKM.clusters"]][is.na(outDF[["mean.RPKM.clusters"]])] <- 0
        SI <- PICBgetchromosomes(REFERENCE.GENOME, SEQ.LEVELS.STYLE)$SeqInfo
        GenomeSize <- sum(Seqinfo::seqlengths(SI))
        outDF[["fraction.of.genome.space.clusters"]] <- outDF[["total.width.clusters"]] / (2 * GenomeSize)
    }
    return(outDF)
}

test_that("PICBoptimize throws error when IN.ALIGNMENTS is NULL", {
    expect_error(
        PICBoptimize(
            IN.ALIGNMENTS = NULL,
            REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
            VERBOSITY = 0
        ),
        "Please provide IN.ALIGNMENTS !"
    )
})

test_that("PICBoptimize throws error when REFERENCE.GENOME is NULL", {
    expect_error(
        PICBoptimize(
            IN.ALIGNMENTS = test_alignments,
            REFERENCE.GENOME = NULL,
            VERBOSITY = 0
        ), "Please provide REFERENCE.GENOME !"
    ) 
})

test_that("PICBoptimize throws error when nothing to optimize for", {
    expect_error(
        PICBoptimize(
            IN.ALIGNMENTS = test_alignments,
            REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
            VERBOSITY = 0
        ), "Provide arguments to iterave over. See example."
    ) 
})

test_that("PICBoptimize handles basic parameter iteration correctly", {
    # Test single parameter iteration
    expect_equal(nrow(test_optimization), 2)
    expect_true(all(c("MIN.UNIQUE.ALIGNMENTS.PER.WINDOW") %in% colnames(test_optimization)))
    
    # Test multiple parameter iteration
    result2 <- PICBoptimize(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        VERBOSITY = 0,
        MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2),
        THRESHOLD.CLUSTERS.GAP = c(0, 50)
    )
    expect_equal(nrow(result2), 4)  # 2x2 combinations
    expect_true(all(c("MIN.UNIQUE.ALIGNMENTS.PER.WINDOW", "THRESHOLD.CLUSTERS.GAP") %in% colnames(result2)))
})

test_that("PICBoptimize handles library size correctly", {
    custom_library_size <- 1000000
    # Library size warning should be triggered
    expect_warning(
        PICBoptimize(
            IN.ALIGNMENTS = test_alignments,
            REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
            LIBRARY.SIZE = custom_library_size,
            THRESHOLD.SEEDS.GAP = c(0, 50)
        ),
        regexp = "^The total number of primary alignments is not equal to the total number of read names"
    )
})

test_that("PICBoptimize output contains expected columns", {
    
    expected_cols <- c(
        "MIN.UNIQUE.ALIGNMENTS.PER.WINDOW",
        "number.of.clusters",
        "total.width.clusters",
        "reads.explained.by.clusters",
        "fraction.of.library.explained.by.clusters",
        "mean.RPKM.clusters",
        "fraction.of.genome.space.clusters"
    )
    expect_true(all(expected_cols %in% colnames(test_optimization)))
})

test_that("PICBoptimize handles PROVIDE.INFO.SEEDS.AND.CORES correctly", {
    result <- PICBoptimize(
        IN.ALIGNMENTS = test_alignments,
        REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
        VERBOSITY = 0,
        PROVIDE.INFO.SEEDS.AND.CORES = TRUE,
        THRESHOLD.CORES.GAP = c(0, 50)
    )
    
    # Base metrics that apply to each type
    base_metrics <- c(
        "number.of",
        "total.width",
        "reads.explained.by",
        "fraction.of.library.explained.by",
        "mean.RPKM",
        "fraction.of.genome.space"
    )
    
    # Types of regions
    region_types <- c("seeds", "cores", "clusters")
    
    # Generate all expected column combinations
    expected_cols <- c(
        "THRESHOLD.CORES.GAP",  # Include any fixed columns first
        outer(base_metrics, region_types, function(x, y) paste0(x, ".", y))
    )
    
    expect_true(all(expected_cols %in% colnames(result)))
})

#' PICB: piRNA Cluster Builder
#'
#' piRNAs (short for PIWI-interacting RNAs) and their PIWI protein partners play a key role in fertility 
#' and maintaining genome integrity by restricting mobile genetic elements (transposons) in germ cells. 
#' piRNAs originate from genomic regions known as piRNA clusters. The piRNA Cluster Builder (PICB) is 
#' a versatile toolkit designed to identify genomic regions with a high density of piRNAs. It constructs 
#' piRNA clusters through a stepwise integration of unique and multimapping piRNAs and offers wide-ranging 
#' parameter settings, supported by an optimization function that allows users to test different parameter 
#' combinations to tailor the analysis to their specific piRNA system. The output includes extensive metadata 
#' columns, enabling researchers to rank clusters and extract cluster characteristics. 
#'
#' @section Main Functions:
#' The package provides several core functions:
#' \itemize{
#'   \item \code{PICBload}: Load and preprocess BAM files containing piRNA alignments
#'   \item \code{PICBbuild}:  Build piRNA seeds/cores/clusters from alignments
#'   \item \code{PICBoptimize}: Optimize parameters for cluster building
#'   \item \code{PICBstrandanalysis}: Add sense/antisense ratio of unique piRNAs per piRNAcluster
#'   \item \code{PICBannotate}: Annotate GRanges according to a piRNA library
#'   \item \code{PICBloadfasta}: Get SeqInfo object from a fasta file
#'   \item \code{PICBexporttoexcel}: Export cluster object into an Excel file
#'   \item \code{PICBimporttoexcel}: Import cluster object from an Excel file
#' }
#'
#' @section Workflow:
#' A typical PICB workflow consists of:
#' \enumerate{
#'   \item Loading alignments with \code{PICBload}
#'   \item Building clusters with \code{PICBbuild}
#'   \item Optional parameter optimization with \code{PICBoptimize}
#'   \item Optional strand analysis with \code{PICBstrandanalysis}
#'   \item Exporting results with \code{PICBexporttoexcel}
#' }
#' 
#' @examples
#' # 0. Load PICB
#' library(PICB)
#' 
#' # 1. Load Required Genome from e.g. Seqinfo (check all options in Vignette or ReadMe)
#' myGenome <- Seqinfo::Seqinfo(
#'     seqnames = c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY"),
#'     seqlengths = c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352)
#' )
#'
#' # 2. Load Example Data and Process Alignments
#' bam_file <- system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", 
#'     package = "PICB")
#'
#' myAlignments <- PICBload(
#'     BAMFILE = bam_file,
#'     REFERENCE.GENOME = myGenome,
#'     VERBOSE = FALSE
#' )
#'
#' # 3. Build piRNA Clusters
#' myClusters <- PICBbuild(
#'     IN.ALIGNMENTS = myAlignments,
#'     REFERENCE.GENOME = myGenome,
#'     LIBRARY.SIZE = 12799826,  # Usually calculated automatically
#'     VERBOSITY = 0
#' )$clusters
#'
#' # 4. Optimize Parameters (Optional)
#' parameterExploration <- PICBoptimize(
#'     IN.ALIGNMENTS = myAlignments,
#'     REFERENCE.GENOME = myGenome,
#'     MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2, 3, 4, 5),
#'     LIBRARY.SIZE = 12799826,  # Usually calculated automatically
#'     VERBOSITY = 1
#' )
#'
#' # 5. Perform Strand Analysis
#' myClustersWithStrand <- PICBstrandanalysis(
#'     IN.ALIGNMENTS = myAlignments,
#'     IN.RANGES = myClusters
#' )
#'
#' # 6. Export Clusters
#' PICBexporttoexcel(
#'     IN.RANGES = myClustersWithStrand,
#'     EXCEL.FILE.NAME = "myClusters_demonstration.xlsx"
#' )
#' # 7. Import Ranges
#' importedClusters <- PICBimportfromexcel(
#'     EXCEL.FILE.NAME = system.file("extdata", "myClusters_demonstration.xlsx", package = "PICB")
#' )
#' 
#' @keywords internal
"_PACKAGE"

#' Load and preprocess BAM files containing piRNA alignments
#'
#' @param BAMFILE name of the bam file to load. Should be sorted and indexed.
#' @param REFERENCE.GENOME name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"
#' @param SIMPLE.CIGAR simpleCigar parameter of Rsamtools::ScanBamParam
#' @param IS.SECONDARY.ALIGNMENT defines loading of primary/secondary alignments. Default value NA loads both primary and secondary.
#' @param STANDARD.CONTIGS.ONLY use only standard chromosomes
#' @param PERFECT.MATCH.ONLY load only alignments without mismatches
#' @param FILTER.BY.FLAG enables filtering by flag. TRUE by default.
#' @param SELECT.FLAG vector of flags to use. Default value c(0,16, 272, 256).
#' @param USE.SIZE.FILTER enables filter by alignment size. True by default.
#' @param READ.SIZE.RANGE allowed alignment sizes. c(18,50) by default.
#' @param TAGS tags to import from bam file. c("NH","NM") by default.
#' @param WHAT "what" parameter of Rsamtools::ScanBamParam. c("flag") by default.
#' @param SEQ.LEVELS.STYLE naming of chromosomes style. "UCSC" by default.
#' @param GET.ORIGINAL.SEQUENCE adds "seq" to WHAT. False by default.
#' @param VERBOSE enables progress output. True by default.
#'
#' @author Pavol Genzor
#' @author Daniel Stoyko
#' @author Aleksandr Friman
#' @author Franziska Ahrend
#' @return list of GRanges objects named "unique" for unique mapping alignments,
#' "multi.primary" for primary multimapping alignments,
#' "multi.secondary" for secondary multimapping alignments
#' @export
#'
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#' PICBload(
#'     BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     VERBOSE = FALSE
#' )
#'
PICBload <- function(
    ## INPUTS
    BAMFILE = NULL,
    REFERENCE.GENOME = NULL,
    ## OPTIONS
    SIMPLE.CIGAR = TRUE,
    IS.SECONDARY.ALIGNMENT = NA,
    STANDARD.CONTIGS.ONLY = TRUE,
    PERFECT.MATCH.ONLY = FALSE,
    FILTER.BY.FLAG = TRUE,
    SELECT.FLAG = c(0, 16, 272, 256),
    USE.SIZE.FILTER = TRUE,
    READ.SIZE.RANGE = c(18, 50),
    TAGS = c("NH", "NM"),
    WHAT = c("flag"),
    ## EXTRA OPTIONS
    SEQ.LEVELS.STYLE = "UCSC",
    GET.ORIGINAL.SEQUENCE = FALSE,
    VERBOSE = TRUE) {
    ## NOTE ON BAM INFO
    ## Flag: 256 = not primary alignment; 272 = reverse strand not primary alignment;
    ## Flag: 0 = forward unpaired unique alignment; 16 = reverse unpaired unique alignment
    ## Tags: NH:i:1 = unique alignment; NM = edit distance to the reference

    outputAlignments <- list()
    chrom_mismatch_flag <- FALSE
    ## check input
    if (is.null(BAMFILE)) stop("Please provide full path to a .bam file!")
    if (is.null(REFERENCE.GENOME)) stop("Please provide REFERENCE.GENOME")
    if (isTRUE(GET.ORIGINAL.SEQUENCE)) {
        WHAT <- c(WHAT, "seq")
    }
    ## for report

    ##
    if (VERBOSE) message("PICB v", utils::packageVersion("PICB"), " Processing ... ")
    ##

    justPrimaryOrSecondary <- function(IS.SECONDARY.ALIGNMENT) {
        ## PARAMETERS FOR LOADING BAM FILE
        if (isTRUE(STANDARD.CONTIGS.ONLY) && !is.na(SEQ.LEVELS.STYLE)) {
            genome_result <- PICBgetchromosomes(REFERENCE.GENOME, SEQ.LEVELS.STYLE)
            SI <- genome_result$SeqInfo
            chrom_mismatch_flag <- genome_result$chrom_mismatch

            REG.CHR <- Seqinfo::seqnames(SI)
            BAM.FILE.HEADER <- Rsamtools::BamFile(BAMFILE)
            BAM.FILE.CHR <- Seqinfo::seqnames(Seqinfo::seqinfo(BAM.FILE.HEADER))
            REG.CHR <- REG.CHR[REG.CHR %in% BAM.FILE.CHR]
            WHICH <- GenomicRanges::GRanges(
                seqnames = REG.CHR,
                ranges = IRanges::IRanges(start = rep(1, length(REG.CHR)), end = Seqinfo::seqlengths(SI)[REG.CHR]),
                strand = rep("*", length(REG.CHR))
            )
            PARAM <- Rsamtools::ScanBamParam(
                flag = Rsamtools::scanBamFlag(
                    isUnmappedQuery = FALSE,
                    isSecondaryAlignment = IS.SECONDARY.ALIGNMENT
                ),
                tag = TAGS, simpleCigar = SIMPLE.CIGAR, what = WHAT, which = WHICH
            )
        } else {
            SI <- PICBgetchromosomes(REFERENCE.GENOME, NA)$SeqInfo
            PARAM <- Rsamtools::ScanBamParam(
                flag = Rsamtools::scanBamFlag(
                    isUnmappedQuery = FALSE,
                    isSecondaryAlignment = IS.SECONDARY.ALIGNMENT
                ),
                tag = TAGS, simpleCigar = SIMPLE.CIGAR, what = WHAT
            )
        }


        if (VERBOSE) {
            message("\nprepared loading parameters")
            message("\tTAGS:\t", toString(TAGS))
            message("\tCIGAR:\t", ifelse(isTRUE(SIMPLE.CIGAR), "simple cigar", "all cigar"))
            message("\tWHAT:\t", toString(WHAT))
            message("loading .bam file into GAlignments")

            if (is.na(IS.SECONDARY.ALIGNMENT)) {
                message("\n******")
                message("SLOW - Loading all reads")
                message(" => to load unique and primary alignments set")
                message(" => IS.SECONDARY.ALIGNMENT=FALSE")
                message("******\n")
            } else if (!IS.SECONDARY.ALIGNMENT) {
                message("\n******")
                message("Loading unique and primary alignments !!!")
                message("******\n")
            } else if (IS.SECONDARY.ALIGNMENT) {
                message("\n******")
                message("Loading secondary alignments only !!!")
                message("******\n")
            }
        }

        GA <- GenomicAlignments::readGAlignments(
            file = BAMFILE,
            use.names = TRUE,
            param = PARAM
        )
        # checking the tags consistency
        for (tagcheck in TAGS) {
            if (any(is.na(GenomicRanges::mcols(GA)[[tagcheck]]))) {
                stop("Tag ", tagcheck, " contains NA values. Check your bam file.")
            }
        }
        ## ***
        GA.IN <- length(GA)
        if (VERBOSE) message("\tIMPORTED: ", GA.IN)

        if (isTRUE(USE.SIZE.FILTER)) {
            if (VERBOSE) {
                message("\nfiltering by read size")
                message(
                "\tRANGE:\t",
                min(READ.SIZE.RANGE), "-", max(READ.SIZE.RANGE)
                )
            }
            GA <- GA[GenomicAlignments::width(GA) %in% seq(min(READ.SIZE.RANGE), max(READ.SIZE.RANGE), by = 1)]

            ## ***
            REMAINDER <- (length(GA) / GA.IN) * 100
            if (VERBOSE) message("\tREMAINDER: ", round(REMAINDER, digits = 2), " %")
        }



        if (isTRUE(FILTER.BY.FLAG)) {
            if (VERBOSE) message("\nfiltering based on flags")
            GARP <- GA[GenomicRanges::mcols(GA)[["flag"]] %in% SELECT.FLAG]
            # mcols(GARP)[["flag"]] <- NULL

            ## ***
            REMAINDER <- (length(GARP) / GA.IN) * 100
            if (VERBOSE) message("\tREMAINDER: ", round(REMAINDER, digits = 2), " %")
        } else {
            GARP <- GA
        }

        if (isTRUE(PERFECT.MATCH.ONLY)) {
            if (VERBOSE) message("\nremoving reads with mismatches")
            GARP <- GARP[GenomicRanges::mcols(GARP)[["NM"]] %in% c(0)]

            ## ***
            REMAINDER <- (length(GARP) / GA.IN) * 100
            if (VERBOSE) message("\tREMAINDER: ", round(REMAINDER, digits = 2), " %")
        }

        if (isTRUE(GET.ORIGINAL.SEQUENCE)) {
            if (VERBOSE) message("\nretrieving original read sequences")
            BAMSEQ <- GenomicRanges::mcols(GARP)[["seq"]]
            ISONMINUS <- as.logical(GenomicAlignments::strand(GARP) == "-")
            BAMSEQ[ISONMINUS] <- Biostrings::reverseComplement(BAMSEQ[ISONMINUS])
            GenomicRanges::mcols(GARP)[["seq"]] <- BAMSEQ
        }

        if (VERBOSE) message("\nconverting to GRanges")
        GARP.GR <- GenomicRanges::granges(GARP, use.names = TRUE, use.mcols = TRUE)
        # if no chromosomes mismatch and SEQ.LEVELS.STYLE is not NA, set the seqlevels style
        if (!chrom_mismatch_flag && !is.na(SEQ.LEVELS.STYLE)) {
            if (VERBOSE) message("\nSeqlevels style set to: ", SEQ.LEVELS.STYLE)
            GenomeInfoDb::seqlevelsStyle(GARP.GR) <- SEQ.LEVELS.STYLE
        }

        if (!all(unique(Seqinfo::seqnames(GARP.GR)) %in% Seqinfo::seqnames(SI))) {
            chrom_mismatch_flag <- TRUE
            warning("Some or all chromosome names in BAM file do not match those in REFERENCE.GENOME.",
                "\nNot filtered by chromosome names. Ensure REFERENCE.GENOME is compatible with the BAMFILE.")
        }

        return(GARP.GR)
    }
    if (VERBOSE) message("\nSorting into uniquemappers vs multimappers and primary vs secondary alignments")

    if (is.na(IS.SECONDARY.ALIGNMENT)) { # justPrimaryOrSecondary
        if (VERBOSE) message("Loading primary only")
        PrimaryAlignments <- justPrimaryOrSecondary(IS.SECONDARY.ALIGNMENT = FALSE)
        outputAlignments[["unique"]] <- PrimaryAlignments[GenomicRanges::mcols(PrimaryAlignments)[["NH"]] == 1]
        outputAlignments[["multi.primary"]] <- PrimaryAlignments[(GenomicRanges::mcols(PrimaryAlignments)[["NH"]] > 1)]
        if (VERBOSE) message("Loading secondary only")
        outputAlignments[["multi.secondary"]] <- justPrimaryOrSecondary(IS.SECONDARY.ALIGNMENT = TRUE)
    } else if (IS.SECONDARY.ALIGNMENT) {
        outputAlignments[["unique"]] <- NULL
        outputAlignments[["multi.primary"]] <- NULL
        if (VERBOSE) message("Loading secondary only")
        outputAlignments[["multi.secondary"]] <- justPrimaryOrSecondary(IS.SECONDARY.ALIGNMENT = TRUE)
    } else if (!IS.SECONDARY.ALIGNMENT) {
        if (VERBOSE) message("Loading primary only")
        PrimaryAlignments <- justPrimaryOrSecondary(IS.SECONDARY.ALIGNMENT = FALSE)
        outputAlignments[["unique"]] <- PrimaryAlignments[GenomicRanges::mcols(PrimaryAlignments)[["NH"]] == 1]
        outputAlignments[["multi.primary"]] <- PrimaryAlignments[(GenomicRanges::mcols(PrimaryAlignments)[["NH"]] > 1)]
        outputAlignments[["multi.secondary"]] <- NULL
    }
    ## RETURN
    if (chrom_mismatch_flag && !is.na(SEQ.LEVELS.STYLE)) {
        warning("No chromosome names in REFERENCE.GENOME match GenomeInfoDb::genomeStyle().",
        "\nKeeping all chromosomes given in REFERENCE.GENOME. No change of SEQ.LEVELS.STYLE.",
        "\nIf desired, filter REFERENCE.GENOME's SeqInfo for chromosomes (e.g. keeping only standard chromosomes).", 
        "\nTo not attempt to filter chromosomes or change SEQ.LEVELS.STYLE, set SEQ.LEVELS.STYLE to NA.")
    }
    if (VERBOSE) {
        message("\nDone!")
        message("")
    }
    return(outputAlignments)
}

# tests/testthat/setup.R
library("BSgenome.Dmelanogaster.UCSC.dm6")

# Common test data setup
test_bam <- system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB")

# Load alignments once
test_alignments <- PICBload(
    BAMFILE = test_bam,
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSE = FALSE
)

# Build ranges (seeds, cores, clusters) once
test_ranges <- PICBbuild(
    IN.ALIGNMENTS = test_alignments,
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSITY = 0
)

# Optimize once
test_optimization <- PICBoptimize(
    IN.ALIGNMENTS = test_alignments,
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSITY = 0,
    MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2)
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----install_dependencies, eval = FALSE---------------------------------------
# install.packages(c(
#     "data.table",
#     "seqinr",
#     "openxlsx",
#     "dplyr"
# ))
# if (!require("BiocManager", quietly = TRUE)) {
#     install.packages("BiocManager")
# }
# BiocManager::install(c(
#     "IRanges",
#     "GenomicRanges",
#     "GenomicAlignments",
#     "Rsamtools",
#     "Biostrings",
#     "Seqinfo",
#     "BSgenome",
#     "rtracklayer"
# ))

## ----install_PICB, eval = FALSE-----------------------------------------------
# BiocManager::install("PICB")

## ----install_PICB_dev, eval = FALSE-------------------------------------------
# devtools::install_github("HaaseLab/PICB")

## ----load_PICB----------------------------------------------------------------
library(PICB)

## ----load_BSgenome1, results="hide", message=FALSE----------------------------
# Example for Drosophila melanogaster (make sure to install the package first), replace with your own genome
library(BSgenome.Dmelanogaster.UCSC.dm6)
myGenome <- "BSgenome.Dmelanogaster.UCSC.dm6"

## ----load_myGenome2-----------------------------------------------------------
myGenome <- Seqinfo::Seqinfo(
    seqnames = c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY"),
    seqlengths = c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352)
)

## ----load_myGenome3-----------------------------------------------------------
myGenome <- Seqinfo::Seqinfo(genome = "dm6")

## ----load_myGenome4, eval = FALSE---------------------------------------------
# myGenome <- PICBloadfasta(FASTA.NAME = "/path/to/your/genome.fa")

## ----load_alignments----------------------------------------------------------
# demonstration bam_file is in the "inst/extdata/" folder of the PICB package
bam_file <- system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB")
myAlignments <- PICBload(
    BAMFILE = bam_file,
    REFERENCE.GENOME = myGenome
)

## ----build_clusters-----------------------------------------------------------
myClusters <- PICBbuild(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    LIBRARY.SIZE = 12799826 #usually not necessary
)$clusters

## ----optimize_parameters------------------------------------------------------
parameterExploration <- PICBoptimize(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2, 3, 4, 5), 
    LIBRARY.SIZE = 12799826, #usually not necessary
    VERBOSITY = 1
)

## ----example_specify_parameter------------------------------------------------
x_column <- "MIN.UNIQUE.ALIGNMENTS.PER.WINDOW" # change parameter to optimize, if applicable

## ----example_plot_optimize_parameters, message=FALSE--------------------------
library(ggplot2)
# Determine a scaling factor for the secondary axis
scaling_factor <- max(parameterExploration$fraction.of.library.explained.by.clusters) / max(parameterExploration$fraction.of.genome.space.clusters)
# plot graph
ggplot(parameterExploration, aes(x = .data[[x_column]])) +
    geom_line(aes(y = fraction.of.library.explained.by.clusters * 100, color = "piRNAs Explained"), linewidth = 1) +
    geom_point(aes(y = fraction.of.library.explained.by.clusters * 100, color = "piRNAs Explained"), size = 3) +
    geom_line(aes(y = fraction.of.genome.space.clusters * scaling_factor * 100, color = "Genome Space"), linewidth = 1) +
    geom_point(aes(y = fraction.of.genome.space.clusters * scaling_factor * 100, color = "Genome Space"), size = 3) +
    scale_y_continuous(name = "piRNAs Explained by Clusters (%, piRNA sample)", sec.axis = sec_axis(~ . / scaling_factor, name = "Total piRNA cluster-length (Genome, %)")) +
    scale_x_reverse(name = paste0("Parameter chosen: ", x_column), breaks = parameterExploration[[x_column]], labels = parameterExploration[[x_column]]) +
    scale_color_manual(name = "Metrics", values = c("piRNAs Explained" = "#00a100", "Genome Space" = "black")) +
    theme_classic() +
    theme(axis.title.y.left = element_text(color = "#00a100"), axis.title.y.right = element_text(color = "black"), legend.position = "top")

## ----strand_specific_analysis-------------------------------------------------
myClustersWithStrandAnalysis <- PICBstrandanalysis(
    IN.ALIGNMENTS = myAlignments,
    IN.RANGES = myClusters
)

## ----output_clusters----------------------------------------------------------
myClustersWithStrandAnalysis

## ----export_results-----------------------------------------------------------
PICBexporttoexcel(
    IN.RANGES = myClustersWithStrandAnalysis,
    EXCEL.FILE.NAME = "myClusters_demonstration.xlsx"
)

## ----import_results-----------------------------------------------------------
myClustersWithStrandAnalysis <- PICBimportfromexcel(EXCEL.FILE.NAME = "myClusters_demonstration.xlsx")

## ----SessionInfo--------------------------------------------------------------
sessionInfo()

#' Build piRNA seeds/cores/clusters from alignments
#'
#' @param IN.ALIGNMENTS list of alignments from PICBload
#' @param REFERENCE.GENOME name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"
#' @param UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH width of sliding window for unique mappers. 350 nt by default
#' @param UNIQUEMAPPERS.SLIDING.WINDOW.STEP step of sliding windows for unique mappers. width/10 by default
#' @param PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH width of sliding window for primary multimapping alignments. 350 nt by default
#' @param PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP step of sliding windows for primary multimapping alignments. width/10 by default
#' @param SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH width of sliding window for secondary multimapping alignments. 1000 nt by default
#' @param SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP step of sliding windows for secondary multimapping alignments. width/10 by default
#' @param LIBRARY.SIZE number of reads in the library. By default computed as number of unique mapping alignments + number of primary multimapping alignments.
#' @param MIN.UNIQUE.ALIGNMENTS.PER.WINDOW absolute number of unique mapping alignments per window to call it. By default computed as 2 FPKM.
#' @param MIN.UNIQUE.SEQUENCES.PER.WINDOW absolute number of unique mapping sequences per window to call it. By default computed as width/50.
#' @param MIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW absolute number of primary multimapping alignments per window to call it. By default computed as 4 FPKM.
#' @param MIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW absolute number of secondary multimapping alignments per window to call it. By default computed as 0.2 FPKM.
#' @param MIN.SEED.LENGTH minimum length of a seed. By default computed as 2x unique mapper window size + 100.
#' @param MIN.COVERED.SEED.LENGTH minimum number of seed nucleotides covered by unique mappers. 0 by default.
#' @param THRESHOLD.SEEDS.GAP minimum gap between seeds to not merge them. 0 by default.
#' @param THRESHOLD.CORES.GAP minimum gap between cores to not merge them. 0 by default.
#' @param THRESHOLD.CLUSTERS.GAP minimum gap between clusters to not merge them. 0 by default.
#' @param SEQ.LEVELS.STYLE naming of chromosomes style. "UCSC" by default.
#' @param MIN.OVERLAP minimum overlap between seeds and cores, as well as between cores and clusters 5 nt by default.
#' @param PROVIDE.NON.NORMALIZED include non-normalized to the library size statistics in the output annotations
#' @param COMPUTE.1U.10A.FRACTIONS for each locus and each alignments type (unique mapping, primary multimapping, secodnary multimapping) compute fraction 1U and 10A containing reads overlapping the locus. Default FALSE.
#' @param VERBOSITY verbosity level 0/1/2/3. 2 by default.
#'
#' @author Pavol Genzor
#' @author Daniel Stoyko
#' @author Aleksandr Friman
#' @author Franziska Ahrend
#' @return list of annotated GRanges objects named "seeds" for seeds,
#' "cores" for cores,
#' "clusters" for clusters
#' @export
#'
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#' myAlignmentsFromPICBload <- PICBload(
#'     BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     VERBOSE = FALSE
#' )
#'
#' outputOfPICBbuild <- PICBbuild(
#'     IN.ALIGNMENTS = myAlignmentsFromPICBload,
#'     REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
#'     VERBOSITY = 0
#' )
PICBbuild <- function(
        ## MAIN
        IN.ALIGNMENTS,
        REFERENCE.GENOME,
        ## OPTIONS
        UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH = 350,
        UNIQUEMAPPERS.SLIDING.WINDOW.STEP = round(UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH / 10, 0),
        PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH = 350,
        PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP = round(PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH / 10, 0),
        SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH = 1000,
        SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP = round(SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH / 10, 0),
        LIBRARY.SIZE = length(IN.ALIGNMENTS$unique) + length(IN.ALIGNMENTS$multi.primary),
        MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = 2 * (UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH / 1e3) * (LIBRARY.SIZE / 1e6),
        MIN.UNIQUE.SEQUENCES.PER.WINDOW = min(MIN.UNIQUE.ALIGNMENTS.PER.WINDOW, round(UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH / 50, 0)),
        MIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW = 4 * (PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH / 1e3) * (LIBRARY.SIZE / 1e6),
        MIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW = 0.2 * (SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH / 1e3) * (LIBRARY.SIZE / 1e6),
        MIN.SEED.LENGTH = 2 * UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH + 100,
        MIN.COVERED.SEED.LENGTH = 0,
        ## CLUSTER FILTERS
        THRESHOLD.SEEDS.GAP = 0, THRESHOLD.CORES.GAP = 0, THRESHOLD.CLUSTERS.GAP = 0,
        ## EXTRA OPTIONS
        SEQ.LEVELS.STYLE = "UCSC",
        MIN.OVERLAP = 5,
        PROVIDE.NON.NORMALIZED = FALSE,
        COMPUTE.1U.10A.FRACTIONS = FALSE,
        VERBOSITY = 2) {
    ## Authors: Pavol Genzor, Alex Friman, Daniel Stoyko
    ## Usage: Build piRNA clusters using piRNA sequencing data

    ## OUTPUT FOR DEBUGGING PURPOSES
    if (VERBOSITY > 2) {
        tmpENV <- as.list(environment())
        tmpENV$IN.ALIGNMENTS <- NULL
        message(tmpENV)
    }
    ## Check input

    if (is.null(IN.ALIGNMENTS)) stop("Please provide IN.ALIGNMENTS !")
    if (is.null(REFERENCE.GENOME)) stop("Please provide REFERENCE.GENOME !")
    # check if IN.ALIGNMENTS contains columns "unique" and "multi.primary"
    typeAlignments <- names(IN.ALIGNMENTS)
    if (!"unique" %in% typeAlignments) stop("IN.ALIGNMENTS must contain 'unique' column! Use PICBload() with IS.SECONDARY.ALIGNMENT = NA (default) to load piRNAs. \n")
    if (!"multi.primary" %in% typeAlignments) stop("IN.ALIGNMENTS must contain 'multi.primary' column! Use PICBload() with IS.SECONDARY.ALIGNMENT = NA (default) to load piRNAs. We recommend including both primary and secondary multimappers; however, you could extract results using only unique alignments by utilizing the 'core' component from the output of PICBbuild.\n")
    if (!"multi.secondary" %in% typeAlignments) warning("IN.ALIGNMENTS does not contain secondary multimappers (column 'multi.secondary')! We recommend including secondary multimappers; please set IS.SECONDARY.ALIGNMENT = NA (default) in PICBload. \n\tContinuing without secondary multimappers...\n")

    for (columnName in typeAlignments) {
        if (!"NH" %in% colnames(GenomicRanges::mcols(IN.ALIGNMENTS[[columnName]]))) stop("The IN.ALIGNMENTS must contain NH information !")
    }

    ##
    if (VERBOSITY > 0) message("PICB v", utils::packageVersion("PICB"), " Starting ... ")
    ##
    chrom_mismatch_flag <- FALSE
    ## KEEP STANDARD CHROMOSOMES
    genome_result <- PICBgetchromosomes(REFERENCE.GENOME, SEQ.LEVELS.STYLE)
    SI <- genome_result$SeqInfo
    chrom_mismatch_flag <- genome_result$chrom_mismatch 
    ##
    ## Clean data
    if (VERBOSITY > 1) {
        if (!is.na(SEQ.LEVELS.STYLE)) {
            message("\n\tKeeping standard linear chromosomes")
        } else {
            message("\n\tKeeping all chromosomes in REFERENCE.GENOME")
        }
    }
    # Change seqlevels style unless chrom_mismatch_flag is TRUE. Only show 
    for (columnName in typeAlignments) {
        if (!chrom_mismatch_flag) {
            GenomeInfoDb::seqlevelsStyle(IN.ALIGNMENTS[[columnName]]) <- SEQ.LEVELS.STYLE
        } else if (!is.na(SEQ.LEVELS.STYLE) && chrom_mismatch_flag) {
            warning(
                "Unable to change SEQ.LEVELS.STYLE: chromosome names in REFERENCE.GENOME do not match any supported genome style.",
                "\nTo see supported styles, use: GenomeInfoDb::genomeStyles() and read https://github.com/Bioconductor/GenomeInfoDb/blob/devel/inst/extdata/dataFiles/README.",
                "\nContinuing using original chromosome names. If you do not want to filter or change chromosome names set SEQ.LEVELS.STYLE to NA."
            )
            SEQ.LEVELS.STYLE <- NA
        }
        if (is.na(SEQ.LEVELS.STYLE)) {
            KEEP.SEQLEVELS <- Seqinfo::seqlevels(SI)
            KEEP.SEQLEVELS <- KEEP.SEQLEVELS[KEEP.SEQLEVELS %in% Seqinfo::seqlevels(IN.ALIGNMENTS[[columnName]])]
            IN.ALIGNMENTS[[columnName]] <- GenomeInfoDb::keepSeqlevels(x = IN.ALIGNMENTS[[columnName]], value = KEEP.SEQLEVELS, pruning.mode = "coarse")
        }
    }

    WGRU <- IN.ALIGNMENTS$unique
    WGRMP <- IN.ALIGNMENTS$multi.primary

    ##
    if (VERBOSITY > 0) message("\nBuilding ... STEP 1... Searching using windows\n")
    ##

    ##
    ## Make genome sliding window for unique mappers

    if (VERBOSITY > 1) message("\tSliding window analysis")
    if (VERBOSITY > 1) message("\t\tUNIQUE MAPPERS\n\t\tWINDOW: ", UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH, "\n\t\tSTEP: ", UNIQUEMAPPERS.SLIDING.WINDOW.STEP)
    AG.gr <- GenomicRanges::GRanges(data.table::data.table("chr" = GenomicRanges::seqnames(SI), "start" = 1, "end" = Seqinfo::seqlengths(SI), "strand" = "*"))
    AG.sw.uniq <- unlist(GenomicRanges::slidingWindows(x = AG.gr, width = UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH, step = UNIQUEMAPPERS.SLIDING.WINDOW.STEP))

    ## Make genome sliding window for multi mappers

    if (VERBOSITY > 1) message("\t\tPRIMARY MULTI MAPPERS\n\t\tWINDOW: ", PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH, "\n\t\tSTEP: ", PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP)
    # AG.gr <- GRanges(data.table("chr"=seqnames(SI),"start"=1,"end"=seqlengths(SI),"strand"="*"))
    AG.sw.primary.mult <- unlist(GenomicRanges::slidingWindows(x = AG.gr, width = PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH, step = PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP))
    if (VERBOSITY > 1 && "multi.secondary" %in% typeAlignments) message("\t\tSECONDARY MULTI MAPPERS\n\t\tWINDOW: ", SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH, "\n\t\tSTEP: ", SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP)

    ##
    ## Counting piRNAs per window: Uniq and all multi
    GenomicRanges::mcols(AG.sw.uniq)[["uniq_piRNA_plus"]] <- GenomicRanges::countOverlaps(query = AG.sw.uniq, subject = WGRU[GenomicRanges::strand(WGRU) == "+"])
    GenomicRanges::mcols(AG.sw.uniq)[["uniq_piRNA_minus"]] <- GenomicRanges::countOverlaps(query = AG.sw.uniq, subject = WGRU[GenomicRanges::strand(WGRU) == "-"])
    GenomicRanges::mcols(AG.sw.uniq)[["uniq_intervals_plus"]] <- GenomicRanges::countOverlaps(query = AG.sw.uniq, subject = unique(WGRU[GenomicRanges::strand(WGRU) == "+"]))
    GenomicRanges::mcols(AG.sw.uniq)[["uniq_intervals_minus"]] <- GenomicRanges::countOverlaps(query = AG.sw.uniq, subject = unique(WGRU[GenomicRanges::strand(WGRU) == "-"]))
    GenomicRanges::mcols(AG.sw.primary.mult)[["primary_mult_piRNA_plus"]] <- GenomicRanges::countOverlaps(query = AG.sw.primary.mult, subject = WGRMP[GenomicRanges::strand(WGRMP) == "+"])
    GenomicRanges::mcols(AG.sw.primary.mult)[["primary_mult_piRNA_minus"]] <- GenomicRanges::countOverlaps(query = AG.sw.primary.mult, subject = WGRMP[GenomicRanges::strand(WGRMP) == "-"])
    if ("multi.secondary" %in% typeAlignments) {
        AG.sw.secondary.mult <- unlist(GenomicRanges::slidingWindows(x = AG.gr, width = SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH, step = SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP))
        GenomicRanges::mcols(AG.sw.secondary.mult)[["secondary_mult_piRNA_plus"]] <- GenomicRanges::countOverlaps(query = AG.sw.secondary.mult, subject = IN.ALIGNMENTS$multi.secondary[GenomicRanges::strand(IN.ALIGNMENTS$multi.secondary) == "+"])
        GenomicRanges::mcols(AG.sw.secondary.mult)[["secondary_mult_piRNA_minus"]] <- GenomicRanges::countOverlaps(query = AG.sw.secondary.mult, subject = IN.ALIGNMENTS$multi.secondary[GenomicRanges::strand(IN.ALIGNMENTS$multi.secondary) == "-"])
    }

    ##
    ## Filter by minimum reads per window and reduce to new range
    if (VERBOSITY > 1) message("\t\tMIN.UNIQUE.ALIGNMENTS.PER.WINDOW >= ", MIN.UNIQUE.ALIGNMENTS.PER.WINDOW)
    sw.UNIQ.PLUS <- AG.sw.uniq[(GenomicRanges::mcols(AG.sw.uniq)[["uniq_piRNA_plus"]] >= MIN.UNIQUE.ALIGNMENTS.PER.WINDOW) & (GenomicRanges::mcols(AG.sw.uniq)[["uniq_intervals_plus"]] >= MIN.UNIQUE.SEQUENCES.PER.WINDOW)]
    sw.UNIQ.MINUS <- AG.sw.uniq[(GenomicRanges::mcols(AG.sw.uniq)[["uniq_piRNA_minus"]] >= MIN.UNIQUE.ALIGNMENTS.PER.WINDOW) & (GenomicRanges::mcols(AG.sw.uniq)[["uniq_intervals_minus"]] >= MIN.UNIQUE.SEQUENCES.PER.WINDOW)]
    AG.sw.uniq <- NULL # clearing memory
    GenomicRanges::strand(sw.UNIQ.PLUS) <- "+"
    GenomicRanges::strand(sw.UNIQ.MINUS) <- "-"
    sw.UNIQ.RED <- GenomicRanges::sort.GenomicRanges(GenomicRanges::reduce(c(sw.UNIQ.PLUS, sw.UNIQ.MINUS)))
    sw.UNIQ.PLUS <- NULL # clearing memory
    ## Removing gaps between SEEDS
    if (THRESHOLD.SEEDS.GAP > 0) {
        GR.GAPS <- GenomicRanges::gaps(sw.UNIQ.RED)
        GR.GAPS <- GenomicRanges::sort.GenomicRanges(GR.GAPS)
        GR.GAPS.WIDER.THAN.MIN <- GR.GAPS[GenomicRanges::width(GR.GAPS) >= THRESHOLD.SEEDS.GAP]
        sw.UNIQ.RED <- GenomicRanges::gaps(GR.GAPS.WIDER.THAN.MIN) # gaps of gaps
        sw.UNIQ.RED <- GenomicRanges::sort.GenomicRanges(sw.UNIQ.RED)
    }
    sw.UNIQ.RED <- GenomicRanges::reduce(sw.UNIQ.RED[GenomicRanges::width(sw.UNIQ.RED) >= MIN.SEED.LENGTH])
    if (VERBOSITY > 1) message("Annotating seeds")
    BM.SEEDS <- PICBannotate(sw.UNIQ.RED, IN.ALIGNMENTS,
        REFERENCE.GENOME = REFERENCE.GENOME, LIBRARY.SIZE = LIBRARY.SIZE,
        SEQ.LEVELS.STYLE = SEQ.LEVELS.STYLE, PROVIDE.NON.NORMALIZED = TRUE
    )
    if (VERBOSITY > 1) message("Removing seed with unique mapping coverage less than ", MIN.COVERED.SEED.LENGTH, " nt")
    BM.SEEDS <- BM.SEEDS[GenomicRanges::mcols(BM.SEEDS)[["width_covered_by_unique_alignments"]] >= MIN.COVERED.SEED.LENGTH]
    outputList <- list()
    outputList[[uniqueonly]] <- BM.SEEDS

    if (VERBOSITY > 1) message("\t\tMIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW >= ", MIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW)
    sw.PRIMARY.MULT.PLUS <- AG.sw.primary.mult[GenomicRanges::mcols(AG.sw.primary.mult)[["primary_mult_piRNA_plus"]] >= MIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW]
    sw.PRIMARY.MULT.MINUS <- AG.sw.primary.mult[GenomicRanges::mcols(AG.sw.primary.mult)[["primary_mult_piRNA_minus"]] >= MIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW]
    GenomicRanges::strand(sw.PRIMARY.MULT.PLUS) <- "+"
    GenomicRanges::strand(sw.PRIMARY.MULT.MINUS) <- "-"
    sw.PRIMARY.MULT.PLUS.RED <- GenomicRanges::reduce(sw.PRIMARY.MULT.PLUS)
    sw.PRIMARY.MULT.MINUS.RED <- GenomicRanges::reduce(sw.PRIMARY.MULT.MINUS)
    sw.PRIMARY.MULT.PLUS.ANCHORED <- IRanges::subsetByOverlaps(x = sw.PRIMARY.MULT.PLUS.RED, ranges = BM.SEEDS, minoverlap = MIN.OVERLAP)
    sw.PRIMARY.MULT.MINUS.ANCHORED <- IRanges::subsetByOverlaps(x = sw.PRIMARY.MULT.MINUS.RED, ranges = BM.SEEDS, minoverlap = MIN.OVERLAP)

    sw.PRIMARY.MULT.RED.ANCHORED <- GenomicRanges::sort.GenomicRanges(GenomicRanges::reduce(c(sw.PRIMARY.MULT.PLUS.ANCHORED, sw.PRIMARY.MULT.MINUS.ANCHORED)))
    sw.PRIMARY.MULT.RED.ANCHORED <- GenomicRanges::sort.GenomicRanges(GenomicRanges::reduce(c(sw.PRIMARY.MULT.RED.ANCHORED, BM.SEEDS)))

    ONE.WINDOW.GR <- c(BM.SEEDS, sw.PRIMARY.MULT.RED.ANCHORED)
    BM.CORES <- GenomicRanges::reduce(ONE.WINDOW.GR)

    ##
    ## Removing gaps between cores
    if (THRESHOLD.CORES.GAP > 0) {
        GR.GAPS <- GenomicRanges::gaps(BM.CORES)
        GR.GAPS <- GenomicRanges::sort.GenomicRanges(GR.GAPS)
        GR.GAPS.WIDER.THAN.MIN <- GR.GAPS[GenomicRanges::width(GR.GAPS) >= THRESHOLD.CORES.GAP]
        BM.CORES <- GenomicRanges::gaps(GR.GAPS.WIDER.THAN.MIN) # gaps of gaps
        BM.CORES <- GenomicRanges::sort.GenomicRanges(BM.CORES)
    }

    ## Regions
    if ("multi.secondary" %in% typeAlignments) {
        if (VERBOSITY > 1) message("\t\tMIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW >= ", MIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW)

        sw.SECONDARY.MULT.PLUS <- AG.sw.secondary.mult[GenomicRanges::mcols(AG.sw.secondary.mult)[["secondary_mult_piRNA_plus"]] >= MIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW]
        sw.SECONDARY.MULT.MINUS <- AG.sw.secondary.mult[GenomicRanges::mcols(AG.sw.secondary.mult)[["secondary_mult_piRNA_minus"]] >= MIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW]
        GenomicRanges::strand(sw.SECONDARY.MULT.PLUS) <- "+"
        GenomicRanges::strand(sw.SECONDARY.MULT.MINUS) <- "-"
        sw.SECONDARY.MULT.PLUS.RED <- GenomicRanges::reduce(sw.SECONDARY.MULT.PLUS)
        sw.SECONDARY.MULT.MINUS.RED <- GenomicRanges::reduce(sw.SECONDARY.MULT.MINUS)
        sw.SECONDARY.MULT.PLUS.ANCHORED <- IRanges::subsetByOverlaps(x = sw.SECONDARY.MULT.PLUS.RED, ranges = BM.CORES, minoverlap = MIN.OVERLAP)
        sw.SECONDARY.MULT.MINUS.ANCHORED <- IRanges::subsetByOverlaps(x = sw.SECONDARY.MULT.MINUS.RED, ranges = BM.CORES, minoverlap = MIN.OVERLAP)

        sw.SECONDARY.MULT.RED.ANCHORED <- GenomicRanges::sort.GenomicRanges(GenomicRanges::reduce(c(sw.SECONDARY.MULT.PLUS.ANCHORED, sw.SECONDARY.MULT.MINUS.ANCHORED)))
        sw.SECONDARY.MULT.RED.ANCHORED <- GenomicRanges::sort.GenomicRanges(GenomicRanges::reduce(c(sw.SECONDARY.MULT.RED.ANCHORED, BM.CORES)))

        CLUSTERS.GR <- c(BM.CORES, sw.SECONDARY.MULT.RED.ANCHORED)
    } else {
        CLUSTERS.GR <- BM.CORES
    }

    CLUSTERS <- GenomicRanges::reduce(CLUSTERS.GR)
    ## Removing gaps between CLUSTERS
    if (THRESHOLD.CLUSTERS.GAP > 0) {
        GR.GAPS <- GenomicRanges::gaps(CLUSTERS)
        GR.GAPS <- GenomicRanges::sort.GenomicRanges(GR.GAPS)
        GR.GAPS.WIDER.THAN.MIN <- GR.GAPS[GenomicRanges::width(GR.GAPS) >= THRESHOLD.CLUSTERS.GAP]
        CLUSTERS <- GenomicRanges::gaps(GR.GAPS.WIDER.THAN.MIN) # gaps of gaps
        CLUSTERS <- GenomicRanges::sort.GenomicRanges(CLUSTERS)
    }
    ##
    if (VERBOSITY > 0) message("\nBuilding ... STEP 2 ... Annotating & filtering\n")
    ##

    ##
    ## Annotate the cores

    if (VERBOSITY > 1) message("\tAnnotating cores and clusters")
    outputList[[uniqueandprimary]] <- BM.CORES
    outputList[[allalignments]] <- CLUSTERS
    outputList <- PICBannotate(outputList, IN.ALIGNMENTS, ,
        REFERENCE.GENOME = REFERENCE.GENOME,
        SEQ.LEVELS.STYLE = SEQ.LEVELS.STYLE, PROVIDE.NON.NORMALIZED = TRUE, COMPUTE.1U.10A.FRACTIONS = COMPUTE.1U.10A.FRACTIONS, LIBRARY.SIZE = LIBRARY.SIZE
    )

    ##
    ## REPORT

    if (VERBOSITY > 1) message("\n\tClusters stats: ")
    used_uniq <- sum(GenomicRanges::mcols(outputList[[allalignments]])[["uniq_reads"]])
    used_uniq_piRNA <- round(((used_uniq / length(IN.ALIGNMENTS$unique)) * 100), 3)
    used_mult <- sum(GenomicRanges::mcols(outputList[[allalignments]])[["multimapping_reads_primary_alignments"]])
    used_mult_primary_piRNA <- round(((used_mult / length(WGRMP)) * 100), 3)
    if (VERBOSITY > 1) {
        message("\t\tAccomodated UNIQUE MAPPERS: ", used_uniq, " (", used_uniq_piRNA, " %)")
        message("\t\tAccomodated PRIMARY MULTI MAPPERS: ", used_mult, " (", used_mult_primary_piRNA, " %)")
    }

    if (!PROVIDE.NON.NORMALIZED) { # removing stats hard to understand
        for (t in c(uniqueonly, uniqueandprimary, allalignments)) {
            GenomicRanges::mcols(outputList[[t]])[["uniq_reads"]] <- NULL
            GenomicRanges::mcols(outputList[[t]])[["uniq_sequences"]] <- NULL
            GenomicRanges::mcols(outputList[[t]])[["multimapping_reads_primary_alignments"]] <- NULL
            GenomicRanges::mcols(outputList[[t]])[["all_reads_primary_alignments"]] <- NULL
            GenomicRanges::mcols(outputList[[t]])[["width_covered_by_unique_alignments"]] <- NULL
        }
    }
    if (VERBOSITY > 0) message("\nDone!")
    return(outputList)
}

utils::globalVariables(c("queryHits", "oneT", "tenA"))
#' Annotate GRanges according to a piRNA library
#'
#' @param INPUT.GRANGES GRanges (seeds/cores/clusters) to annotate
#' @param ALIGNMENTS list of alignments from PICBload
#' @param REFERENCE.GENOME name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"
#' @param REPLICATE.NAME name of the replicate. NULL by default.
#' @param LIBRARY.SIZE number of reads in the library. By default computed as number of unique mapping alignments + number of primary multimapping alignments.
#' @param PROVIDE.NON.NORMALIZED provide annotations in non-normalized format. False by default.
#' @param SEQ.LEVELS.STYLE naming of chromosomes style. "UCSC" by default.
#' @param COMPUTE.1U.10A.FRACTIONS for each locus and each alignments type (unique mapping, primary multimapping, secodnary multimapping) compute fraction 1U and 10A containing reads overlapping the locus. Default FALSE.
#'
#' @author Aleksandr Friman
#' @return GRanges object with extra annotation columns
#' @export
#'
#' @examples
#' library(BSgenome.Dmelanogaster.UCSC.dm6)
#' myGenome <- "BSgenome.Dmelanogaster.UCSC.dm6"
#' myAlignmentsFromPICBload <- PICBload(
#'     BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
#'     REFERENCE.GENOME = myGenome,
#'     VERBOSE = FALSE
#' )
#' myRangesFromPICBbuild <- PICBbuild(
#'     IN.ALIGNMENTS = myAlignmentsFromPICBload,
#'     REFERENCE.GENOME = myGenome,
#'     VERBOSITY = 0
#' )
#'
#' myClustersFromPICBbuildAnnotationsRemoved <- GenomicRanges::granges(myRangesFromPICBbuild$clusters)
#'
#' PICBannotate(
#'     INPUT.GRANGES = myClustersFromPICBbuildAnnotationsRemoved,
#'     ALIGNMENTS = myAlignmentsFromPICBload,
#'     REFERENCE.GENOME = myGenome,
#'     PROVIDE.NON.NORMALIZED = TRUE
#' )
PICBannotate <- function(
        INPUT.GRANGES,
        ALIGNMENTS,
        REFERENCE.GENOME = NULL,
        REPLICATE.NAME = NULL,
        LIBRARY.SIZE = length(ALIGNMENTS$unique) + length(ALIGNMENTS$multi.primary),
        PROVIDE.NON.NORMALIZED = FALSE,
        SEQ.LEVELS.STYLE = "UCSC",
        COMPUTE.1U.10A.FRACTIONS = FALSE
    ) {
    if (is.null(REFERENCE.GENOME)) stop("Please provide REFERENCE.GENOME")
    if (!is.null(REPLICATE.NAME)) {
        SUFFIX <- paste0(".", REPLICATE.NAME)
    } else {
        SUFFIX <- ""
    }

    SI <- PICBgetchromosomes(REFERENCE.GENOME, SEQ.LEVELS.STYLE)$SeqInfo

    PICBfixCoverage <- function(inCov, SI) {
        for (i in seq_along(SI)) {
            chrname <- Seqinfo::seqnames(SI)[i]
            if (chrname %in% names(inCov)) {
                targetLen <- Seqinfo::seqlengths(SI)[chrname]
                currentLen <- length(inCov[[chrname]])
                deltaLen <- targetLen - currentLen
                if (deltaLen > 0) {
                    inCov[[chrname]] <- c(inCov[[chrname]], rep(0, deltaLen))
                }
            }
        }
        return(inCov)
    }

    PICBannotateGranges <- function(INPUT.GRANGES, ALIGNMENTS,
                                    SI, LIBRARY.SIZE,
                                    PROVIDE.NON.NORMALIZED, SEQ.LEVELS.STYLE, COMPUTE.1U.10A.FRACTIONS) {
        GenomicRanges::mcols(INPUT.GRANGES)[["width_in_nt"]] <- GenomicRanges::width(INPUT.GRANGES)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_reads", SUFFIX)]] <- GenomicRanges::countOverlaps(INPUT.GRANGES, ALIGNMENTS$unique)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("multimapping_reads_primary_alignments", SUFFIX)]] <- GenomicRanges::countOverlaps(INPUT.GRANGES, ALIGNMENTS$multi.primary)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("all_reads_primary_alignments", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_reads", SUFFIX)]] + GenomicRanges::mcols(INPUT.GRANGES)[[paste0("multimapping_reads_primary_alignments", SUFFIX)]]

        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_reads_FPM", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_reads", SUFFIX)]] * 1e6 / LIBRARY.SIZE
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("multimapping_reads_primary_alignments_FPM", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("multimapping_reads_primary_alignments", SUFFIX)]] * 1e6 / LIBRARY.SIZE
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("all_reads_primary_alignments_FPM", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("all_reads_primary_alignments", SUFFIX)]] * 1e6 / LIBRARY.SIZE

        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_reads_FPKM", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_reads_FPM", SUFFIX)]] * 1e3 / GenomicRanges::width(INPUT.GRANGES)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("multimapping_reads_primary_alignments_FPKM", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("multimapping_reads_primary_alignments_FPM", SUFFIX)]] * 1e3 / GenomicRanges::width(INPUT.GRANGES)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("all_reads_primary_alignments_FPKM", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("all_reads_primary_alignments_FPM", SUFFIX)]] * 1e3 / GenomicRanges::width(INPUT.GRANGES)
        # coverage
        UniqueMappingCoveragePlus <- GenomicRanges::coverage(ALIGNMENTS$unique[GenomicRanges::strand(ALIGNMENTS$unique) == "+"])
        UniqueMappingCoveragePlus <- PICBfixCoverage(UniqueMappingCoveragePlus, SI)
        UniqueMappingCoverageMinus <- GenomicRanges::coverage(ALIGNMENTS$unique[GenomicRanges::strand(ALIGNMENTS$unique) == "-"])
        UniqueMappingCoverageMinus <- PICBfixCoverage(UniqueMappingCoverageMinus, SI)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("width_covered_by_unique_alignments", SUFFIX)]][as.vector(GenomicRanges::strand(INPUT.GRANGES) == "+")] <- sum(UniqueMappingCoveragePlus[INPUT.GRANGES[GenomicRanges::strand(INPUT.GRANGES) == "+"]] > 0)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("width_covered_by_unique_alignments", SUFFIX)]][as.vector(GenomicRanges::strand(INPUT.GRANGES) == "-")] <- sum(UniqueMappingCoverageMinus[INPUT.GRANGES[GenomicRanges::strand(INPUT.GRANGES) == "-"]] > 0)
        GenomicRanges::mcols(INPUT.GRANGES)[[paste0("fraction_of_width_covered_by_unique_alignments", SUFFIX)]] <- GenomicRanges::mcols(INPUT.GRANGES)[[paste0("width_covered_by_unique_alignments", SUFFIX)]] / GenomicRanges::width(INPUT.GRANGES)

        if (COMPUTE.1U.10A.FRACTIONS) {
            for (alignmentType in c("unique", "multi.primary", "multi.secondary")) {
            alignmentsAtWork <- ALIGNMENTS[[alignmentType]]
                if ("seq" %in% colnames(GenomicRanges::mcols(alignmentsAtWork))) {
                    GenomicRanges::mcols(alignmentsAtWork)[["seq"]] <- as.character(GenomicRanges::mcols(alignmentsAtWork)[["seq"]])
                    pairsAtWork <- GenomicRanges::findOverlaps(INPUT.GRANGES, alignmentsAtWork)
                    pairsAtWork.DF <- data.frame(
                    queryHits = S4Vectors::queryHits(pairsAtWork),
                    subjectHits = S4Vectors::subjectHits(pairsAtWork)
                    )
                    pairsAtWork.DF$seq <- GenomicRanges::mcols(alignmentsAtWork)[["seq"]][pairsAtWork.DF$subjectHits]
                    pairsAtWork.DF$oneT <- (substring(pairsAtWork.DF$seq, 1, 1) == "T")
                    pairsAtWork.DF$tenA <- (substring(pairsAtWork.DF$seq, 10, 10) == "A")
                    pairsAtWork.DF_by_cluster <- dplyr::group_by(pairsAtWork.DF, queryHits)
                    summedDF <- dplyr::summarise(pairsAtWork.DF_by_cluster, oneT = sum(oneT), tenA = sum(tenA), normCoef = dplyr::n())
                    summedDF$oneT <- summedDF$oneT / summedDF$normCoef
                    summedDF$tenA <- summedDF$tenA / summedDF$normCoef

                    GenomicRanges::mcols(INPUT.GRANGES)[[paste0("oneU.frac.", alignmentType, SUFFIX)]] <- 0
                    GenomicRanges::mcols(INPUT.GRANGES)[[paste0("oneU.frac.", alignmentType, SUFFIX)]][summedDF$queryHits] <- summedDF$oneT
                    GenomicRanges::mcols(INPUT.GRANGES)[[paste0("tenA.frac.", alignmentType, SUFFIX)]] <- 0
                    GenomicRanges::mcols(INPUT.GRANGES)[[paste0("tenA.frac.", alignmentType, SUFFIX)]][summedDF$queryHits] <- summedDF$tenA
                } else {
                    stop("Alignments '", alignmentType, "' does not contain sequence ('seq' field). Run PICBload with parameter GET.ORIGINAL.SEQUENCE=TRUE to get 'seq' column. Cannot COMPUTE.1U.10A.FRACTIONS.")
                }
            }
        }

        if (!PROVIDE.NON.NORMALIZED) {
            GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_reads", SUFFIX)]] <- NULL
            GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_sequences", SUFFIX)]] <- NULL
            GenomicRanges::mcols(INPUT.GRANGES)[[paste0("multimapping_reads_primary_alignments", SUFFIX)]] <- NULL
            GenomicRanges::mcols(INPUT.GRANGES)[[paste0("all_reads_primary_alignments", SUFFIX)]] <- NULL
            GenomicRanges::mcols(INPUT.GRANGES)[[paste0("width_covered_by_unique_alignments", SUFFIX)]] <- NULL
        } else {
            GenomicRanges::mcols(INPUT.GRANGES)[[paste0("uniq_sequences", SUFFIX)]] <- GenomicRanges::countOverlaps(INPUT.GRANGES, unique(ALIGNMENTS$unique))
        }
        return(INPUT.GRANGES)
    }
    # identification of cluster types
    PICBannotateTypesOfClusters <- function(INPUT.CLUSTERS, INPUT.CORES, SI) {
        INPUT.CLUSTERS$cores_intersected <- GenomicRanges::countOverlaps(INPUT.CLUSTERS, INPUT.CORES)
        INPUT.CORES.plus <- INPUT.CORES[GenomicRanges::strand(INPUT.CORES) == "+"]
        INPUT.CORES.minus <- INPUT.CORES[GenomicRanges::strand(INPUT.CORES) == "-"]
        CORES.COV.PLUS <- GenomicRanges::coverage(INPUT.CORES.plus)
        CORES.COV.PLUS <- PICBfixCoverage(inCov = CORES.COV.PLUS, SI = SI)
        CORES.COV.MINUS <- GenomicRanges::coverage(INPUT.CORES.minus)
        CORES.COV.MINUS <- PICBfixCoverage(inCov = CORES.COV.MINUS, SI = SI)
        INPUT.CLUSTERS.plus <- INPUT.CLUSTERS[GenomicRanges::strand(INPUT.CLUSTERS) == "+"]
        INPUT.CLUSTERS.minus <- INPUT.CLUSTERS[GenomicRanges::strand(INPUT.CLUSTERS) == "-"]
        INPUT.CLUSTERS.plus$core_coverage <- sum(CORES.COV.PLUS[INPUT.CLUSTERS.plus])
        INPUT.CLUSTERS.minus$core_coverage <- sum(CORES.COV.MINUS[INPUT.CLUSTERS.minus])
        INPUT.CLUSTERS <- c(INPUT.CLUSTERS.plus, INPUT.CLUSTERS.minus)
        INPUT.CLUSTERS$core_coverage_fraction <- INPUT.CLUSTERS$core_coverage / GenomicRanges::width(INPUT.CLUSTERS)
        GenomicRanges::mcols(INPUT.CLUSTERS)[[paste0("type", SUFFIX)]] <- NA
        GenomicRanges::mcols(INPUT.CLUSTERS)[[paste0("type", SUFFIX)]][INPUT.CLUSTERS$cores_intersected > 1] <- "MultiCore"
        GenomicRanges::mcols(INPUT.CLUSTERS)[[paste0("type", SUFFIX)]][(INPUT.CLUSTERS$cores_intersected == 1) & (INPUT.CLUSTERS$core_coverage_fraction < 1)] <- "ExtendedCore"
        GenomicRanges::mcols(INPUT.CLUSTERS)[[paste0("type", SUFFIX)]][(INPUT.CLUSTERS$cores_intersected == 1) & (INPUT.CLUSTERS$core_coverage_fraction == 1)] <- "SingleCore"
        # cleaning up
        INPUT.CLUSTERS$cores_intersected <- NULL
        INPUT.CLUSTERS$core_coverage <- NULL
        INPUT.CLUSTERS$core_coverage_fraction <- NULL
        return(INPUT.CLUSTERS)
    }
    # processing the input genomic intervals
    if (typeof(INPUT.GRANGES) == "list") {
        for (t in intersect(c(uniqueonly, uniqueandprimary, allalignments), names(INPUT.GRANGES))) {
            if (length(INPUT.GRANGES[[t]]) > 0) {
                INPUT.GRANGES[[t]] <- PICBannotateGranges(INPUT.GRANGES[[t]], ALIGNMENTS,
                    SI, LIBRARY.SIZE,
                    PROVIDE.NON.NORMALIZED, SEQ.LEVELS.STYLE,
                    COMPUTE.1U.10A.FRACTIONS = COMPUTE.1U.10A.FRACTIONS
                )
            }
        }
        if (allalignments %in% names(INPUT.GRANGES) && length(INPUT.GRANGES[[allalignments]]) > 0) {
            INPUT.GRANGES[[allalignments]] <- PICBannotateTypesOfClusters(INPUT.GRANGES[[allalignments]], INPUT.GRANGES[[uniqueandprimary]], SI)
        }
    } else {
        if (length(INPUT.GRANGES) > 0) {
            INPUT.GRANGES <- PICBannotateGranges(INPUT.GRANGES, ALIGNMENTS,
                SI, LIBRARY.SIZE,
                PROVIDE.NON.NORMALIZED, SEQ.LEVELS.STYLE,
                COMPUTE.1U.10A.FRACTIONS = COMPUTE.1U.10A.FRACTIONS
            )
        }
    }

    return(INPUT.GRANGES)
}

---
title: "Introduction to the piRNA Cluster Builder (PICB)"
shorttitle: "The piRNA Cluster Builder (PICB)"
author: "Haase Lab"
package: PICB
output: 
    BiocStyle::html_document:
        toc_float: true
abstract: >
    piRNAs (short for PIWI-interacting RNAs) and their PIWI protein partners play a key role in fertility 
    and maintaining genome integrity by restricting mobile genetic elements (transposons) in germ cells. 
    piRNAs originate from genomic regions known as piRNA clusters. The piRNA Cluster Builder (PICB) is 
    a versatile toolkit designed to identify genomic regions with a high density of piRNAs. It constructs 
    piRNA clusters through a stepwise integration of unique and multimapping piRNAs and offers wide-ranging 
    parameter settings, supported by an optimization function that allows users to test different parameter 
    combinations to tailor the analysis to their specific piRNA system. The output includes extensive metadata 
    columns, enabling researchers to rank clusters and extract cluster characteristics. 
    
vignette: >
    %\VignetteIndexEntry{Introduction to the piRNA Cluster Builder (PICB)}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8}
---
```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```
# Introduction
PICB is a flexible toolkit for assembling, prioritizing, and characterizing piRNA clusters. piRNAs (PIWI-interacting RNAs) and their PIWI protein partners play a key role in fertility and maintaining genome integrity by restricting mobile genetic elements (transposons) in germ cells. piRNAs originate from genomic regions called _piRNA clusters_. 

PICB identifies genomic regions with a high density of piRNAs through stepwise integration of unique and multimapping piRNAs. 

![](PICB_stepwiseIntegration.jpg)

Figure 1: PICB considers unique mapping piRNAs (NH=1), primary alignments of multimapping piRNAs (NH>1), and all possible alignments stepwise to build seeds, cores, and clusters. Find additional information in our recent [publication](https://www.sciencedirect.com/science/article/pii/S2211124724011288).

Only very limited programming knowledge is needed to run PICB. Check out the following instructions and demonstration. 

# Getting Started

It is possible to run PICB in RStudio, in an R script on your local machine or with High Performance Computing (HPC) resources or in Jupyter Notebook (using R). Keep in mind that as for any handling of large-scale sequencing data you need to have sufficient memory on your device (or cluster) allocated. 

`PICB` requires the CRAN packages `r CRANpkg("data.table")`, `r CRANpkg("seqinr")`, `r CRANpkg("openxlsx")`, `r CRANpkg("dplyr")`, 
and the Bioconductor packages `r Biocpkg("IRanges")`, `r Biocpkg("GenomicRanges")`, `r Biocpkg("GenomicAlignments")`, `r Biocpkg("Rsamtools")`, `r Biocpkg("Biostrings")`, `r Biocpkg("Seqinfo")`, `r Biocpkg("BSgenome")` and `r Biocpkg("rtracklayer")` installed. 

Load PICB in your R environment:


```{r install_dependencies, eval = FALSE}
install.packages(c(
    "data.table",
    "seqinr",
    "openxlsx",
    "dplyr"
))
if (!require("BiocManager", quietly = TRUE)) {
    install.packages("BiocManager")
}
BiocManager::install(c(
    "IRanges",
    "GenomicRanges",
    "GenomicAlignments",
    "Rsamtools",
    "Biostrings",
    "Seqinfo",
    "BSgenome",
    "rtracklayer"
))
```

Next, install PICB from Bioconductor:

```{r install_PICB, eval = FALSE}
BiocManager::install("PICB")
```

Or install the latest development version from GitHub. Note that this version may be less stable than the Bioconductor release. Make sure to have `r CRANpkg("devtools")` installed.

```{r install_PICB_dev, eval = FALSE}
devtools::install_github("HaaseLab/PICB")
```

Then load PICB in your R environment:
```{r load_PICB}
library(PICB)
```

## How to run PICB

PICB requires only two main inputs: a **BAM file** containing the piRNA alignments and the **reference genome**.

### BAM File Requirements

- **piRNA dataset**: Verify your system! piRNAs (short for PIWI-interacting RNAs) are defined by their association with PIWI proteins. Verify the **expression of PIWI proteins** before looking for piRNAs. If possible, work with **‘bona-fide’ piRNAs** that are prepared from purified PIWI-piRNA complexes [Girard et al., 2006](https://doi.org/10.1038/nature04917). Unrelated RNA fragments can be easily mistaken for ‘piRNAs’ in small RNA sequencing data. 
- **NH and NM tags**: The BAM file must include these tags. PICB stops when these tags are not included and provides a descriptive error message.
- **Coordinate-sorted and indexed**: Ensure that your BAM file is sorted and accompanied by a .bai index file. PICB's dependency, GenomicAlignments, stops if the index file is not found and provides a descriptive error message.


### Information on the demo dataset

In the following example, we demonstrate the basic workflow of PICB using a sample dataset.

For this demonstration, we utilize a subset of our mapped small RNAs from Drosophila ovaries [SRR7346241](https://www.ncbi.nlm.nih.gov/sra/?term=SRR7346241). In short, we trimmed the 3’ adaptor and filtered for reads ≥24 nucleotides. 
Next, we removed abundant cellular RNAs (rRNAs, tRNAs, snRNAs, snoRNAs) by annotation. Finally, we mapped the processed reads to the Dm6 genome using STAR ([Dobin et al., 2013](https://doi.org/10.1093/bioinformatics/bts635)) allowing up to one mismatch and 100 alignments per read. 
Using samtools ([Li et al., 2009](https://doi.org/10.1093/bioinformatics/btp352)), we extracted alignments from chr2L:20000000-21000000. 
To further reduce the file size, we retained only the essential tags for PICB (NH, NM) in the BAM file, replaced the sequence quality field (QUAL) with a placeholder (*) and downsampled the file to retain 22% of the reads. 
Lastly, we applied a high level of compression to further reduce its size and indexed the compressed and sorted BAM file. More information on the processing steps can be found in the inst/script folder.

The purpose of this entire reduction process was to significantly minimize the file size, making it manageable for storage. 
It is important to note that this downsampling, combined with the prior restriction to a smaller genomic region (chr2L:20000000-21000000), 
has drastically reduced the coverage and scope of the data, rendering it no longer representative of a biologically relevant sample.


### Reference Genome

You can provide the reference genome to PICB in **one of four ways**. Make sure to use the same reference genome version as used for mapping.

a. **Using a BSgenome package**

If your genome is available in a BSgenome package, load it directly:

```{r load_BSgenome1, results="hide", message=FALSE}
# Example for Drosophila melanogaster (make sure to install the package first), replace with your own genome
library(BSgenome.Dmelanogaster.UCSC.dm6)
myGenome <- "BSgenome.Dmelanogaster.UCSC.dm6"
```


b. **Using chromosome names and lengths**

If you have the chromosome names and lengths, create a `Seqinfo` object:
```{r load_myGenome2}
myGenome <- Seqinfo::Seqinfo(
    seqnames = c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY"),
    seqlengths = c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352)
)
```


c. **Using a Seqinfo object**

Or use an existing `Seqinfo` object:

```{r load_myGenome3}
myGenome <- Seqinfo::Seqinfo(genome = "dm6")
```

d. **Using a FASTA file**

If you have a FASTA file of the assembled genome sequence:

```{r load_myGenome4, eval = FALSE}
myGenome <- PICBloadfasta(FASTA.NAME = "/path/to/your/genome.fa")
```

# Running PICB

PICB involves two main functions:

1. `PICBload`: Loads the piRNA alignments from the BAM file.
2. `PICBbuild`: Builds the piRNA clusters from the loaded alignments.

Next to these two main functions, PICB also includes a function to optimize parameters for `PICBbuild` (`PICBoptimize`) and a function to perform strand-specific analysis (`PICBstrandanalysis`).

## Load Alignments with `PICBload`

Load the sample BAM file provided with the package or provide your own BAM file path.

```{r load_alignments}
# demonstration bam_file is in the "inst/extdata/" folder of the PICB package
bam_file <- system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB")
myAlignments <- PICBload(
    BAMFILE = bam_file,
    REFERENCE.GENOME = myGenome
)
```

The output is a list with three `GenomicRanges` objects:

- `myAlignments$unique`: Unique mapping piRNAs (read mapped only to a single location on the reference genome)
- `myAlignments$multi.primary`: Primary multimapping piRNAs (read mapped to multiple locations on the reference genome, one alignment is assigned as primary by the alignment software based on the mapping quality score. If multiple alignments have the same mapping quality score, the primary alignment is selected randomly.)
- `myAlignments$multi.secondary`: Secondary multimapping piRNAs (read mapped to multiple locations on the reference genome, all other alignments are secondary)

## Build piRNA Clusters with `PICBbuild`

Next, we want to construct the piRNA clusters using the `PICBbuild` function. Usually you would not need to include the size of the library (`LIBRARY.SIZE`) since PICB calculates it automatically. However, just for this demo, please include this parameter to adjust to the subset we chose to create this demo. 

```{r build_clusters}
myClusters <- PICBbuild(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    LIBRARY.SIZE = 12799826 #usually not necessary
)$clusters
```

We note that ranking piRNA clusters is essential for proper interpretation. PICB provides various [output columns](#output) as options to rank the clusters.

💡 Both `PICBload` and `PICBbuild` allow parameter adjustments. Refer to chapter [Parameter Adjustments](#parameter-adjustments) for insights into customization options. [PICBoptimize](#optimize-parameters) can help you find the optimal parameters for your dataset.


## Optimize parameters with `PICBoptimize`

You can use `PICBoptimize` to find the optimal parameters for `PICBbuild` for your dataset.
This is especially useful if you are not sure about the quality of your data or your reference genome.
The goal is to find a set of parameters that maximize the number of piRNAs accomodated in the clusters while minimizing genomic space occupied by the clusters. 

Make sure to provide the values for the parameters as a vector. An example is shown below:

```{r optimize_parameters}
parameterExploration <- PICBoptimize(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2, 3, 4, 5), 
    LIBRARY.SIZE = 12799826, #usually not necessary
    VERBOSITY = 1
)
```

This function works on numerical parameters (see list of parameters in [Parameter Adjustments](#parameter-adjustments)) and returns a dataframe with any combination of your chosen parameters and their corresponding number of clusters, the total width of the clusters in nucleotides, the number of reads explained by the clusters, the mean RPKM of the clusters, the fraction of reads explained by the clusters and the fraction of genomic space occupied by the clusters.

Be aware that for most datasets, the default parameters will yield great results and you do not need to optimize parameters. 

As an example, we can plot the results to find the optimal parameters for your dataset. First specify the parameter you want to optimize.

```{r example_specify_parameter}
x_column <- "MIN.UNIQUE.ALIGNMENTS.PER.WINDOW" # change parameter to optimize, if applicable
```

Then plot the results.

```{r example_plot_optimize_parameters, message=FALSE}
library(ggplot2)
# Determine a scaling factor for the secondary axis
scaling_factor <- max(parameterExploration$fraction.of.library.explained.by.clusters) / max(parameterExploration$fraction.of.genome.space.clusters)
# plot graph
ggplot(parameterExploration, aes(x = .data[[x_column]])) +
    geom_line(aes(y = fraction.of.library.explained.by.clusters * 100, color = "piRNAs Explained"), linewidth = 1) +
    geom_point(aes(y = fraction.of.library.explained.by.clusters * 100, color = "piRNAs Explained"), size = 3) +
    geom_line(aes(y = fraction.of.genome.space.clusters * scaling_factor * 100, color = "Genome Space"), linewidth = 1) +
    geom_point(aes(y = fraction.of.genome.space.clusters * scaling_factor * 100, color = "Genome Space"), size = 3) +
    scale_y_continuous(name = "piRNAs Explained by Clusters (%, piRNA sample)", sec.axis = sec_axis(~ . / scaling_factor, name = "Total piRNA cluster-length (Genome, %)")) +
    scale_x_reverse(name = paste0("Parameter chosen: ", x_column), breaks = parameterExploration[[x_column]], labels = parameterExploration[[x_column]]) +
    scale_color_manual(name = "Metrics", values = c("piRNAs Explained" = "#00a100", "Genome Space" = "black")) +
    theme_classic() +
    theme(axis.title.y.left = element_text(color = "#00a100"), axis.title.y.right = element_text(color = "black"), legend.position = "top")
```

As described earlier, we subset the dataset to include only reads from chromosome 2L and further downsampled the file to retain 22% of the reads for this demonstration.
This is why the proportion of reads explained by clusters and the genome space occupied by the clusters is so low, as shown in the plot. The plot can help you identify the optimal parameters for your dataset.
However, for this demonstration, optimization is neither necessary nor meaningful.
Refer to the [paper](https://www.sciencedirect.com/science/article/pii/S2211124724011288) for more information on how to interpret the results. 
In most cases, the default parameters yield excellent results (high piRNA coverage and low genome space occupation), so parameter optimization is usually unnecessary. 



## Strand-specific analysis with `PICBstrandanalysis`

Next to the core-PICB analysis, `PICBstrandanalysis` allows for strand-specific analysis of piRNA clusters. This computes the sense/antisense ratio of piRNAs per cluster (non-collapsed) and adds this information as a new metadata column (`s_as_ratio`). 

```{r strand_specific_analysis}
myClustersWithStrandAnalysis <- PICBstrandanalysis(
    IN.ALIGNMENTS = myAlignments,
    IN.RANGES = myClusters
)
```

To visualize the sense/antisense ratio of the clusters, you can plot the sense/antisense ratio of the clusters in a violin plot. Higher numbers indicate a higher ratio of sense to antisense piRNAs, while values close to 1 indicate an equal ratio of sense and antisense piRNAs.

In [Explore Results](#explore-results) you can see that the sense/antisense ratio is added as a new metadata column (`s_as_ratio`) and is quite low. 
This indicates that the demonstration example is dealing with dual-stranded piRNA clusters.

## Explore Results

Inspect the piRNA clusters:

```{r output_clusters}
myClustersWithStrandAnalysis
```

Explainations of the output columns are listed under the chapter [Output](#output).

## Export Results

You can export the results to an excel file using `PICBexporttoexcel`.

```{r export_results}
PICBexporttoexcel(
    IN.RANGES = myClustersWithStrandAnalysis,
    EXCEL.FILE.NAME = "myClusters_demonstration.xlsx"
)
```

When exporting clusters only (PICBbuildOutput$clusters), you receive an excel file with just one sheet containing the clusters. 
When you export all resulting list items of `PICBbuild` (PICBbuildOutput with all 3 elements: seeds, cores and clusters), you receive an excel file with three sheets containing the seeds, cores and clusters. 


## Import Results

When needed, you can import the results from an excel file using `PICBimportfromexcel`.

```{r import_results}
myClustersWithStrandAnalysis <- PICBimportfromexcel(EXCEL.FILE.NAME = "myClusters_demonstration.xlsx")
```


> 💡 Keep in mind that you just run a demo. These are not representative clusters! Though you can apply these steps to your organism of choice and sequencing data. 


# Parameter Adjustments

As previously mentioned, PICB allows wide-ranging parameter adjustments to adapt to e.g. sparse reference genomes and specific limitations of the data set. Lists of possible parameter adjustments for `PICBload` and `PICBbuild` are shown below. 

## Parameters for `PICBload`

Required parameters: 

- **BAMFILE**
- **REFERENCE.GENOME** 

Optional parameters:

- **VERBOSE** default TRUE • Allows disabling progress messages while running `PICBload`.
- **IS.SECONDARY.ALIGNMENT** default NA (all alignments) • Determines which alignment types (primary multimappers and secondary multimappers) will be loaded.
- **SEQ.LEVELS.STYLE** default "UCSC" • Allows changing the style of chromosome names in the output. Supported styles are listed in GenomeInfoDb::genomeStyles(). NA allows keeping all chromosomes from REFERENCE.GENOME and does not filter by standard linear chromosome names.
- **STANDARD.CONTIGS.ONLY** default TRUE • Determines whether alignments from non-standard contigs are used.
- **FILTER.BY.FLAG** default TRUE • Allows only those alignments with flag values present in the vector of allowed flags SELECT.FLAG. Default values of SELECT.FLAG are 0, 16, 272 and 256 (primary and secondary alignments on plus and minus strands). If FALSE, includes all flags.
- **USE.SIZE.FILTER** default TRUE • Allows filtering of alignments based on size. Default value is 18-50 nt.
- **TAGS** default c("NH","NM") • Indicates list of tags to be extracted from given bam file. The “NH” tag is required to deduce if the alignment is unique mapping or multimapping. The “NM” is required to identify mismatches if required in further analysis.
- **GET.ORIGINAL.SEQUENCE** default FALSE • Allows extraction of original read sequence from the bam file.
- **SIMPLE.CIGAR** default TRUE • Allows filtering out spliced alignments.
- **PERFECT.MATCH.ONLY** default FALSE • Allows filtering out alignments with mismatches.
- **WHAT** default c("flag") • Allows importing flags. Corresponds to “what” from ScanBamParam-class `r Biocpkg("Rsamtools")`.


## Parameters for `PICBbuild`

Required parameters: 

- **IN.ALIGNMENTS** (output of `PICBload`)
- **REFERENCE.GENOME** (reference genome previously used in `PICBload`)

The library size can be adjusted as shown in our [PICB demo](#example-workflow), however in most cases this is not necessary.

- **LIBRARY.SIZE**: default number of unique mapping alignments + number of primary multimapping alignments • number of reads in the library.
- **VERBOSITY** default 2 • Allows choosing the quantity of progress messages while running `PICBbuild`. Depending on VERBOSITY's value, printed messages are missing (0), include current processing step (1), include additionally current processing sub-step (2) or include chosen parameters for `PICBbuild`.
- **COMPUTE.1U.10A.FRACTIONS** default FALSE • Calculates the fractions of piRNAs with a 1U (Uridine at the 5' most piRNA position) and a 10A (adenine at position 10). Requirement: GET.ORIGINAL.SEQUENCE set to TRUE in `PICBload`.

PICBbuild integrates unique mapping (seeds), primary multimapping (cores) and secondary alignments stepwise using the sliding window algorithm. Each step allows following parameter adjustments. 

Adjustable parameters for processing **unique mapping alignments**:

- **UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH**  default 350 nt • Sets length of sliding windows for uniquely mapping alignments.
- **UNIQUEMAPPERS.SLIDING.WINDOW.STEP** default UNIQUEMAPPERS.SLIDING. WINDOW.WIDTH divided by 10 and rounded to the nearest integer • Sets distance between starts of the windows for uniquely mapping alignments.
- **MIN.UNIQUE.ALIGNMENTS.PER.WINDOW** default Value corresponding to 2 FPKM • Sets coverage threshold for seed discovery. Corresponds to normalized counts of uniquely mapping alignments throughout the window (in FPKM).
- **MIN.UNIQUE.SEQUENCES.PER.WINDOW** default 1 per 50 nt of window length or MIN.UNIQUE.ALIGNMENTS.PER.WINDOW whichever is smaller • Minimum number of distinct piRNA sequences within the sliding window. This parameter allows identification of clusters supported by a diverse set of piRNA sequences rather than a high number of reads from one or a few piRNA sequences. 

The called windows are reduced into **seeds** and further filtered: 

- **THRESHOLD.SEEDS.GAP** default 0 nt • Removes gaps between seeds if below the given length value.
- **MIN.SEED.LENGTH** default 2 * UNIQUEMAPPERS.SLIDING. WINDOW.WIDTH + 100 nt = 800 nt • Removes seeds below a certain length (Default requires an actual piRNA coverage of at least 100 nt: UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH (see above, default: 350 nt) at both the 5’ end and the 3’ end, thus making a 350+350+100=800 nt long seed).
- **MIN.COVERED.SEED.LENGTH** default 0 nt • Allows filtering by minimum number of seed bases covered by unique mapping alignments.

The next step includes **primary multimapping alignments** using a similar but simplified algorithm. Following parameters can be adjusted:


- **PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH**: default 350 nt • Sets lengths of sliding windows for primary multimapping alignments.
- **PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP**: default PRIMARY.MULTIMAPPERS. SLIDING.WINDOW.WIDTH divided by 10 and rounded to the nearest integer • Sets distances between starts of windows for primary multimapping alignments.
- **MIN.PRIMARY.MULTIMAPING.ALIGNMENTS.PER.WINDOW**: default 4 FPKM • Sets coverage threshold for calling primary multimapping alignments windows. Corresponds to normalized counts of primary multimapping alignments throughout the window (in FPKM).

These resulting windows are being reduced into **cores** and are further filtered:


- **THRESHOLD.CORES.GAP**: default 0 nt • Removes gaps between cores if below the given length value.

**Seeds** and **primary multimapping windows** overlapping **seeds** merge into **cores**. Standalone **seeds** are also considered **cores**. **Cores** not overlapping with a **seed** are being filtered out as their transcription cannot be verified.

In the next step **secondary alignments** are processed identically as primary multimapping alignments. This step is dependent on whether `PICBload` loaded the secondary alignments (parameter *IS.SECONDARY.ALIGNMENT*).


- **SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH**: default 1000 nt • Sets lengths of sliding windows for secondary alignments.
- **SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP**: default 100 nt • Sets distances between starts of the windows for secondary alignments.
- **MIN.SECONDARY.MULTIMAPING.ALIGNMENTS.PER.WINDOW**: default 0.2 FPKM • Sets coverage threshold for calling secondary alignments windows. Corresponds to normalized counts of secondary alignments throughout the window (in FPKM).

**Cores** and **secondary alignments** overlapping **cores** merge into **clusters**. Standalone **cores** are also considered **clusters**. piRNA **clusters** were formed and include all alignment information. 


# Output

The output of `PICBbuild` includes **seeds**, **cores** and  **clusters**, each in *GenomicRanges* format. In [Running PICB](#running-picb), we extracted directly the **clusters** using `$clusters`. Extracting the **seeds** and **cores** can be done similarly using `$seeds` and `$cores`. 

The piRNA **clusters** follow *GRanges* convention including the genomic coordinates (seqnames, ranges, and strand) and metacolumns. There are in total 9 metacolumns when running PICB with default parameters:

## Metadata Columns

- **`type`**: Type of cluster (`SingleCore`, `ExtendedCore`, `MultiCore`).
- **`width_in_nt`**: Width of the seed/core/cluster.
- **`uniq_reads_FPM`**: Uniquely mapping piRNA reads aligned to the seed/core/cluster (normalized to the number of all aligned reads)
- **`multimapping_reads_primary_alignments_FPM`**: Number of primary multimapping alignments overlapping to the seed/core/cluster (normalized to the number of all aligned reads)
- **`all_reads_primary_alignments_FPM`**: Number of primary alignments overlapping the seed/core/cluster (normalized to the number of all aligned reads)
- **`uniq_reads_FPKM`**: Number of unique reads normalized to the number of all aligned reads and to the size of the corresponding seed/core/cluster length (unique reads FPM per kilobase)
- **`multimapping_reads_primary_alignments_FPKM`**: Number of multimapping reads normalized to the number of all aligned reads and to the size of the corresponding seed/core/cluster length (primary alignments of multimapping reads FPM per kilobase)
- **`all_reads_primary_alignments_FPKM`**: Number of all reads primary alignments normalized to the number of all aligned reads and to the size of the corresponding seed/core/cluster length (primary alignments of all reads FPM per kilobase)
- **`fraction_of_width_covered_by_unique_alignments`**: Fraction of seed/core/cluster length coreved by at least 1 unique mapping piRNA read

Further analyze and visualize your piRNA clusters as needed. 
PICB provides an excellent starting point for identifying and characterizing piRNA clusters and serves as a stepping stone for further piRNA research.
# Acknowledgments

We thank all contributors and supporters of PICB.

**Authors:**

The Haase Lab at the National Institute of Health (NIH)

Visit the lab website of [Astrid D. Haase, M.D., Ph.D.](https://www.niddk.nih.gov/about-niddk/staff-directory/biography/haase-astrid)

# Session Information

```{r SessionInfo}
sessionInfo()
```

*For any questions or issues, please contact the PICB development team through [email](mailto:haase-lab-bioinfo@nih.gov).*

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBloadfasta.R
\name{PICBloadfasta}
\alias{PICBloadfasta}
\title{Get SeqInfo object from a fasta file}
\usage{
PICBloadfasta(FASTA.NAME = NULL)
}
\arguments{
\item{FASTA.NAME}{path to the fasta file}
}
\value{
SeqInfo object with all chromosome names and lengths from the fasta file
}
\description{
Get SeqInfo object from a fasta file
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)

# create temporary fasta file
chr2L_seq <- BSgenome.Dmelanogaster.UCSC.dm6[["chr2L"]]
chr2L_seq_set <- DNAStringSet(chr2L_seq)
names(chr2L_seq_set) <- "chr2L"
temp_fasta <- tempfile(fileext = ".fasta")
writeXStringSet(chr2L_seq_set, temp_fasta)

myGenome <- PICBloadfasta(FASTA.NAME = temp_fasta)
unlink(temp_fasta)
}
\author{
Aleksandr Friman
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICB.R
\docType{package}
\name{PICB-package}
\alias{PICB}
\alias{PICB-package}
\title{PICB: piRNA Cluster Builder}
\description{
piRNAs (short for PIWI-interacting RNAs) and their PIWI protein partners play a key role in fertility 
and maintaining genome integrity by restricting mobile genetic elements (transposons) in germ cells. 
piRNAs originate from genomic regions known as piRNA clusters. The piRNA Cluster Builder (PICB) is 
a versatile toolkit designed to identify genomic regions with a high density of piRNAs. It constructs 
piRNA clusters through a stepwise integration of unique and multimapping piRNAs and offers wide-ranging 
parameter settings, supported by an optimization function that allows users to test different parameter 
combinations to tailor the analysis to their specific piRNA system. The output includes extensive metadata 
columns, enabling researchers to rank clusters and extract cluster characteristics.
}
\section{Main Functions}{

The package provides several core functions:
\itemize{
  \item \code{PICBload}: Load and preprocess BAM files containing piRNA alignments
  \item \code{PICBbuild}:  Build piRNA seeds/cores/clusters from alignments
  \item \code{PICBoptimize}: Optimize parameters for cluster building
  \item \code{PICBstrandanalysis}: Add sense/antisense ratio of unique piRNAs per piRNAcluster
  \item \code{PICBannotate}: Annotate GRanges according to a piRNA library
  \item \code{PICBloadfasta}: Get SeqInfo object from a fasta file
  \item \code{PICBexporttoexcel}: Export cluster object into an Excel file
  \item \code{PICBimporttoexcel}: Import cluster object from an Excel file
}
}

\section{Workflow}{

A typical PICB workflow consists of:
\enumerate{
  \item Loading alignments with \code{PICBload}
  \item Building clusters with \code{PICBbuild}
  \item Optional parameter optimization with \code{PICBoptimize}
  \item Optional strand analysis with \code{PICBstrandanalysis}
  \item Exporting results with \code{PICBexporttoexcel}
}
}

\examples{
# 0. Load PICB
library(PICB)

# 1. Load Required Genome from e.g. Seqinfo (check all options in Vignette or ReadMe)
myGenome <- Seqinfo::Seqinfo(
    seqnames = c("chr2L", "chr2R", "chr3L", "chr3R", "chr4", "chrX", "chrY"),
    seqlengths = c(23513712, 25286936, 28110227, 32079331, 1348131, 23542271, 3667352)
)

# 2. Load Example Data and Process Alignments
bam_file <- system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", 
    package = "PICB")

myAlignments <- PICBload(
    BAMFILE = bam_file,
    REFERENCE.GENOME = myGenome,
    VERBOSE = FALSE
)

# 3. Build piRNA Clusters
myClusters <- PICBbuild(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    LIBRARY.SIZE = 12799826,  # Usually calculated automatically
    VERBOSITY = 0
)$clusters

# 4. Optimize Parameters (Optional)
parameterExploration <- PICBoptimize(
    IN.ALIGNMENTS = myAlignments,
    REFERENCE.GENOME = myGenome,
    MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2, 3, 4, 5),
    LIBRARY.SIZE = 12799826,  # Usually calculated automatically
    VERBOSITY = 1
)

# 5. Perform Strand Analysis
myClustersWithStrand <- PICBstrandanalysis(
    IN.ALIGNMENTS = myAlignments,
    IN.RANGES = myClusters
)

# 6. Export Clusters
PICBexporttoexcel(
    IN.RANGES = myClustersWithStrand,
    EXCEL.FILE.NAME = "myClusters_demonstration.xlsx"
)
# 7. Import Ranges
importedClusters <- PICBimportfromexcel(
    EXCEL.FILE.NAME = system.file("extdata", "myClusters_demonstration.xlsx", package = "PICB")
)

}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/HaaseLab/PICB}
  \item Report bugs at \url{https://github.com/HaaseLab/PICB/issues}
}

}
\author{
\strong{Maintainer}: Franziska Ahrend \email{haase-lab-bioinfo@nih.gov} (\href{https://orcid.org/0009-0004-7464-3444}{ORCID})

Authors:
\itemize{
  \item Pavol Genzor
  \item Aleksandr Friman \email{haase-lab-bioinfo@nih.gov}
  \item Daniel Stoyko
  \item Parthena Konstantinidou \email{haase-lab-bioinfo@nih.gov}
  \item Zuzana Loubalova \email{haase-lab-bioinfo@nih.gov}
  \item Yuejun Wang \email{haase-lab-bioinfo@nih.gov}
  \item Hernan Lorenzi \email{haase-lab-bioinfo@nih.gov}
  \item Astrid D Haase \email{haase-lab-bioinfo@nih.gov}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBstrandanalysis.R
\name{PICBstrandanalysis}
\alias{PICBstrandanalysis}
\title{Add sense/antisense ratio of unique piRNAs per piRNA cluster}
\usage{
PICBstrandanalysis(IN.ALIGNMENTS, IN.RANGES, VERBOSE = TRUE)
}
\arguments{
\item{IN.ALIGNMENTS}{list of alignments from PICBload}

\item{IN.RANGES}{single GRanges object (seeds, cores or clusters from PICBbuild)}

\item{VERBOSE}{enables progress output. True by default.}
}
\value{
GRanges object with an additional annotation column
}
\description{
Add sense/antisense ratio of unique piRNAs per piRNA cluster
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)
myAlignmentsFromPICBload <- PICBload(
    BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSE = FALSE
)

outputOfPICBbuild <- PICBbuild(
    IN.ALIGNMENTS = myAlignmentsFromPICBload,
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSITY = 0
)

outputOfPICBbuild$clusters <- PICBstrandanalysis(
    IN.ALIGNMENTS = myAlignmentsFromPICBload,
    IN.RANGES = outputOfPICBbuild$clusters
)
}
\author{
Parthena Konstantinidou

Zuzana Loubalova

Franziska Ahrend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBload.R
\name{PICBload}
\alias{PICBload}
\title{Load and preprocess BAM files containing piRNA alignments}
\usage{
PICBload(
  BAMFILE = NULL,
  REFERENCE.GENOME = NULL,
  SIMPLE.CIGAR = TRUE,
  IS.SECONDARY.ALIGNMENT = NA,
  STANDARD.CONTIGS.ONLY = TRUE,
  PERFECT.MATCH.ONLY = FALSE,
  FILTER.BY.FLAG = TRUE,
  SELECT.FLAG = c(0, 16, 272, 256),
  USE.SIZE.FILTER = TRUE,
  READ.SIZE.RANGE = c(18, 50),
  TAGS = c("NH", "NM"),
  WHAT = c("flag"),
  SEQ.LEVELS.STYLE = "UCSC",
  GET.ORIGINAL.SEQUENCE = FALSE,
  VERBOSE = TRUE
)
}
\arguments{
\item{BAMFILE}{name of the bam file to load. Should be sorted and indexed.}

\item{REFERENCE.GENOME}{name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"}

\item{SIMPLE.CIGAR}{simpleCigar parameter of Rsamtools::ScanBamParam}

\item{IS.SECONDARY.ALIGNMENT}{defines loading of primary/secondary alignments. Default value NA loads both primary and secondary.}

\item{STANDARD.CONTIGS.ONLY}{use only standard chromosomes}

\item{PERFECT.MATCH.ONLY}{load only alignments without mismatches}

\item{FILTER.BY.FLAG}{enables filtering by flag. TRUE by default.}

\item{SELECT.FLAG}{vector of flags to use. Default value c(0,16, 272, 256).}

\item{USE.SIZE.FILTER}{enables filter by alignment size. True by default.}

\item{READ.SIZE.RANGE}{allowed alignment sizes. c(18,50) by default.}

\item{TAGS}{tags to import from bam file. c("NH","NM") by default.}

\item{WHAT}{"what" parameter of Rsamtools::ScanBamParam. c("flag") by default.}

\item{SEQ.LEVELS.STYLE}{naming of chromosomes style. "UCSC" by default.}

\item{GET.ORIGINAL.SEQUENCE}{adds "seq" to WHAT. False by default.}

\item{VERBOSE}{enables progress output. True by default.}
}
\value{
list of GRanges objects named "unique" for unique mapping alignments,
"multi.primary" for primary multimapping alignments,
"multi.secondary" for secondary multimapping alignments
}
\description{
Load and preprocess BAM files containing piRNA alignments
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)
PICBload(
    BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSE = FALSE
)

}
\author{
Pavol Genzor

Daniel Stoyko

Aleksandr Friman

Franziska Ahrend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBbuild.R
\name{PICBbuild}
\alias{PICBbuild}
\title{Build piRNA seeds/cores/clusters from alignments}
\usage{
PICBbuild(
  IN.ALIGNMENTS,
  REFERENCE.GENOME,
  UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH = 350,
  UNIQUEMAPPERS.SLIDING.WINDOW.STEP = round(UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH/10, 0),
  PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH = 350,
  PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP =
    round(PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH/10, 0),
  SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH = 1000,
  SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP =
    round(SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH/10, 0),
  LIBRARY.SIZE = length(IN.ALIGNMENTS$unique) + length(IN.ALIGNMENTS$multi.primary),
  MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = 2 * (UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH/1000) *
    (LIBRARY.SIZE/1e+06),
  MIN.UNIQUE.SEQUENCES.PER.WINDOW = min(MIN.UNIQUE.ALIGNMENTS.PER.WINDOW,
    round(UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH/50, 0)),
  MIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW = 4 *
    (PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH/1000) * (LIBRARY.SIZE/1e+06),
  MIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW = 0.2 *
    (SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH/1000) * (LIBRARY.SIZE/1e+06),
  MIN.SEED.LENGTH = 2 * UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH + 100,
  MIN.COVERED.SEED.LENGTH = 0,
  THRESHOLD.SEEDS.GAP = 0,
  THRESHOLD.CORES.GAP = 0,
  THRESHOLD.CLUSTERS.GAP = 0,
  SEQ.LEVELS.STYLE = "UCSC",
  MIN.OVERLAP = 5,
  PROVIDE.NON.NORMALIZED = FALSE,
  COMPUTE.1U.10A.FRACTIONS = FALSE,
  VERBOSITY = 2
)
}
\arguments{
\item{IN.ALIGNMENTS}{list of alignments from PICBload}

\item{REFERENCE.GENOME}{name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"}

\item{UNIQUEMAPPERS.SLIDING.WINDOW.WIDTH}{width of sliding window for unique mappers. 350 nt by default}

\item{UNIQUEMAPPERS.SLIDING.WINDOW.STEP}{step of sliding windows for unique mappers. width/10 by default}

\item{PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH}{width of sliding window for primary multimapping alignments. 350 nt by default}

\item{PRIMARY.MULTIMAPPERS.SLIDING.WINDOW.STEP}{step of sliding windows for primary multimapping alignments. width/10 by default}

\item{SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.WIDTH}{width of sliding window for secondary multimapping alignments. 1000 nt by default}

\item{SECONDARY.MULTIMAPPERS.SLIDING.WINDOW.STEP}{step of sliding windows for secondary multimapping alignments. width/10 by default}

\item{LIBRARY.SIZE}{number of reads in the library. By default computed as number of unique mapping alignments + number of primary multimapping alignments.}

\item{MIN.UNIQUE.ALIGNMENTS.PER.WINDOW}{absolute number of unique mapping alignments per window to call it. By default computed as 2 FPKM.}

\item{MIN.UNIQUE.SEQUENCES.PER.WINDOW}{absolute number of unique mapping sequences per window to call it. By default computed as width/50.}

\item{MIN.PRIMARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW}{absolute number of primary multimapping alignments per window to call it. By default computed as 4 FPKM.}

\item{MIN.SECONDARY.MULTIMAPPING.ALIGNMENTS.PER.WINDOW}{absolute number of secondary multimapping alignments per window to call it. By default computed as 0.2 FPKM.}

\item{MIN.SEED.LENGTH}{minimum length of a seed. By default computed as 2x unique mapper window size + 100.}

\item{MIN.COVERED.SEED.LENGTH}{minimum number of seed nucleotides covered by unique mappers. 0 by default.}

\item{THRESHOLD.SEEDS.GAP}{minimum gap between seeds to not merge them. 0 by default.}

\item{THRESHOLD.CORES.GAP}{minimum gap between cores to not merge them. 0 by default.}

\item{THRESHOLD.CLUSTERS.GAP}{minimum gap between clusters to not merge them. 0 by default.}

\item{SEQ.LEVELS.STYLE}{naming of chromosomes style. "UCSC" by default.}

\item{MIN.OVERLAP}{minimum overlap between seeds and cores, as well as between cores and clusters 5 nt by default.}

\item{PROVIDE.NON.NORMALIZED}{include non-normalized to the library size statistics in the output annotations}

\item{COMPUTE.1U.10A.FRACTIONS}{for each locus and each alignments type (unique mapping, primary multimapping, secodnary multimapping) compute fraction 1U and 10A containing reads overlapping the locus. Default FALSE.}

\item{VERBOSITY}{verbosity level 0/1/2/3. 2 by default.}
}
\value{
list of annotated GRanges objects named "seeds" for seeds,
"cores" for cores,
"clusters" for clusters
}
\description{
Build piRNA seeds/cores/clusters from alignments
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)
myAlignmentsFromPICBload <- PICBload(
    BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSE = FALSE
)

outputOfPICBbuild <- PICBbuild(
    IN.ALIGNMENTS = myAlignmentsFromPICBload,
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSITY = 0
)
}
\author{
Pavol Genzor

Daniel Stoyko

Aleksandr Friman

Franziska Ahrend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBexporttoexcel.R
\name{PICBexporttoexcel}
\alias{PICBexporttoexcel}
\title{Export cluster object into an Excel file}
\usage{
PICBexporttoexcel(IN.RANGES = NULL, EXCEL.FILE.NAME = NULL)
}
\arguments{
\item{IN.RANGES}{clustering object to export}

\item{EXCEL.FILE.NAME}{file name to save}
}
\value{
no values returned
}
\description{
Export cluster object into an Excel file
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)
myAlignmentsFromPICBload <- PICBload(
    BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSE = FALSE
)

outputOfPICBbuild <- PICBbuild(
    IN.ALIGNMENTS = myAlignmentsFromPICBload,
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    LIBRARY.SIZE = 12799826, #usually not necessary
    VERBOSITY = 0
)
PICBexporttoexcel(
    IN.RANGES = outputOfPICBbuild,
    EXCEL.FILE.NAME = "inst/extdata/myClusters_demonstration.xlsx"
)
}
\author{
Aleksandr Friman

Franziska Ahrend
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBimportfromexcel.R
\name{PICBimportfromexcel}
\alias{PICBimportfromexcel}
\title{Import cluster object from an Excel file}
\usage{
PICBimportfromexcel(EXCEL.FILE.NAME = NULL)
}
\arguments{
\item{EXCEL.FILE.NAME}{file name to import from}
}
\value{
list of annotated GRanges objects named "seeds" for seeds,
"cores" for cores,
"clusters" for clusters
}
\description{
Import cluster object from an Excel file
}
\examples{
importedClusters <- PICBimportfromexcel(
    EXCEL.FILE.NAME = system.file("extdata", "myClusters_demonstration.xlsx", package = "PICB")
)
}
\author{
Aleksandr Friman
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBannotate.R
\name{PICBannotate}
\alias{PICBannotate}
\title{Annotate GRanges according to a piRNA library}
\usage{
PICBannotate(
  INPUT.GRANGES,
  ALIGNMENTS,
  REFERENCE.GENOME = NULL,
  REPLICATE.NAME = NULL,
  LIBRARY.SIZE = length(ALIGNMENTS$unique) + length(ALIGNMENTS$multi.primary),
  PROVIDE.NON.NORMALIZED = FALSE,
  SEQ.LEVELS.STYLE = "UCSC",
  COMPUTE.1U.10A.FRACTIONS = FALSE
)
}
\arguments{
\item{INPUT.GRANGES}{GRanges (seeds/cores/clusters) to annotate}

\item{ALIGNMENTS}{list of alignments from PICBload}

\item{REFERENCE.GENOME}{name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"}

\item{REPLICATE.NAME}{name of the replicate. NULL by default.}

\item{LIBRARY.SIZE}{number of reads in the library. By default computed as number of unique mapping alignments + number of primary multimapping alignments.}

\item{PROVIDE.NON.NORMALIZED}{provide annotations in non-normalized format. False by default.}

\item{SEQ.LEVELS.STYLE}{naming of chromosomes style. "UCSC" by default.}

\item{COMPUTE.1U.10A.FRACTIONS}{for each locus and each alignments type (unique mapping, primary multimapping, secodnary multimapping) compute fraction 1U and 10A containing reads overlapping the locus. Default FALSE.}
}
\value{
GRanges object with extra annotation columns
}
\description{
Annotate GRanges according to a piRNA library
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)
myGenome <- "BSgenome.Dmelanogaster.UCSC.dm6"
myAlignmentsFromPICBload <- PICBload(
    BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
    REFERENCE.GENOME = myGenome,
    VERBOSE = FALSE
)
myRangesFromPICBbuild <- PICBbuild(
    IN.ALIGNMENTS = myAlignmentsFromPICBload,
    REFERENCE.GENOME = myGenome,
    VERBOSITY = 0
)

myClustersFromPICBbuildAnnotationsRemoved <- GenomicRanges::granges(myRangesFromPICBbuild$clusters)

PICBannotate(
    INPUT.GRANGES = myClustersFromPICBbuildAnnotationsRemoved,
    ALIGNMENTS = myAlignmentsFromPICBload,
    REFERENCE.GENOME = myGenome,
    PROVIDE.NON.NORMALIZED = TRUE
)
}
\author{
Aleksandr Friman
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBoptimize.R
\name{PICBoptimize}
\alias{PICBoptimize}
\title{Runs PICBbuild multiple times with provided parameters and returns optimization data frame.}
\usage{
PICBoptimize(
  IN.ALIGNMENTS,
  REFERENCE.GENOME,
  LIBRARY.SIZE = length(IN.ALIGNMENTS$unique) + length(IN.ALIGNMENTS$multi.primary),
  VERBOSITY = 2,
  PROVIDE.INFO.SEEDS.AND.CORES = FALSE,
  SEQ.LEVELS.STYLE = "UCSC",
  ...
)
}
\arguments{
\item{IN.ALIGNMENTS}{list of alignments from PICBload}

\item{REFERENCE.GENOME}{name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6"}

\item{LIBRARY.SIZE}{number of reads in the library. By default computed as number of unique mapping alignments + number of primary multimapping alignments.}

\item{VERBOSITY}{verbosity level 0/1/2/3. 2 by default.}

\item{PROVIDE.INFO.SEEDS.AND.CORES}{FALSE by default.}

\item{SEQ.LEVELS.STYLE}{naming of chromosomes style. "UCSC" by default.}

\item{...}{rest of the parameters used by PICBbuild and provided as iteratable vectors}
}
\value{
Optimization values dataframe
}
\description{
Runs PICBbuild multiple times with provided parameters and returns optimization data frame.
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)
myAlignmentsFromPICBload <- PICBload(
    BAMFILE = system.file("extdata", "Fly_Ov1_chr2L_20To21mb_filtered.bam", package = "PICB"),
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    VERBOSE = FALSE
)

PICBoptimize(
    IN.ALIGNMENTS = myAlignmentsFromPICBload,
    REFERENCE.GENOME = "BSgenome.Dmelanogaster.UCSC.dm6",
    MIN.UNIQUE.ALIGNMENTS.PER.WINDOW = c(1, 2, 3, 4, 5)
)
}
\author{
Aleksandr Friman
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PICBgetchromosomes.R
\name{PICBgetchromosomes}
\alias{PICBgetchromosomes}
\title{Get SeqInfo object from standard non-circular chromosome names from your genome}
\usage{
PICBgetchromosomes(REFERENCE.GENOME, SEQ.LEVELS.STYLE = "UCSC")
}
\arguments{
\item{REFERENCE.GENOME}{name of genome. For example "BSgenome.Dmelanogaster.UCSC.dm6", or directly a SeqInfo object.}

\item{SEQ.LEVELS.STYLE}{naming of chromosomes style. "UCSC" by default.}
}
\value{
SeqInfo object with standard non-circular chromosome names
}
\description{
Get SeqInfo object from standard non-circular chromosome names from your genome
}
\examples{
library(BSgenome.Dmelanogaster.UCSC.dm6)
mySI <- PICBgetchromosomes("BSgenome.Dmelanogaster.UCSC.dm6", "UCSC")
}
\author{
Aleksandr Friman

Franziska Ahrend
}
